/**
 * @file MinosEventBuilder.cpp
 * @date 9 sept. 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the DaqMinos software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "MinosEventBuilder.h"
#include <minos/fem/SourceEvent.h>
#include <utl/compat/ArrayHelper.hpp>
#include <utl/Logging.h>

namespace minos
{
namespace acq
{
namespace back
{
using namespace utl::io;
using namespace mdcs::acq;
using namespace mdcs::utl;
using namespace fem;
namespace array = ::utl::array;

// =============================================================================
// minos::acq::back::MinosEventBuilder::Params class methods body
// -----------------------------------------------------------------------------

size_t MinosEventBuilder::Params::countActiveSources() const
{
	size_t count = 0;
	for(size_t i = 0; i < array::size(activeSources); ++i)
	{
		if(activeSources[i]) ++count;
	}
	return count;
}

// =============================================================================
// minos::acq::back::MinosEventBuilder::Setup class methods body
// -----------------------------------------------------------------------------

MinosEventBuilder::Setup::Setup(ApplicationState& state, const Params& params)
	: BaseSetupData<Policy>(state, params),
	  activeSourcesCount(params.countActiveSources()),
	  datablock(new uint8_t[activeSourcesCount * (params.maxSrcEventSize_B + params.bufferSize_B)])
{
}

// =============================================================================
// minos::acq::back::MinosEventBuilder::RunImpl class methods body
// -----------------------------------------------------------------------------

MinosEventBuilder::RunImpl::RunImpl(ApplicationState& state, Setup& setup, const RunInfo& runInfo)
	: BaseRunImpl<Policy>(state, setup, runInfo),
	  eventBuffer(setup.datablock.get()),
	  emptyBufferCount(setup.activeSourcesCount),
	  lastEventId(0)
{
	LOG_DEBUG() << "EMPTY: " << emptyBufferCount << " buffers";
	for(size_t i = 0; i < array::size(maskedSources); ++i)
	{
		maskedSources[i] = false;
	}
	uint8_t* inputPtr = eventBuffer;
	inputPtr += params.maxSrcEventSize_B * setup.activeSourcesCount;
	for(size_t i = 0; i < array::size(sourcesBuffer); ++i)
	{
		if(params.activeSources[i])
		{
			sourcesBuffer[i].init(inputPtr, params.bufferSize_B);
			inputPtr += params.bufferSize_B;
		}
	}
}

MinosEventBuilder::RunImpl::~RunImpl()
{
}

void MinosEventBuilder::RunImpl::process(const uint8_t* buf, uint32_t nBytes)
{
	SourceEvent::UntilCard srcEvent(buffer(buf, nBytes));
	uint16_t feminosId = srcEvent.card();

	LOG_DEBUG() << "Event received: Fem #" << feminosId << " id=" << srcEvent.eventId() << " ts=" << srcEvent.timestamp() << " size=" << nBytes;

	if(maskedSources[feminosId])
	{
		LOG_WARN() << "Receiving data from disabled Feminos board, enabling it !";
		maskedSources[feminosId] = false;
		++emptyBufferCount;
	}

	if(srcEvent.eventId() > lastEventId)
	{
		if(feminosId >= fem::FeminosMaxCount)
		{
			throw std::range_error("out of range feminos id!");
		}
		if(!params.activeSources[feminosId])
		{
			throw std::runtime_error("receiving data from inactive feminos board!");
		}

		recordSourceEvent(buf, nBytes, sourcesBuffer[feminosId]);
		if(isAlmostFull(feminosId))
		{
			LOG_WARN() << "Fem #" << feminosId << " almost full!!";
			maskMissingSources();
		}
		processCompleteEvents();
	}
	else
	{
		LOG_WARN() << "LATE EVENT DISCARDED: Fem #" << feminosId << " id=" << srcEvent.eventId() << " ts=" << srcEvent.timestamp() << " size=" << nBytes;
	}
}

void MinosEventBuilder::RunImpl::recordSourceEvent(const uint8_t* data, uint32_t nBytes, mdcs::utl::ContiguousCircularBuffer& srcBuffer)
{
	bool wasEmpty = srcBuffer.empty();
	mdcs::utl::ContiguousCircularBuffer::WriteSession session = srcBuffer.write(nBytes + sizeof(uint32_t));
	uint8_t* wPtr = session.pointer();
	memcpy(wPtr, &nBytes, sizeof(uint32_t));
	wPtr += sizeof(uint32_t);
	memcpy(wPtr, data, nBytes);
	session.commit();
	if(wasEmpty)
	{
		--emptyBufferCount;
		LOG_DEBUG() << "EMPTY: " << emptyBufferCount << " buffers";
	}
}

bool MinosEventBuilder::RunImpl::isAlmostFull(size_t femId)
{
	return sourcesBuffer[femId].sizeMaxContiguousFree() < params.maxSrcEventSize_B;
}

void MinosEventBuilder::RunImpl::maskMissingSources()
{
	for(size_t i = 0; i < array::size(sourcesBuffer); ++i)
	{
		if(params.activeSources[i] and !maskedSources[i] and sourcesBuffer[i].empty())
		{
			LOG_WARN() << "Masking Feminos #" << i << "!!";
			maskedSources[i] = true;
			--emptyBufferCount;
			LOG_DEBUG() << "EMPTY: " << emptyBufferCount << " buffers";
		}
	}
}

void MinosEventBuilder::RunImpl::processCompleteEvents()
{
	while(emptyBufferCount == 0)
	{
		releaseOldestEvent();
	}
}

void MinosEventBuilder::RunImpl::releaseOldestEvent()
{
	LOG_DEBUG() << "Releasing oldest event";
	EventInfo eventInfo[fem::FeminosMaxCount];
	uint32_t minId;
	analyseBuffers(minId, eventInfo);

	EventInfo builtEventInfo;
	buildEvent(minId, eventInfo, builtEventInfo);

	nextStageProcess(eventBuffer, builtEventInfo.size);
}

void MinosEventBuilder::RunImpl::analyseBuffers(uint32_t& minId, EventInfoArray& eventInfo)
{
	minId = std::numeric_limits<uint32_t>::max();
	for(size_t i = 0; i < array::size(sourcesBuffer); ++i)
	{
		if(params.activeSources[i] and !maskedSources[i])
		{
			readEventInfo(sourcesBuffer[i].readPtr(), eventInfo[i]);
			minId = std::min(minId, eventInfo[i].id);
		}
		else
		{
			eventInfo[i].size = 0;
		}
	}
	LOG_DEBUG() << "oldest event ID is " << minId;
}

void MinosEventBuilder::RunImpl::buildEvent(uint32_t id, const EventInfoArray& srcEventInfo, EventInfo& builtEventInfo)
{
	lastEventId = id;
	computeBuiltEventInfo(id, srcEventInfo, builtEventInfo);

	uint8_t* writePtr = eventBuffer;
	writeEventHeader(writePtr, builtEventInfo);

	for(size_t i = 0; i < array::size(srcEventInfo); ++i)
	{
		if(srcEventInfo[i].size != 0 and srcEventInfo[i].id == id)
		{
			copySourceEvent(writePtr, srcEventInfo[i], sourcesBuffer[i]);
		}
	}
	writeEventFooter(writePtr);
}

void MinosEventBuilder::RunImpl::readEventInfo(uint8_t* buf, EventInfo& eventInfo)
{
	eventInfo.size = *reinterpret_cast<uint32_t*>(buf);
	buf += sizeof(uint32_t);
	SourceEvent::UntilId event(buffer(buf, sizeof(EventHeader)));
	eventInfo.id = event.eventId();
	eventInfo.timestamp = event.timestamp();
	eventInfo.type = event.type();
}

void MinosEventBuilder::RunImpl::computeBuiltEventInfo(uint32_t id, const EventInfoArray& srcEventInfo, EventInfo& builtEventInfo)
{
	builtEventInfo.size = sizeof(BuiltEventHeader) + sizeof(EventHeader) + sizeof(BuiltEventFooter);
	builtEventInfo.id = id;
	builtEventInfo.timestamp = 0;
	builtEventInfo.type = 0;
	for(size_t i = 0; i < array::size(srcEventInfo); ++i)
	{
		if(srcEventInfo[i].size != 0)
		{
			if(srcEventInfo[i].id != id)
			{
				LOG_WARN() << "MISSING: Event " << id << " for source " << i;
			}
			else
			{
				builtEventInfo.size += srcEventInfo[i].size - sizeof(EventHeader) - sizeof(EventEpilogue);
				if(builtEventInfo.timestamp == 0)
				{
					builtEventInfo.timestamp = srcEventInfo[i].timestamp;
					builtEventInfo.type      = srcEventInfo[i].type;
				}
				else
				{
					if(srcEventInfo[i].timestamp != builtEventInfo.timestamp)
					{
						throw std::runtime_error("mismatched timestamps !!");
					}
					if(srcEventInfo[i].type != builtEventInfo.type)
					{
						throw std::runtime_error("mismatched types !!");
					}
				}
			}
		}
	}
	LOG_DEBUG() << "Building Event: id=" << builtEventInfo.id
			<< " ts=" << builtEventInfo.timestamp
			<< " type=" << builtEventInfo.type
			<< " size=" << builtEventInfo.size;

	if(builtEventInfo.size > setup.activeSourcesCount * params.maxSrcEventSize_B)
	{
		throw std::runtime_error("Event is too big for the buffer, please adjust the event max size");
	}
}

void MinosEventBuilder::RunImpl::writeEventHeader(uint8_t*& writePtr, const EventInfo& builtEventInfo)
{
	BuiltEventHeader* header = reinterpret_cast<BuiltEventHeader*>(writePtr);
	header->prefix = header->PREFIX;
	header->size = builtEventInfo.size;
	writePtr += sizeof(BuiltEventHeader);

	EventHeader* evtHeader = reinterpret_cast<EventHeader*>(writePtr);
	evtHeader->prefix    = evtHeader->PREFIX;
	evtHeader->type      = builtEventInfo.type;
	evtHeader->timestamp = builtEventInfo.timestamp;
	evtHeader->eventId   = builtEventInfo.id;
	writePtr += sizeof(EventHeader);
}

void MinosEventBuilder::RunImpl::copySourceEvent(uint8_t*& writePtr, const EventInfo& srcEventInfo, mdcs::utl::ContiguousCircularBuffer& srcBuffer)
{
	uint8_t* readPtr = srcBuffer.readPtr() + sizeof(uint32_t) + sizeof(EventHeader);
	size_t wSize = srcEventInfo.size - sizeof(EventHeader) - sizeof(EventEpilogue);
	memcpy(writePtr, readPtr, wSize);
	writePtr += wSize;
	srcBuffer.moveReadPtr(srcEventInfo.size + sizeof(uint32_t));

	if(srcBuffer.empty())
	{
		++emptyBufferCount;
		LOG_DEBUG() << "EMPTY: " << emptyBufferCount << " buffers";
	}
}

void MinosEventBuilder::RunImpl::writeEventFooter(uint8_t*& writePtr)
{
	BuiltEventFooter* footer = reinterpret_cast<BuiltEventFooter*>(writePtr);
	footer->prefix = footer->PREFIX;
	writePtr += sizeof(BuiltEventFooter);
}

} /* namespace back */
} /* namespace acq */
} /* namespace minos */
