/**
 * @file FemAcquisitionBuffer.cpp
 * @date 15 mai 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the DaqMinos software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "FemAcquisitionBuffer.h"
#include <minos/fem/FeminosController.h>
#include <minos/fem/FrameTypes.h>
#include <mdcs/utl/net/UdpTransceiver.h>
#include <utl/io/InplaceParser.ipp>
#include <utl/Logging.h>
#include ICEUTIL_HEADER(Monitor.h)
#include ICEUTIL_HEADER(Mutex.h)

namespace minos
{
namespace acq
{
namespace front
{
using namespace fem;
using namespace mdcs::utl;
using namespace mdcs::utl::net;
using namespace utl::io;

// =============================================================================
// minos::acq::front::FemAcquisitionBuffer::FeminosReceiver class definition
// -----------------------------------------------------------------------------

class FemAcquisitionBuffer::FeminosReceiver : public FeminosController::AbstractReceiver
{
public:
	void prepareForReply();
	ReplyBufferPtr waitForReply(int timeout_ms);

	void setReply(std::auto_ptr<ReplyBuffer> rep);

private:
	std::auto_ptr<ReplyBuffer> reply;
	IceUtil::Monitor<IceUtil::Mutex> monitor;
};

// =============================================================================
// minos::acq::front::FemAcquisitionBuffer class methods body
// -----------------------------------------------------------------------------

FemAcquisitionBuffer::EventChunk::EventChunk(const void* begin)
	: segment(begin), size(), eventEnd()
{
}

void FemAcquisitionBuffer::EventChunk::setChunkEnd(const void* end)
{
	size = static_cast<const uint8_t*>(end) - static_cast<const uint8_t*>(segment);
}

FemAcquisitionBuffer::FemAcquisitionBuffer(
	const FemDataProducer::Params::FlowControl& params,
	fem::FeminosController& ctrl
)
	: feminos(ctrl),
	  flowManager(params, ctrl),
	  datablock(new uint8_t[params.circularBufferSize * 1024]),
	  frameBuffer(datablock.get(), params.circularBufferSize * 1024),
	  stopping(false), lastSendCount(0xFF),
	  lastEventSize(0), lastFrameSize(0), avgFrameSize(0),
	  lastTime(), sizeDataExtracted(0)
{
	lastTime.tv_sec = 0;
	lastTime.tv_nsec = 0;
	chunks.reserve(10000);

	// resets the frame sequence number on the board
	feminos.daq_clearCredits();
	feminos.daq_addCredits(0, FrameCredits, true);

	feminos.setReceiver(fem::FeminosController::ReceiverPtr(new FeminosReceiver));

	if(params.socketRecvBufferSize > 0)
	{
		feminos.transceiver().set_rcveBufSize(params.socketRecvBufferSize * 1024);
	}
	LOG_INFO() << "Receive socket buffer: " << feminos.transceiver().rcveBufSize();
}

FemAcquisitionBuffer::~FemAcquisitionBuffer()
{
	feminos.resetReceiver();
}

FemAcquisitionBuffer::Handle
FemAcquisitionBuffer::read(const uint32_t maxSize_B, const uint32_t maxWait_ms, bool /*allowSplitFrame*/)
{
	ValueRef<ChannelHitIndex> lastChannel;
	Milliseconds timeSpent = 0;
	size_t sizeChunks = 0; // the sum of the size of all chunks accumulated so far.
	size_t sizeEvents = 0; // the sum of the size of all completed events so far.
	size_t sizeCompletedFrames = 0; // the sum of the size of all frames entirely read so far.
	enum ReadState { WaitingNext = 0, Reconstructing, Dropping };
	ReadState state = WaitingNext;
	size_t lostCount = 0;
	Status status;
	chunks.clear();
	try
	{
		do
		{
			if(parser.isEndOfBuffer()) // no more data to read
			{
				sizeCompletedFrames += lastFrameSize;
				lastFrameSize = 0;
				if(state == WaitingNext and timeSpent >= maxWait_ms)
				{
					LOG_DEBUG() << "Reaching the expected event size takes too much time";
					break;
				}
				Milliseconds allowedWait =
					(stopping and state == WaitingNext and maxWait_ms > 250) ?
						250 : maxWait_ms;
				Milliseconds remainingWait = allowedWait;

				lostCount = receiveFrame(remainingWait);
				timeSpent += allowedWait - remainingWait;
			}
			if(state == Dropping)
			{
				LOG_DEBUG() << "dropping frame data !";
				const EventHeader* eventHeader = 0;
				ValueRef<FrameEnd> frameEnd;
				while(!(isDataField(parser, eventHeader) or parser.parse_1(frameEnd) or parser.isEndOfBuffer()))
				{
					parser.ignore<uint16_t>();
				}
				if(isDataField(parser, eventHeader))
				{
					LOG_DEBUG() << "found an event header, stop dropping data !";
					state = WaitingNext;
				}
				else
				{
					continue;
				}
			}

			chunks.push_back(EventChunk(parser.cur()));
			if(state == WaitingNext)
			{
				ValueRef<EventHeader> eventHeader;
				if(!parser.parse_1(eventHeader))
				{
					if(lostCount > 0)
					{
						LOG_WARN() << "Major data loss: start-of-event lost, dropping data until the next start-of-event !";
					}
					else
					{
						LOG_WARN() << "Invalid data: start-of-event expected but missing !";
					}
					state = Dropping;
					chunks.pop_back();
					continue;
				}
				else
				{
					LOG_DEBUG() << "boe, id=" << eventHeader->eventId;
					state = Reconstructing;
				}
			}

			ValueRef<EventEpilogue> eventEpilogue;
			ValueRef<FrameEnd> frameEnd;
			while(!(parser.parse_1(frameEnd) or parser.parse_1(eventEpilogue) or parser.isEndOfBuffer()))
			{
				ValueRef<ChannelHitIndex> channel;
				if(parser.parse_1(channel))
				{
					if(lostCount > 0 and lastChannel.isSet())
					{
						int lastChipChan = lastChannel->chip * 128 + lastChannel->channel;
						int curChipChan  = channel->chip * 128 + channel->channel;
						if(lastEventSize == 0)
						{
							lastEventSize = sizeChunks;
						}
						if(lastChipChan >= curChipChan or lostCount * avgFrameSize > lastEventSize)
						{
							if(lostCount * avgFrameSize > lastEventSize)
							{
								LOG_WARN() << "Lost more than one event: " << lostCount * avgFrameSize << " > " << lastEventSize;
							}
							LOG_WARN() << "Major data loss: end-of-event lost, closing event !";
							// reconstructing a valid event epilogue
							EventChunk& chunk = chunks.back();
							chunk.segment = &missingEpilogueBuffer;
							chunk.size = sizeof(missingEpilogueBuffer);
							chunk.eventEnd = true;

							sizeChunks += chunk.size;
							lastEventSize = sizeChunks - sizeEvents;
							sizeEvents = sizeChunks;
							state = Dropping;

							missingEpilogueBuffer.prefix = missingEpilogueBuffer.PREFIX;
							missingEpilogueBuffer.setEventSize(lastEventSize);

							break;
						}
						else
						{
							LOG_INFO() << "Minor data loss, continuing event.";
							lostCount = 0;
						}
					}
					lastChannel = channel;
				}
				else
				{
					parser.ignore<uint16_t>();
				}
			}
			if(state == Reconstructing)
			{
				EventChunk& chunk = chunks.back();
				chunk.setChunkEnd(parser.cur());
				if(frameEnd.isSet())
				{
					chunk.size -= sizeof(FrameEnd);
					chunk.eventEnd = false;
					sizeChunks += chunk.size;
				}
				else if(eventEpilogue.isSet())
				{
					chunk.eventEnd = true;
					sizeChunks += chunk.size;
					lastEventSize = sizeChunks - sizeEvents;
					sizeEvents = sizeChunks;
					state = WaitingNext;
					const_cast<EventEpilogue*>(eventEpilogue.get())->setEventSize(lastEventSize);
					LOG_DEBUG() << "eoe: size=" << chunk.size
						<< " eventSize=" << eventEpilogue->eventSize()
						<< " sizeEvents=" << sizeEvents;
					parser.parse_1(frameEnd); // try to parse the end-of-frame
				}
				else // end of buffer
				{
					LOG_WARN() << "no frame end marker!";
				}
			}
		}
		while(sizeEvents < maxSize_B);
		status = READ_OK;
	}
	catch(const Socket::TimeoutException& e)
	{
		LOG_DEBUG() << "timeout";
		if(stopping)
		{
			if(state == WaitingNext) status = DEPLETED;
			else                     status = INTERRUPTED;
		}
		else
		{
			status = TIMEOUT;
		}
	}
	LOG_DEBUG() << "Iterator " << status << ' ' << sizeCompletedFrames;
	return Handle(new(iterator)Iterator(*this, status, sizeCompletedFrames));
}

void FemAcquisitionBuffer::interruptRead()
{
	stopping = true;
}

size_t FemAcquisitionBuffer::receiveFrame(Milliseconds& recvTimeout)
{
	flowManager.sendCredits();

	ValueRef<DataPreamble>   preamble;
	ValueRef<DataFrameStart> header;
	do
	{
		BufferWSession session = frameBuffer.write(8192);
		uint8_t* frame = session.pointer();
		SocketAddress sender;
		const size_t len = feminos.transceiver().tryRcve(sender, frame, 8192, recvTimeout); //TODO: timeout will throw exception

		updateThroughputStats(len);

		InplaceParser p(buffer(frame, len));
		if(p.parse_1(preamble) and p.parse_1(header)) // check if it is a data frame
		{
			lastFrameSize = len;
			avgFrameSize = len; // TODO: compute a true average
			LOG_DEBUG() << "frame, size=" << header->payloadSize << " seq=" << preamble->recvCount << '/' << preamble->sendCount;
			if(header->payloadSize + sizeof(DataPreamble) != len)
			{
				LOG_WARN() << "mismatched frame size, payload=" << header->payloadSize << ", size=" << len;
			}
			session.commit(len);
			p.commit(parser);
		}
		else
		{
			handleNonDataFrame(buffer(frame, len));
			session.rollback(); // not keeping non-data frames in the circular buffer
		}
	}
	while(not header.isSet());

	size_t lostCount = (static_cast<size_t>(preamble->sendCount) - (lastSendCount + 1)) % 256;
	if(lostCount > 0)
	{
		LOG_INFO() << "Lost " << lostCount << " frames!";
	}
	flowManager.adjustCredits(lostCount);
	lastSendCount = preamble->sendCount;

	return lostCount;
}

void FemAcquisitionBuffer::updateThroughputStats(size_t len)
{
	sizeDataExtracted += len;
	// TODO: get rid of this function as soon as monitoring charts per source are implemented
	// for now, MacOS version is not necessary
#ifndef __APPLE__
	timespec t;
	clock_gettime(CLOCK_MONOTONIC, &t);
	if(lastTime.tv_sec == 0)
	{
		lastTime = t;
	}
	else if(t.tv_sec > lastTime.tv_sec)
	{
		uint64_t microSecs = static_cast<uint64_t>(t.tv_sec - lastTime.tv_sec) * 1000000 + (t.tv_nsec - lastTime.tv_nsec) / 1000;
		LOG_DEBUG() << "Time elapsed=" << microSecs << "us, transfer speed=" << static_cast<double>(sizeDataExtracted) * 1000000 / microSecs / (1024*1024) << " MB/s";
		lastTime = t;
		sizeDataExtracted = 0;
	}
#endif
}

void FemAcquisitionBuffer::handleNonDataFrame(const ::utl::io::ConstBuffer& buf)
{
	std::auto_ptr<ReplyBuffer> reply(new ReplyBuffer);
	reply->buffer.assign(buf.begin(), buf.end());
	reply->length = buf.maxBytes;
	static_cast<FeminosReceiver&>(feminos.receiver()).setReply(reply);
}

// =============================================================================
// minos::acq::front::FemAcquisitionBuffer::FlowManager class methods body
// -----------------------------------------------------------------------------

FemAcquisitionBuffer::FlowManager::FlowManager(
	const FemDataProducer::Params::FlowControl& params,
	fem::FeminosController& feminos
)
	: params(params), feminos(feminos),
	credits(0),
	creditIncrement(params.creditIncrement),
	creditLowThreshold(1),
	thresholdRaiseIncr(4),
	safeThreshold(1),
	unsafeThreshold(params.creditLowThreshold + 1),
	consecutiveFramesNeeded(creditLowThreshold * CONSECUTIVE_MULTIPLIER),
	missedFramesNeeded(MISSED_MAX)
{
}

void FemAcquisitionBuffer::FlowManager::sendCredits()
{
	if(credits < creditLowThreshold)
	{
		LOG_DEBUG() << "Asked " << creditIncrement << " credits (" << credits << " remaining)";
		feminos.daq_addCredits(creditIncrement, FrameCredits);

		credits += creditIncrement;
		if(creditLowThreshold < unsafeThreshold - 1)
		{
			if(consecutiveFramesNeeded <= 0)
			{
				safeThreshold = creditLowThreshold;
				creditLowThreshold = std::min(creditLowThreshold + thresholdRaiseIncr, params.creditLowThreshold);
				consecutiveFramesNeeded = creditLowThreshold * CONSECUTIVE_MULTIPLIER;
				LOG_INFO() << "Credits low threshold: " << creditLowThreshold << " (+)";
			}
		}
	}
}

void FemAcquisitionBuffer::FlowManager::adjustCredits(size_t lostCount)
{
	--credits;
	if(lostCount > 0)
	{
		credits -= lostCount;
		missedFramesNeeded -= lostCount;
		consecutiveFramesNeeded = creditLowThreshold * CONSECUTIVE_MULTIPLIER;
		if(creditLowThreshold > safeThreshold or missedFramesNeeded <= 0)
		{
			if(missedFramesNeeded <= 0 and safeThreshold > 1)
			{
				--safeThreshold;
			}
			unsafeThreshold = creditLowThreshold;
			creditLowThreshold = safeThreshold;
			missedFramesNeeded = MISSED_MAX;
			thresholdRaiseIncr = 1;
			LOG_INFO() << "Credits low threshold: " << creditLowThreshold << " (-)";
		}
	}
	else
	{
		if(consecutiveFramesNeeded > 0)
		{
			--consecutiveFramesNeeded;
		}
	}
}

// =============================================================================
// minos::acq::front::FemAcquisitionBuffer::Iterator class methods body
// -----------------------------------------------------------------------------

FemAcquisitionBuffer::Iterator::Iterator(FemAcquisitionBuffer& stream, Status status, size_t bytesRead)
	: stream_(stream),
	  status_(status),
	  bytesRead_(bytesRead),
	  currentChunk(stream_.chunks.begin())
{
}

FemAcquisitionBuffer::Iterator::~Iterator()
{
	try
	{
		if(bytesRead_ > 0)
		{
			stream_.frameBuffer.moveReadPtr(bytesRead_);
		}
	}
	catch(std::exception& e)
	{
		// TODO: fix the problem with invalid bytesRead call
		LOG_ERROR() << e.what();
	}
}

FemAcquisitionBuffer::Status FemAcquisitionBuffer::Iterator::getStatus() const
{
	return status_;
}

bool FemAcquisitionBuffer::Iterator::hasNext() const
{
	return currentChunk != stream_.chunks.end();
}

ConstBuffer FemAcquisitionBuffer::Iterator::next(bool& frameEnd)
{
	ConstBuffer buf(currentChunk->segment, currentChunk->size);
	frameEnd = currentChunk->eventEnd;
	++currentChunk;
	return buf;
}

void FemAcquisitionBuffer::Iterator::dispose()
{
	this->~Iterator();
}

// =============================================================================
// minos::acq::front::FemAcquisitionBuffer::FeminosReceiver class methods body
// -----------------------------------------------------------------------------

ReplyBufferPtr FemAcquisitionBuffer::FeminosReceiver::waitForReply(int timeout_ms)
{
	ReplyBufferPtr ret;
	IceUtil::Monitor<IceUtil::Mutex>::Lock lock(monitor);
	bool succeeded = true;
	while(reply.get() == 0 and succeeded)
	{
		succeeded = monitor.timedWait(IceUtil::Time::milliSeconds(timeout_ms));
	}
	ret = reply;
	return ret;
}

void FemAcquisitionBuffer::FeminosReceiver::setReply(std::auto_ptr<ReplyBuffer> rep)
{
	IceUtil::Monitor<IceUtil::Mutex>::Lock lock(monitor);
	reply = rep;
	monitor.notify();
}

void FemAcquisitionBuffer::FeminosReceiver::prepareForReply()
{
	IceUtil::Monitor<IceUtil::Mutex>::Lock lock(monitor);
	reply.reset();
}

} /* namespace front */
} /* namespace acq */
} /* namespace minos */
