/**
 * @file FemDataProducer.cpp
 * @date 16 mai 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the DaqMinos software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "FemDataProducer.h"
#include "FemAcquisitionBuffer.h"
#include <mdcs/acq/ServerApplication.h>
#include <mdcs/acq/mon/MetricsCollector.h>
#include <utl/compat/ArrayHelper.hpp>
#include <utl/Factory.ipp>
#include <utl/Singleton.ipp>
#include <utl/Logging.h>
#include <iostream>
#include <iomanip>

namespace minos
{
namespace acq
{
namespace front
{
using namespace fem;
namespace array = ::utl::array;

// =============================================================================
// minos::acq::front::FemDataProducer::Params class methods body
// -----------------------------------------------------------------------------

FemDataProducer::Params::Params()
{
	lowerCable.isUpperChips = false;
	upperCable.isUpperChips = true;
}

utl::net::SocketAddress FemDataProducer::Params::feminosSockAddr() const
{
	utl::net::SocketAddress addr(network, port);
	addr.ipAddress()[3] = feminosId + 1;
	return addr;
}

// =============================================================================
// minos::acq::front::FemDataProducer::Setup class methods body
// -----------------------------------------------------------------------------

FemDataProducer::Setup::Setup(ApplicationState& state, const Params& params)
	: mdcs::utl::BaseSetupData<Policy>(state, params),
	  feminos(params.feminosSockAddr()),
	  wiringMetric_  (state.application().collector().createMetric("Feminos.wiring")),
	  pedestalMetric_(state.application().collector().createMetric("Feminos.pedestals"))
{
	FemNum femId = feminos.getFeminosNumber();
	if(femId != static_cast<FemNum>(params.feminosId))
	{
		std::ostringstream msg;
		msg << "the actual Feminos number (" << femId << ") does not match the one specified in configuration (" << params.feminosId << ')';
		throw std::runtime_error(msg.str().c_str());
	}

	mapping_.setWiring(params.lowerCable);
	mapping_.setWiring(params.upperCable);

	bool doStartup = params.fec.reboot or isStartupNeeded();
	if(doStartup)
	{
		startupBoard();
	}
	bool doPedestals = doStartup or params.pedestalRun.isForced;
	if(doPedestals)
	{
		flushData();
		reportWiring();
		runPedestals();
	}
	if(params.doFlush or doPedestals)
	{
		flushData();
	}
	acquisitionGetReady();
}

bool FemDataProducer::Setup::isStartupNeeded()
{
	Reg1 reg1 = feminos.register_read<Reg1>();
	return reg1.fec_enable == 0;
}

void FemDataProducer::Setup::startupBoard()
{
	LOG_INFO() << "Executing Feminos board startup...";

	feminos.setAsicType(params.fec.asicType == AfterAsic ? "after" : "aget");

	Reg1 reg1 = feminos.register_read<Reg1>();
	bool powerCycleFec = (reg1.fec_enable != 0);
	reg1.power_inv  = 0;
	reg1.fec_enable = 0;
	feminos.register_write(reg1);

	if(powerCycleFec)
	{
		sleep(1);
	}

	reg1.fec_enable = 1;
	reg1.asic_mask  = params.fec.asicMask;
	feminos.register_write(reg1);

	if(params.fec.asicType == AfterAsic)
	{
		// reset reg1 because of a random bug on power on, causing an invalid value to be used.
		LOG_DEBUG() << "resetting After register 1 to 0";
		feminos.after_write(AllAsics, AfterCfgReg1::ID, 0);
	}

	LOG_INFO() << "Startup done.";
}

void FemDataProducer::Setup::runPedestals()
{
	LOG_INFO() << "Pedestal run init...";
	feminos.pedestal_set(AllAsics, AllChannels, 0);
	feminos.threshold_set(AllAsics, AllChannels, 0);
	feminos.hped_clear(AllAsics, AllChannels);

	const Params::FeminosConfig& pedestals = params.fem[Params::RunType_Pedestal];
	configureFeminos(pedestals);

	feminos.daq_setTarget(DataToPedestal);

	Reg0 reg0 = feminos.register_read<Reg0>();
	reg0.ped_subtract = 0;
	reg0.sca_enable = 1;
	feminos.register_write(reg0);

	LOG_INFO() << "First pedestal run for equalization started...";
	sleep(params.pedestalRun.duration);

	reg0.sca_enable = 0;
	feminos.register_write(reg0);

	reportPedestals();

	feminos.hped_center(AllAsics, AllChannels, params.pedestalRun.pedestal);
	// clear pedestals before next run
	feminos.hped_clear(AllAsics, AllChannels);

	reg0.ped_subtract = 1;
	reg0.sca_enable = 1;
	feminos.register_write(reg0);

	LOG_INFO() << "Second pedestal run for thresholds definition started...";
	sleep(params.pedestalRun.duration);

	reg0.sca_enable = 0;
	feminos.register_write(reg0);

	reportPedestals();

	LOG_INFO() << "Pedestals runs finished.";
}

void FemDataProducer::Setup::excludeUnmappedChannels()
{
	fem::Channel channel;
	channel.card = params.feminosId;
	for(channel.chip = 0; channel.chip < AsicCount; ++channel.chip)
	{
		int channelCount = params.fec.asicType == AfterAsic ? AfterChannelCount : AgetChannelCount;
		for(channel.channel = 0; channel.channel < channelCount; ++channel.channel)
		{
			fem::WiringInfo wiring = mapping_.getWiring(channel);
			bool isMapped = true;
			switch(wiring.device)
			{
			case ProjectiveTPC:
			{
				ProjectiveTpcPad pad;
				isMapped = mapping_.daqChannelToTpcPad(channel, pad, params.fec.asicType);
				break;
			}
			case ConstantTPC:
			{
				ConstantTpcPad pad;
				isMapped = mapping_.daqChannelToTpcPad(channel, pad, params.fec.asicType);
				break;
			}
			case Tracker:
			{
				CoaxPin pin;
				isMapped = mapping_.daqChannelToCoaxPin(channel, pin, params.fec.asicType);
				break;
			}
			case None:
				break;
			}
			if(!isMapped)
			{
				LOG_DEBUG() << "Excluding chip " << channel.chip << " channel " << channel.channel;
				feminos.threshold_set(channel.chip, channel.channel, 0xFFF);
			}
		}
	}
}

void FemDataProducer::Setup::flushData()
{
	LOG_INFO() << "Flushing Feminos data...";
	feminos.daq_clearCredits();

	Reg0 reg0 = feminos.register_read<Reg0>();
	reg0.sca_enable = 0;
	feminos.register_write(reg0);

	feminos.daq_setTarget(DataDiscard);
	sleep(params.flushDuration);

	feminos.buffer_setActive(false);
	feminos.buffer_sendPendingData();
	sleep(params.flushDuration);

	feminos.buffer_setActive(true);

	LOG_INFO() << "Flushing done.";
}

void FemDataProducer::Setup::acquisitionGetReady()
{
	LOG_INFO() << "Getting ready for acquisition";

	feminos.daq_clearCredits();

	const Params::FeminosConfig& femConfig = params.fem[Params::RunType_Data];
	configureFeminos(femConfig);
	feminos.hbusy_clear();
	feminos.daq_setTarget(DataToDaq);

	if(femConfig.enableZeroSuppression != 0)
	{
		excludeUnmappedChannels();
	}

	LOG_INFO() << "Ready for acquisition";
}

void FemDataProducer::Setup::configureFeminos(const Params::FeminosConfig& config)
{
	{
		Reg0 reg0 = feminos.register_read<Reg0>();
		reg0.sca_cnt          = config.scaReadCount;
		//reg0.mode_after
		reg0.rst_chan_cnt     = (config.agetResetCycles == 4);
		reg0.force_on_all     = config.agetForceOnAll;
		reg0.zero_suppress    = config.enableZeroSuppression;
		reg0.ped_subtract     = config.enableEqualization;
		reg0.emit_ch_hit_cnt  = config.emitHitCount;
		reg0.keep_rst_chan    = config.keepResetChannels;
		reg0.skip_rst_until   = config.skipResetChannels;
		reg0.modify_hit_reg   = config.modifyHitReg;
		reg0.sca_auto_start   = config.scaAutoStart;
		feminos.register_write(reg0);
	}
	{
		Reg1 reg1 = feminos.register_read<Reg1>();
		reg1.keep_fco = config.keepFCO;
		feminos.register_write(reg1);
	}
	{
		Reg5 reg5 = feminos.register_read<Reg5>();
		reg5.zs_pre           = config.samplesBefore;
		reg5.zs_post          = config.samplesAfter;
		reg5.test_enable      = config.enableTest;
		reg5.test_mode        = config.memArrayMode;
		reg5.test_zbt         = !config.useZBT;
		reg5.emit_lst_cell_rd = config.emitLastCellRead;
		reg5.eof_on_eoe       = config.closeFrameOnEventEnd;
		reg5.emit_empty_ch    = config.emitEmptyChannels;
		reg5.sca_wckdiv       = config.scaClockDivisor;
		reg5.polarity         = config.polarityBelow;
		feminos.register_write(reg5);
	}
	{
		Reg2 reg2(0);
		reg2.trig_delay       = config.triggerDelay / 10;
		reg2.trig_rate_range  = config.triggerRate.range();
		reg2.trig_rate        = config.triggerRate.rate();
		reg2.event_limit      = config.eventLimit;
		reg2.enable_auto_trig = config.enableAutoTrig;
		reg2.enable_ext_trig  = config.enableExtTrig;
		reg2.enable_mult_trig = config.enablePulserTrig;
		reg2.enable_tcm_trig  = config.enableTcmTrig;
		feminos.register_write(reg2);
	}
	for(AsicNum asic = 0; asic < AsicCount; ++asic)
	{
		configureAfter(config.after[asic], asic);
		configureAget(config.aget[asic], asic);
	}
	if(config.patternLength > 0)
	{
		feminos.testdata_fillPattern(config.patternType, config.patternLength);
	}
	feminos.buffer_enableTimeout(config.enableBufferTimeout);
	feminos.buffer_setTimeout(config.bufferTimeout);
	feminos.daq_setLossPolicy(config.lossPolicy);
	feminos.daq_setCreditWaitTime(config.creditWaitTime);
	feminos.hped_setThreshold(AllAsics, AllChannels, config.threshold, config.sigma);
}

void FemDataProducer::Setup::configureAfter(const Params::After& afterCfg, fem::AsicNum asic)
{
	static const uint16_t gains[] = {120, 240, 360, 600};
	static const uint16_t times[] = {116, 200, 412, 505, 610, 695, 912, 993, 1054, 1134, 1343, 1421, 1546, 1626, 1834, 1912};

	if(afterCfg.isPresent)
	{
		const size_t gain = array::find_index(array::begin(gains), array::end(gains), afterCfg.gain);
		const size_t time = array::find_index(array::begin(times), array::end(times), afterCfg.shapingTime);

		if(gain >= array::size(gains))
		{
			throw std::runtime_error("invalid gain");
		}
		if(time >= array::size(times))
		{
			throw std::runtime_error("invalid shapingTime");
		}

		AfterCfgReg1 after1 = feminos.after_read<AfterCfgReg1>(asic);
		after1.gain = gain;
		after1.time = time;
		feminos.after_write(asic, after1);

		AfterCfgReg2 after2 = feminos.after_read<AfterCfgReg2>(asic);
		after2.reset_marker_enabled = afterCfg.enableMarkerReset;
		after2.reset_marker_level   = afterCfg.resetLevel;
		after2.read_from_0          = afterCfg.readFromZero;
		after2.test_digout          = afterCfg.testDigOut;
		feminos.after_write(asic, after2);
	}
}

void FemDataProducer::Setup::configureAget(const Params::AGet& agetCfg, fem::AsicNum asic)
{
	if(agetCfg.isPresent)
	{
		feminos.aget_write(asic)
			.setEnableMarkerReset(agetCfg.enableMarkerReset)
			.setResetLevel(agetCfg.resetLevel)
			.setReadFromZero(agetCfg.readFromZero)
			.setTestDigOut(agetCfg.testDigOut)
			.setModeReadAll(agetCfg.modeReadAll)
			.setPolarityPositive(agetCfg.polarityPositive)
			.setReadFpn(agetCfg.readFpn)
			.setIcmVoltage(agetCfg.icmVoltage)
			.setDacThreshold(agetCfg.dacThreshold)
			.setTriggerVeto(agetCfg.triggerVeto)
			.setSynchroDiscri(agetCfg.synchroDiscri)
			.setTimeOverThreshold(agetCfg.timeOverThreshold)
			.setRangeTriggerWidth(agetCfg.rangeTriggerWidth)
			.setTriggerWidth(agetCfg.triggerWidth)
			.setShortRead(agetCfg.shortRead)
			.setDisableMultiplicityOut(agetCfg.disableMultiplicityOut)
			.setAutoresetBank(agetCfg.autoresetBank);

		for(ChannelNum i = 0; i < AgetConfigurableChannelCount; ++i)
		{
			const Params::AGet::Channel& chan = agetCfg.channel[i];
			feminos.aget_write(asic)
				.setGain(i + 1, chan.gain)
				.setShapingTime(i + 1, chan.shapingTime)
				.setThreshold(i + 1, chan.threshold)
				.setInhibit(i + 1, chan.inhibit);
		}
	}
}

void FemDataProducer::Setup::startDataFlow()
{
	Reg0 reg0 = feminos.register_read<Reg0>();
	reg0.sca_enable = 1;
	feminos.register_write(reg0);
}

void FemDataProducer::Setup::stopDataFlow()
{
	Reg0 reg0 = feminos.register_read<Reg0>();
	reg0.sca_enable = 0;
	feminos.register_write(reg0);
}

void FemDataProducer::Setup::reportWiring()
{
	mon::Metric::SampleTimestamp ts = wiringMetric_->currentTimestamp();
	wiringMetric_->reportSample(int64_t(params.lowerCable.value), ts);
	wiringMetric_->reportSample(int64_t(params.upperCable.value), ts);
}

void FemDataProducer::Setup::reportPedestals()
{
	PedestalSummaryResult summary;
	std::cout << std::fixed << std::setprecision(2);
	for(AsicNum asic = 0; asic < AsicCount; ++asic)
	{
		feminos.hped_getSummary(summary, asic, AllChannels);
		pedestalMetric_->reportSample(summary.buffer(), mon::Metric::currentTimestamp());

		for(ChannelNum channel = 0; channel < summary.pedestalSummary().count(); ++channel)
		{
			const PedestalSummary& p = summary.pedestalSummary()[channel];
			LOG_INFO()
				<< "Chip "    << std::setw(2) << p.channel.chip
				<< " Chan "   << std::setw(2) << p.channel.channel
				<< " Mean "   << p.pedestalMean()
				<< " Stdev "  << p.pedestalStdev();
		}
	}
	std::cout.unsetf(std::ios::floatfield);

//	for(int channel = 2; channel < 5; ++channel)
//	{
//		PedestalHistoResult histo;
//		feminos.hped_getHisto(histo, 0, channel);
//
//		std::cout << "Bin | Value" << std::endl;
//		for(unsigned bin = 0; bin < histo.pedestalHisto().count(); ++bin)
//		{
//			std::cout
//				<< std::setw(3) << histo.pedestalHisto()[bin].binIndex
//				<< std::setw(8) << histo.pedestalHisto()[bin].binValue
//				<< std::endl;
//		}
//	}
}

// =============================================================================
// minos::acq::front::FemDataProducer::RunImpl class methods body
// -----------------------------------------------------------------------------

FemDataProducer::RunImpl::RunImpl(ApplicationState& state, Setup& setup, const RunInfo& runInfo)
	: mdcs::utl::BaseRunImpl<Policy>(state, setup, runInfo)
{
	setup.feminos.cmdStats_clear();
	setup.feminos.tcmStats_clear();
}

InputBufferPtr FemDataProducer::RunImpl::createBuffer()
{
	return InputBufferPtr(new FemAcquisitionBuffer(params.flowControl, setup.feminos)); //TODO: do a better encapsulation
}

void FemDataProducer::RunImpl::start()
{
	setup.startDataFlow();
}

void FemDataProducer::RunImpl::stop()
{
	setup.stopDataFlow();
}

} /* namespace front */
} /* namespace acq */
} /* namespace minos */
