/**
 * @file FemDataProducer.h
 * @date 16 mai 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the DaqMinos software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#ifndef minos_acq_front_FemDataProducer_h_INCLUDED
#define minos_acq_front_FemDataProducer_h_INCLUDED

#include <minos/fem/DetectorMapping.h>
#include <minos/fem/FeminosController.h>
#include <minos/fem/TriggerRate.h>
#include <mdcs/acq/mon/Metric.h>
#include <mdcs/acq/front/AbstractDataProducer.h>
#include <mdcs/utl/Policy.hpp>

namespace minos
{
namespace acq
{
using namespace mdcs::acq;

namespace front
{
using namespace mdcs::acq::front;

BEGIN_POLICY_NAMED(DataProducer, FemDataProducer, "FeminosDataProducer")

	class Params : public mdcs::utl::BaseParamData<Policy>
	{
	public:
		Params();

		utl::net::SocketAddress feminosSockAddr() const;

		template<class C> static void init(Params& r, const C& config);
		template<class C> static void init(fem::WiringInfo& r, const C& config);

		struct FEC
		{
			template<class C> static void init(FEC& p, const C& config);
			bool reboot;
			fem::AsicType asicType;
			short asicMask;
		};
		struct FlowControl
		{
			template<class C> static void init(FlowControl& r, const C& config);
			int creditLowThreshold;
			int creditIncrement;
			int socketRecvBufferSize;
			int circularBufferSize;
		};
		struct PedestalRun
		{
			template<class C> static void init(PedestalRun& r, const C& config);
			bool isForced;
			short pedestal;
			short duration;
		};
		struct After
		{
			template<class C> static void init(After& p, const C& config);
			bool isPresent;
			bool enableMarkerReset;
			bool resetLevel;
			bool readFromZero;
			bool testDigOut;
			short gain;
			short shapingTime;
		};
		struct AGet
		{
			struct Channel
			{
				template<class C> static void init(Channel& p, const C& config);
				short gain;
				short shapingTime;
				short threshold;
				bool inhibit;
			};
			template<class C> static void init(AGet& p, const C& config);
			bool isPresent;
			bool enableMarkerReset;
			bool resetLevel;
			bool readFromZero;
			bool testDigOut;
			bool modeReadAll;
			bool polarityPositive;
			bool readFpn;
			short icmVoltage;
			short dacThreshold;
			short triggerVeto;
			bool synchroDiscri;
			bool timeOverThreshold;
			bool rangeTriggerWidth;
			short triggerWidth;
			bool shortRead;
			bool disableMultiplicityOut;
			bool autoresetBank;
			Channel channel[fem::AgetConfigurableChannelCount];
		};
		struct FeminosConfig
		{
			template<class C> static void init(FeminosConfig& p, const C& config);

			///@{ Buffer
			bool enableBufferTimeout;
			fem::BufferTimeout bufferTimeout;
			short creditWaitTime;
			fem::DaqLossPolicy lossPolicy;
			///@}
			///@{ Framing
			bool  closeFrameOnEventEnd;
			bool  emitEmptyChannels;
			bool  emitLastCellRead;
			bool  emitHitCount;
			bool  keepResetChannels;
			short skipResetChannels;
			///@}
			///@{ Readout
			short scaReadCount;
			short scaClockDivisor;
			bool  modifyHitReg;
			bool  agetForceOnAll;
			short agetResetCycles;
			///@}
			///@{
			bool enableTest;
			bool memArrayMode;
			bool useZBT;
			bool keepFCO;
			fem::TestPattern patternType;
			uint16_t patternLength;
			///@}
			///@{ Trigger
			bool enableAutoTrig;
			bool enableExtTrig;
			bool enablePulserTrig;
			bool enableTcmTrig;
			fem::TriggerRate triggerRate;
			int  triggerDelay; ///< trigger delay in nanoseconds
			int  eventLimit;
			bool scaAutoStart;
			///@}
			//@{ Zero Suppression
			bool  enableZeroSuppression;
			bool  enableEqualization;
			short threshold;
			float sigma;
			short samplesBefore;
			short samplesAfter;
			bool polarityBelow;
			///@}
			After after[fem::AsicCount];
			AGet aget[fem::AsicCount];
		};

		enum RunType { RunType_Pedestal = 0, RunType_Data, RunTypeCount };

		::utl::net::IpAddress network;
		int port;
		int feminosId;
		bool doFlush;
		short flushDuration;
		FEC fec;
		PedestalRun pedestalRun;
		fem::WiringInfo lowerCable;
		fem::WiringInfo upperCable;
		FlowControl flowControl;
		FeminosConfig fem[RunTypeCount];
	};
	class Setup : public mdcs::utl::BaseSetupData<Policy>
	{
	public:
		friend class FemDataProducer::RunImpl;

		Setup(ApplicationState& state, const Params& params);

	private:
		bool isStartupNeeded();
		void startupBoard();
		void runPedestals();
		void flushData();
		void acquisitionGetReady();
		void startDataFlow();
		void stopDataFlow();
		void reportWiring();
		void reportPedestals();
		void excludeUnmappedChannels();
		void configureFeminos(const Params::FeminosConfig& config);
		void configureAfter(const Params::After& afterCfg, fem::AsicNum asic);
		void configureAget(const Params::AGet& agetCfg, fem::AsicNum asic);

		fem::FeminosController feminos;
		fem::DetectorMapping mapping_;
		mon::MetricPtr wiringMetric_;
		mon::MetricPtr pedestalMetric_;
	};
	class RunImpl : public mdcs::utl::BaseRunImpl<Policy>
	{
	public:
		RunImpl(ApplicationState& state, Setup& setup, const RunInfo& runInfo);
		InputBufferPtr createBuffer();
		void start();
		void stop();
	};

END_POLICY(FemDataProducer)

} /* namespace front */
} /* namespace acq */
} /* namespace minos */

#endif /* minos_acq_front_FemDataProducer_h_INCLUDED */
