/**
 * @file MinosAcqFrontCCfg.cpp
 * @date 17 mai 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the DaqMinos software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "MinosAcqFrontCCfg.h"
#include "FemDataProducer.h"
#include "MinosTcmDataController.h"
#include <mdcs/cfg/IdentityProvider.h>
#include <mdcs/cfg/CCfgConfig.h>
#include <CCfg/CConfig.h>
#include <utl/compat/ArrayHelper.hpp>
#include <cmath>

namespace minos
{
using namespace mdcs;
using namespace fem;

namespace acq
{
namespace front
{
using namespace ::utl;

template<>
void MinosTcmDataController::Params::SyncCommands::init(SyncCommands& r, const CCfg::CConfig& subconfig)
{
	CCfg::CConfig config = subconfig;
	r.restart         = config("restart");
	r.clearTimestamp  = config("clearTimestamp");
	r.clearEventCount = config("clearEventCount");
	r.scaStart        = config("scaStart");
	r.wckSynch        = config("wckSynch");
}

template<>
void MinosTcmDataController::Params::init(Params& p, const CCfg::CConfig& subconfig)
{
	CCfg::CConfig config = subconfig;
	p.addr.fromString(config("sockAddr"));
	p.sel_edge       = config("sel_edge");
	p.do_end_of_busy = config("do_end_of_busy");
	p.event_max_time = config("event_max_time");
	for(int i = TcmTriggerAuto; i < TcmTriggerCount; ++i)
	{
		CCfg::CConfig triggerCfg = config("Trigger", Params::triggerId2Name(static_cast<TcmTrigger>(i)));
		p.triggerEnabled[i] = triggerCfg("enabled");
		p.triggerDelay[i]   = triggerCfg("delay");

		if(i == TcmTriggerAuto)
		{
			p.autoTriggerRate = fem::TriggerRate::createFromFreq(triggerCfg("frequency"));
			p.event_limit     = triggerCfg("eventLimit");
		}
	}
	Params::SyncCommands::init(p.commands[Configure], config("Commands", "configure"));
	Params::SyncCommands::init(p.commands[Start],     config("Commands", "start"));
}

template<>
void FemDataProducer::Params::init(fem::WiringInfo& p, const CCfg::CConfig& subconfig)
{
	CCfg::CConfig config = subconfig;

	// TODO: factorize these enums mappers into some template method/class
	static const char* const devices[] = {"ProjectiveTPC", "ConstantTPC", "Tracker", "None"};
	std::string deviceStr = config("device");
	size_t device = array::find_index(array::begin(devices), array::end(devices), deviceStr);
	if(device >= array::size(devices))
	{
		throw std::runtime_error("invalid device");
	}
	p.device       = device;
	p.isDefault    = config("defaultWiring").getBoolValue();
	p.tpcQuadrant  = config("tpcQuadrant").getIntValue();
	p.tpcConnector = config("tpcConnector").getIntValue();
}

template<>
void FemDataProducer::Params::FEC::init(FEC& p, const CCfg::CConfig& subconfig)
{
	static const char* const asicTypes[] = {"after", "aget"};

	CCfg::CConfig config = subconfig;
	p.reboot   = config("reboot");
	p.asicType = static_cast<AsicType>(array::find_index(
			array::begin(asicTypes),
			array::end(asicTypes),
			config("type").getStringValue()));
	p.asicMask = config("mask");
}

template<>
void FemDataProducer::Params::PedestalRun::init(PedestalRun& p, const CCfg::CConfig& subconfig)
{
	CCfg::CConfig config = subconfig;
	p.isForced    = config("forced");
	p.pedestal    = config("pedestal");
	p.duration = config("runDuration");
}

template<>
void FemDataProducer::Params::FlowControl::init(FlowControl& p, const CCfg::CConfig& subconfig)
{
	CCfg::CConfig config = subconfig;
	p.creditLowThreshold   = config("creditLowThreshold");
	p.creditIncrement      = config("creditIncrement");
	p.socketRecvBufferSize = config("socketRecvBufferSize");
	p.circularBufferSize   = config("circularBufferSize");
}

template<>
void FemDataProducer::Params::After::init(After& p, const CCfg::CConfig& subconfig)
{
	CCfg::CConfig config = subconfig;
	p.isPresent    = config("isPresent");
	p.gain         = config("gain");
	p.shapingTime  = config("shapingTime");
	p.enableMarkerReset = config("enableMarkerReset");
	p.resetLevel   = config("resetLevel");
	p.readFromZero = config("readFromZero");
	p.testDigOut   = config("testDigOut");
}

template<>
void FemDataProducer::Params::AGet::Channel::init(Channel& p, const CCfg::CConfig& subconfig)
{
	static const char* const shapingTimes[] = {"69.67", "117.3", "232.3", "279.7", "333.9", "382.7",
			"501.8", "541.3", "568.2", "632.5", "720.9", "760.3", "830.7", "869.2", "976.5", "1014"};
	static const short gains[] = {120, 240, 1000, 10000};

	CCfg::CConfig config = subconfig;
	p.shapingTime = array::find_index(array::begin(shapingTimes), array::end(shapingTimes),	config("shapingTime").getValueAsString());
	p.gain        = array::find_index(array::begin(gains), array::end(gains), config("gain").getIntValue());
	p.threshold   = config("threshold");
	p.inhibit     = config("inhibit");
}

template<>
void FemDataProducer::Params::AGet::init(AGet& p, const CCfg::CConfig& subconfig)
{
	static const char* const icmVoltages[] = {"1.25", "1.35", "1.55", "1.65"};
	static const char* const triggerVetos[] = {"none", "4us", "hit register"};
	static const short triggerWidth[] = {94, 77, 110, 127, 218, 170, 235, 298};

	CCfg::CConfig config = subconfig;
	p.isPresent         = config("isPresent");
	p.enableMarkerReset = config("enableMarkerReset");
	p.resetLevel        = config("resetLevel");
	p.readFromZero      = config("readFromZero");
	p.testDigOut        = config("testDigOut");
	p.modeReadAll       = config("modeReadAll");
	p.polarityPositive  = config("polarityPositive");
	p.readFpn           = config("readFpn");
	p.icmVoltage        = array::find_index(array::begin(icmVoltages), array::end(icmVoltages), config("icmVoltage").getValueAsString());
	p.dacThreshold      = config("dacThreshold");
	p.triggerVeto       = array::find_index(array::begin(triggerVetos), array::end(triggerVetos), config("triggerVeto").getStringValue());
	p.synchroDiscri     = config("synchroDiscri");
	p.timeOverThreshold = config("timeOverThreshold");
	p.triggerWidth      = array::find_index(array::begin(triggerWidth), array::end(triggerWidth), config("triggerWidth").getIntValue());
	p.rangeTriggerWidth = p.triggerWidth / 4;
	p.triggerWidth     %= 4;
	p.shortRead         = config("shortRead");
	p.disableMultiplicityOut = config("disableMultiplicityOut");
	p.autoresetBank          = config("autoresetBank");

	for(unsigned i = 0; i < AgetConfigurableChannelCount; ++i)
	{
		Channel::init(p.channel[i], config("Channel", i));
	}
}

template<>
void FemDataProducer::Params::FeminosConfig::init(FeminosConfig& p, const CCfg::CConfig& subconfig)
{
	CCfg::CConfig config = subconfig;
	{
		CCfg::CConfig bufferCfg = config("Buffer");
		p.enableBufferTimeout = bufferCfg("enableTimeout");
		p.bufferTimeout =
			static_cast<fem::BufferTimeout>(
				static_cast<int>(log10(bufferCfg("timeout").getIntValue()))
			);
		p.creditWaitTime = bufferCfg("creditWaitTime");
		std::string lossPolicy = bufferCfg("lossPolicy");

		static const char* const lossPolicies[] = {"Ignore", "Recredit", "Resend"};
		size_t policy = array::find_index(array::begin(lossPolicies), array::end(lossPolicies), lossPolicy);
		if(policy >= array::size(lossPolicies))
		{
			throw std::runtime_error("invalid loss policy");
		}
		p.lossPolicy = static_cast<fem::DaqLossPolicy>(policy);
	}
	{
		CCfg::CConfig framingCfg = config("Framing");
		p.closeFrameOnEventEnd = framingCfg("closeFrameOnEventEnd");
		p.emitEmptyChannels    = framingCfg("emitEmptyChannels");
		p.emitLastCellRead     = framingCfg("emitLastCellRead");
		p.emitHitCount         = framingCfg("emitHitCount");
		p.keepResetChannels    = framingCfg("keepResetChannels");
		p.skipResetChannels    = framingCfg("skipResetChannels");
	}
	{
		CCfg::CConfig readoutCfg = config("Readout");
		p.scaReadCount    = readoutCfg("scaReadCount");
		p.scaClockDivisor = readoutCfg("scaClockDivisor");
		p.modifyHitReg    = readoutCfg("modifyHitReg");
		p.agetForceOnAll  = readoutCfg("agetForceOnAll");
		p.agetResetCycles = readoutCfg("agetResetCycles");
	}
	{
		std::string patternType;
		CCfg::CConfig testCfg = config("Test");
		p.enableTest    = testCfg("enabled");
		p.memArrayMode  = testCfg("memArrayMode");
		p.useZBT        = testCfg("useZBT");
		p.keepFCO       = testCfg("keepFCO");
		p.patternLength = testCfg("patternLength");
		patternType     = testCfg("patternType").getStringValue();

		static const char* const patterns[] = {"raising ramp", "falling ramp", "pulse"};
		size_t index = array::find_index(array::begin(patterns), array::begin(patterns), patternType);
		if(index >= array::size(patterns))
		{
			throw std::runtime_error("invalid test pattern");
		}
		p.patternType = static_cast<fem::TestPattern>(index);
	}
	{
		CCfg::CConfig triggerCfg = config("Trigger");
		p.enableAutoTrig   = triggerCfg("enableAuto");
		p.enableExtTrig    = triggerCfg("enableExternal");
		p.enablePulserTrig = triggerCfg("enablePulser");
		p.enableTcmTrig    = triggerCfg("enableTcm");
		p.triggerRate  = fem::TriggerRate::createFromFreq(triggerCfg("frequency"));
		p.triggerDelay = triggerCfg("delay");
		p.eventLimit   = triggerCfg("eventLimit");
		p.scaAutoStart = triggerCfg("scaAutoStart");
	}
	{
		CCfg::CConfig zsCfg = config("ZeroSuppression");
		p.enableZeroSuppression = zsCfg("enabled");
		p.enableEqualization    = zsCfg("enableEqualization");
		p.threshold             = zsCfg("threshold");
		p.sigma                 = zsCfg("sigma").getRealValue();
		p.samplesBefore         = zsCfg("samplesBefore");
		p.samplesAfter          = zsCfg("samplesAfter");
		p.polarityBelow         = zsCfg("polarityBelow");
	}
	for(int i = 0; i < fem::AsicCount; ++i)
	{
		After::init(p.after[i], config("After", i));
		AGet::init(p.aget[i], config("AGet", i));
	}
}

template<>
void FemDataProducer::Params::init(Params& p, const CCfg::CConfig& subconfig)
{
	CCfg::CConfig config = subconfig;
	p.network.fromString(config("network"));
	p.port             = config("port");
	p.feminosId        = cfg::IdentityProvider::instance().id().num();
	p.doFlush          = config("doFlush");
	p.flushDuration    = config("flushDuration");

	FEC::init(p.fec, config("FEC"));
	p.lowerCable.board = p.feminosId;
	p.upperCable.board = p.feminosId;
	init(p.lowerCable, config("Wiring")("Cable", "chips01"));
	init(p.upperCable, config("Wiring")("Cable", "chips23"));
	PedestalRun::init(p.pedestalRun, config("PedestalRun"));
	FlowControl::init(p.flowControl, config("FlowControl"));
	FeminosConfig::init(p.fem[RunType_Pedestal], config("FeminosConfig", "pedestals"));
	FeminosConfig::init(p.fem[RunType_Data], config("FeminosConfig", "run"));
}

MinosAcqFrontCCfg::MinosAcqFrontCCfg()
{
	cfg::Configurator::instance()
		.add(&MinosTcmDataController::Params::init<CCfg::CConfig>)
		.add<FemDataProducer::Params>(&FemDataProducer::Params::init<CCfg::CConfig>)
		;
}

} /* namespace front */
} /* namespace acq */
} /* namespace minos */
