/**
 * @file MinosTcmDataController.cpp
 * @date 7 août 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the DaqMinos software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "MinosTcmDataController.h"

namespace minos
{
namespace acq
{
namespace front
{
using namespace fem;

void MinosTcmDataController::Params::SyncCommands::send(fem::TcmController& tcm) const
{
	TcmSwTrigReg tReg = tcm.register_read<TcmSwTrigReg>();
	if(restart)
	{
		tReg.restart = 1;
		tcm.register_write(tReg);
		tReg.restart = 0;
		tcm.register_write(tReg);
	}

	tReg.clr_tstamp = clearTimestamp;
	tReg.clr_evcnt  = clearEventCount;
	tReg.sca_start  = scaStart;
	tReg.wck_synch  = wckSynch;
	tcm.register_write(tReg);

	tReg.clr_tstamp = 0;
	tReg.clr_evcnt  = 0;
	tReg.sca_start  = 0;
	tReg.wck_synch  = 0;
	tcm.register_write(tReg);
}

// =============================================================================
// minos::acq::front::MinosTcmDataController::Params class methods body
// -----------------------------------------------------------------------------

const char* MinosTcmDataController::Params::triggerId2Name(fem::TcmTrigger id)
{
	static const char* const names[] = {"Auto", "NIM", "LVDS", "TTL"};
	return names[id];
}

// =============================================================================
// minos::acq::front::MinosTcmDataController::Setup class methods body
// -----------------------------------------------------------------------------

MinosTcmDataController::Setup::Setup(mdcs::acq::ApplicationState& state, const Params& params)
	: mdcs::utl::BaseSetupData<Policy>(state, params),
	  tcm(params.addr)
{
	TcmConfigReg cfgReg = tcm.register_read<TcmConfigReg>();
	cfgReg.trig_rate        = params.autoTriggerRate.rate();
	cfgReg.trig_rate_range  = params.autoTriggerRate.range();
	cfgReg.event_cnt_limit  = params.event_limit;
	cfgReg.sel_edge         = params.sel_edge;
	cfgReg.auto_trig_enable = 0;
	cfgReg.nim_trig_enable  = 0;
	cfgReg.lvds_trig_enable = 0;
	cfgReg.ttl_trig_enable  = 0;
	cfgReg.do_end_of_busy   = params.do_end_of_busy;
	cfgReg.event_max_time   = params.event_max_time;
	tcm.register_write(cfgReg);

	for(int i = TcmTriggerAuto; i < TcmTriggerCount; ++i)
	{
		tcm.triggerDelay(static_cast<TcmTrigger>(i), params.triggerDelay[i] / 10);
	}

	params.commands[Params::Configure].send(tcm);
}

// =============================================================================
// minos::acq::front::MinosTcmDataController::RunImpl class methods body
// -----------------------------------------------------------------------------

MinosTcmDataController::RunImpl::RunImpl(mdcs::acq::ApplicationState& state, Setup& setup, const mdcs::acq::RunInfo& runInfo)
	: mdcs::utl::BaseRunImpl<Policy>(state, setup, runInfo)
{
	setup.tcm.cmdStats_clear();
	setup.tcm.msgStats_clear();

	params.commands[Params::Start].send(setup.tcm);

	TcmConfigReg cfgReg = setup.tcm.register_read<TcmConfigReg>();
	cfgReg.auto_trig_enable = params.triggerEnabled[TcmTriggerAuto];
	cfgReg.nim_trig_enable  = params.triggerEnabled[TcmTriggerNIM];
	cfgReg.lvds_trig_enable = params.triggerEnabled[TcmTriggerLVDS];
	cfgReg.ttl_trig_enable  = params.triggerEnabled[TcmTriggerTTL];
	setup.tcm.register_write(cfgReg);
}

MinosTcmDataController::RunImpl::~RunImpl()
{
	TcmConfigReg cfgReg = setup.tcm.register_read<TcmConfigReg>();
	cfgReg.auto_trig_enable = 0;
	cfgReg.nim_trig_enable  = 0;
	cfgReg.lvds_trig_enable = 0;
	cfgReg.ttl_trig_enable  = 0;
	setup.tcm.register_write(cfgReg);

	TcmSwTrigReg tReg = setup.tcm.register_read<TcmSwTrigReg>();
	tReg.restart = 1;
	setup.tcm.register_write(tReg);
	tReg.restart = 0;
	setup.tcm.register_write(tReg);
}

} /* namespace front */
} /* namespace acq */
} /* namespace minos */
