/* =============================================================================
 * BoardController.cpp
 * -----------------------------------------------------------------------------
 * class minos::fem::BoardController
 * Created on: 8 août 2013 at Irfu/Sedi/Lilas, CEA Saclay, F-91191, France
 * Author: fchateau
 * -----------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * -----------------------------------------------------------------------------
 * Contributors: fchateau
 * -----------------------------------------------------------------------------
 * This software is part of the DaqMinos project.
 * -----------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding
 * by the rules of distribution of free software. You can use, modify and/or
 * redistribute the software under the terms of the CeCILL license as circulated
 * by CEA, CNRS and INRIA at the following URL: "http://www.cecill.info".
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability. In this
 * respect, the user's attention is drawn to the risks associated with loading,
 * using, modifying and/or developing or reproducing the software by the user in
 * light of its specific status of free software, that may mean that it is
 * complicated to manipulate, and that also therefore means that it is reserved
 * for developers and experienced professionals having in-depth computer
 * knowledge. Users are therefore encouraged to load and test the software's
 * suitability as regards their requirements in conditions enabling the security
 * of their systems and/or data to be ensured and, more generally, to use and
 * operate it in the same conditions as regards security.
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * -----------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for
 * commercial purposes. For this you will need to negotiate a specific contract
 * with a legal representative of CEA.
 * =============================================================================
 */

#include "BoardController.h"
#include <utl/Logging.h>
#include <iomanip>
#include <cstdio>

namespace minos
{
namespace fem
{
using namespace mdcs::utl::net;

BoardController::DefaultReceiver::DefaultReceiver(mdcs::utl::net::UdpTransceiver& transceiver)
	: transceiver_(transceiver)
{
}

void BoardController::DefaultReceiver::prepareForReply()
{
}

ReplyBufferPtr BoardController::DefaultReceiver::waitForReply(int timeout_ms)
{
	ReplyBufferPtr reply = ReplyBufferPtr(new ReplyBuffer);
	reply->buffer.resize(8192);
	SocketAddress srcEndpoint;
	Milliseconds timeout = timeout_ms;
	try
	{
		reply->length = transceiver_.tryRcve(srcEndpoint, buffer(reply->buffer), timeout);
	}
	catch(Socket::TimeoutException& e) //TODO: handle timeout without exception ?
	{
		reply.reset();
	}
	return reply;
}

// =============================================================================
// minos::fem::BoardController class methods body
// -----------------------------------------------------------------------------

BoardController::BoardController(const utl::net::SocketAddress& endpoint)
	: endpoint(endpoint)
{

}

BoardController::~BoardController()
{
}

BoardController::AbstractReceiver& BoardController::receiver()
{
	if(!receiver_)
	{
		setReceiver(ReceiverPtr(new DefaultReceiver(socket)));
	}
	return *receiver_;
}

void BoardController::resetReceiver()
{
	receiver_.reset();
}

void BoardController::doSendCommand(FeminosCommand& cmd)
{
	receiver().prepareForReply();
	int retryCount = RETRY_COUNT;
	do
	{
		LOG_DEBUG() << "Cmd: " << cmd.command.str();
		socket.doSend(endpoint, buffer(cmd.command.str()), sendTimeo);
		cmd.reply_ = receiver().waitForReply(rcveTimeo);
		if(!cmd.reply_)
		{
			LOG_WARN() << "Cmd: reply timeout";
		}
	}
	while(!cmd.reply_ and --retryCount > 0);

	if(!cmd.reply_)
	{
		throw CommandBadReply(cmd, Timeout);
	}

	if(!cmd.parse(buffer(cmd.reply_->buffer, cmd.reply_->length)))
	{
		throw CommandBadReply(cmd, BadFrameFormat);
	}
}

void BoardController::checkConfigFrame(FeminosCommand& frame)
{
	if(!frame.isFrame<ConfigFrame>())
	{
		throw CommandBadReply(frame, UnexpectedFrameContent);
	}
	const ConfigFrame& cframe = frame.frame<ConfigFrame>();
	if(cframe.frameStart().errorCode != 0)
	{
		throw CommandFailure(cframe.frameStart().errorCode, cframe.message().get());
	}
}

void BoardController::checkMonitoringFrame(FeminosCommand& frame)
{
	// When the command results in an error, a config frame is sent.
	if(frame.isFrame<ConfigFrame>())
	{
		checkConfigFrame(frame);
		if(frame.frame<ConfigFrame>().frameStart().errorCode == 0)
		{
			// if there was no error code, this is anormal.
			throw CommandBadReply(frame, UnexpectedFrameContent);
		}
	}
	if(!frame.isFrame<MonitFrame>())
	{
		throw CommandBadReply(frame, UnexpectedFrameContent);
	}
}

RegValue BoardController::register_read(RegisterID reg)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "reg " << reg;
	checkConfigFrame(cmd);
	FemNum femNum = -1;
	RegisterID regNum = -1;
	RegValue regValue = -1;
	char board[4];
	board[3] = '\0';
	if(sscanf(cmd.frame<ConfigFrame>().message().get(), "0 %3c(%u) Reg(%u) = %x", board, &femNum, &regNum, &regValue) != 4
	   or (strncmp(board, "Fem", 3) and strncmp(board, "Tcm", 3)))
	{
		throw CommandBadReply(cmd, UnexpectedMessageFormat);
	}
	if(regValue == 0xFFFFFFFF)
	{
		throw CommandBadReply(cmd, InvalidResult);
	}
	return regValue;
}

void BoardController::register_write(RegisterID reg, RegValue value)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "reg " << reg << ' ' << std::hex << std::showbase << value;
	checkConfigFrame(cmd);
}

void BoardController::cmdStats_clear()
{
	FeminosCommand cmd;
	sendCommand(cmd) << "cmd clr";
	checkConfigFrame(cmd);
}

// =============================================================================
// minos::fem::BoardController::CommandFailure class methods body
// -----------------------------------------------------------------------------

BoardController::CommandFailure::CommandFailure(int16_t code, const std::string& message)
	: errorCode(code), errorMessage(message)
{
}

BoardController::CommandFailure::~CommandFailure() throw()
{
}

void BoardController::CommandFailure::writeMessage(std::ostream& stream) const
{
	stream << "Command failure: " << errorMessage << " (code: " << errorCode << ')';
}

// =============================================================================
// minos::fem::BoardController::CommandBadReply class methods body
// -----------------------------------------------------------------------------

const char* BoardController::CommandBadReply::causeText[BadReplyCauseCount] = {
	"bad frame format",
	"unexpected frame content",
	"unexpected message format",
	"wrong channel",
	"timeout"
};

BoardController::CommandBadReply::CommandBadReply(FeminosCommand& command, CommandBadReplyCause cause)
	: command(command), cause(cause)
{
}

BoardController::CommandBadReply::~CommandBadReply() throw()
{
}

void BoardController::CommandBadReply::writeMessage(std::ostream& stream) const
{
	stream << "Bad reply to Feminos command '" << command.command.str() << "'(";
	if(cause < BadReplyCauseCount)
	{
		stream << causeText[cause];
	}
	else
	{
		stream << "unknown cause";
	}
	stream << ')';
}

// =============================================================================
// minos::fem::AsicRange class methods body
// -----------------------------------------------------------------------------


} /* namespace fem */
} /* namespace minos */
