/* =============================================================================
 * BoardController.h
 * -----------------------------------------------------------------------------
 * class minos::fem::BoardController
 * Created on: 8 août 2013 at Irfu/Sedi/Lilas, CEA Saclay, F-91191, France
 * Author: fchateau
 * -----------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * -----------------------------------------------------------------------------
 * Contributors: fchateau
 * -----------------------------------------------------------------------------
 * This software is part of the DaqMinos project.
 * -----------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding
 * by the rules of distribution of free software. You can use, modify and/or
 * redistribute the software under the terms of the CeCILL license as circulated
 * by CEA, CNRS and INRIA at the following URL: "http://www.cecill.info".
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability. In this
 * respect, the user's attention is drawn to the risks associated with loading,
 * using, modifying and/or developing or reproducing the software by the user in
 * light of its specific status of free software, that may mean that it is
 * complicated to manipulate, and that also therefore means that it is reserved
 * for developers and experienced professionals having in-depth computer
 * knowledge. Users are therefore encouraged to load and test the software's
 * suitability as regards their requirements in conditions enabling the security
 * of their systems and/or data to be ensured and, more generally, to use and
 * operate it in the same conditions as regards security.
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * -----------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for
 * commercial purposes. For this you will need to negotiate a specific contract
 * with a legal representative of CEA.
 * =============================================================================
 */

#ifndef minos_fem_BoardController_h_INCLUDED
#define minos_fem_BoardController_h_INCLUDED

#include "FeminosResult.h"
#include "Frame.h"
#include <minos/Exception.h>
#include <mdcs/utl/net/UdpTransceiver.h>
#include <utl/net/SocketAddress.h>
#include <boost/scoped_ptr.hpp>
#include <sstream>

namespace minos
{
namespace fem
{

class FeminosCommand : public Frame
{
public:

	friend class BoardController;
	ReplyBufferPtr& reply() { return reply_; }
private:
	std::ostringstream command;
	ReplyBufferPtr reply_;
};

typedef unsigned FemNum;
typedef unsigned RegisterID;
typedef uint32_t RegValue;

class BoardController
{
public:
	class AbstractReceiver
	{
	public:
		virtual void prepareForReply() = 0;
		virtual ReplyBufferPtr waitForReply(int timeout_ms) = 0;
	};
	class DefaultReceiver : public AbstractReceiver
	{
	public:
		DefaultReceiver(mdcs::utl::net::UdpTransceiver& transceiver);
		void prepareForReply();
		ReplyBufferPtr waitForReply(int timeout_ms);
	private:
		mdcs::utl::net::UdpTransceiver& transceiver_;
	};
	typedef std::auto_ptr<AbstractReceiver> ReceiverPtr;

	BoardController(const utl::net::SocketAddress& endpoint);
	virtual ~BoardController();

	mdcs::utl::net::UdpTransceiver& transceiver() { return socket; }
	const utl::net::SocketAddress& getEndpoint() const { return endpoint; }

	AbstractReceiver& receiver();
	void setReceiver(ReceiverPtr r) { receiver_.reset(r.release()); }
	void resetReceiver();

	RegValue register_read(RegisterID reg);
	void register_write(RegisterID reg, RegValue value);
	template<class Reg>	Reg register_read();
	template<class Reg>	void register_write(Reg value);

	void cmdStats_clear();

	class CommandBuilder
	{
	public:
		CommandBuilder(BoardController& ctrl, FeminosCommand& cmd) : ctrl(ctrl), cmd(cmd) {}
		~CommandBuilder() { ctrl.doSendCommand(cmd); }

		template<class T>
		CommandBuilder& operator<<(const T& val) { cmd.command << val; return *this; }
	private:
		BoardController& ctrl;
		FeminosCommand& cmd;
	};

	CommandBuilder sendCommand(FeminosCommand& cmd) { return CommandBuilder(*this, cmd); }
	void checkConfigFrame(FeminosCommand& frame);
	void checkMonitoringFrame(FeminosCommand& frame);

	class CommandBadReply;
	class CommandFailure;

protected:
	friend class CommandBuilder;
	void doSendCommand(FeminosCommand& cmd);

	static const mdcs::utl::net::Milliseconds sendTimeo = 1000;
	static const mdcs::utl::net::Milliseconds rcveTimeo = 3000;
	static const int RETRY_COUNT = 3;

	utl::net::SocketAddress endpoint;
	mdcs::utl::net::UdpTransceiver socket;
	boost::scoped_ptr<AbstractReceiver> receiver_;
};


class BoardController::CommandFailure : public Exception
{
public:
	CommandFailure(int16_t code, const std::string& message);
	~CommandFailure() throw();

	void writeMessage(std::ostream& stream) const;

	const int16_t errorCode;
	const std::string errorMessage;
};

enum CommandBadReplyCause
{
	BadFrameFormat = 0,
	UnexpectedFrameContent,
	UnexpectedMessageFormat,
	InvalidResult,
	Timeout,
	BadReplyCauseCount
};

class BoardController::CommandBadReply : public Exception
{
public:
	CommandBadReply(FeminosCommand& command, CommandBadReplyCause cause);
	~CommandBadReply() throw();

	void writeMessage(std::ostream& stream) const;

private:
	FeminosCommand& command;
	CommandBadReplyCause cause;

	static const char* causeText[BadReplyCauseCount];
};

// =============================================================================
// minos::fem::BoardController inline methods body
// -----------------------------------------------------------------------------

template<class Reg>
Reg BoardController::register_read()
{
	Reg reg = register_read(Reg::ID);
	return reg.value;
}

template<class Reg>
void BoardController::register_write(Reg value)
{
	register_write(Reg::ID, value.value);
}

} /* namespace fem */
} /* namespace minos */

#endif /* minos_fem_BoardController_h_INCLUDED */
