/**
 * @file BuiltEvent.cpp
 * @date 18 sept. 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the DaqMinos software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "BuiltEvent.h"
#include <utl/io/InplaceParser.ipp>
#include <utl/Logging.h>

namespace minos
{
namespace fem
{
using ::utl::io::InplaceParser;

// =============================================================================
// minos::fem::BuiltEvent private classes declarations
// -----------------------------------------------------------------------------

class BuiltEvent::Parser
{
public:
	Parser(Visitor& visitor, std::istream& stream);
	void run();

private:
	enum ParserState { Idle = 0, ParsingBuiltEvent, ParsingEvent, ParsingHitcount,
			ParsingCellPointer, ParsingChannelHit, ParsingWaveform, ParsingSamples };
	static const unsigned StateCount = ParsingSamples + 1;

	typedef bool (Parser::*StateParser)(uint16_t);

	bool parseIdle(uint16_t value);
	bool parseBuiltEvent(uint16_t value);
	bool parseEvent(uint16_t value);
	bool parseHitcount(uint16_t value);
	bool parseCellPointer(uint16_t value);
	bool parseChannelHit(uint16_t value);
	bool parseWaveform(uint16_t value);
	bool parseSamples(uint16_t value);

	template<class T> static bool checkPrefix(uint16_t value);
	template<class T> bool processPrefix(uint16_t alreadyRead, void (Visitor::*func)(const T&));

	static const StateParser parser[StateCount];
	Visitor& visitor;
	std::istream& stream;
	ParserState state;
};

const BuiltEvent::Parser::StateParser BuiltEvent::Parser::parser[BuiltEvent::Parser::StateCount] =
{
	&BuiltEvent::Parser::parseIdle,
	&BuiltEvent::Parser::parseBuiltEvent,
	&BuiltEvent::Parser::parseEvent,
	&BuiltEvent::Parser::parseHitcount,
	&BuiltEvent::Parser::parseCellPointer,
	&BuiltEvent::Parser::parseChannelHit,
	&BuiltEvent::Parser::parseWaveform,
	&BuiltEvent::Parser::parseSamples
};

// =============================================================================
// minos::fem::BuiltEvent::Parser class methods body
// -----------------------------------------------------------------------------

BuiltEvent::Parser::Parser(Visitor& visitor, std::istream& stream)
	: visitor(visitor), stream(stream), state(Idle)
{
}

void BuiltEvent::Parser::run()
{
	uint16_t value;
	while(stream.read(reinterpret_cast<char*>(&value), sizeof(uint16_t)))
	{
		while(!(this->*parser[state])(value));
	}
}

bool BuiltEvent::Parser::parseIdle(uint16_t value)
{
	if(processPrefix(value, &Visitor::onBuiltEventBegin))
	{
		state = ParsingBuiltEvent;
	}
	else
	{
		LOG_WARN() << "unrecognized data: " << std::hex << value << std::dec;
		throw std::runtime_error("invalid start of built event");
	}
	return true;
}

bool BuiltEvent::Parser::parseBuiltEvent(uint16_t value)
{
	if(processPrefix(value, &Visitor::onEventBegin))
	{
		state = ParsingEvent;
	}
	else if(processPrefix(value, &Visitor::onBuiltEventEnd))
	{
		state = Idle;
	}
	else
	{
		LOG_WARN() << "unrecognized data: " << std::hex << value << std::dec;
		throw std::runtime_error("invalid start of event");
	}
	return true;
}

bool BuiltEvent::Parser::parseEvent(uint16_t value)
{
	if(checkPrefix<fem::ChannelHitIndex>(value))
	{
		state = ParsingChannelHit;
	}
	else if(checkPrefix<fem::HitChannelCount>(value))
	{
		state = ParsingHitcount;
	}
	else if(checkPrefix<fem::LastCellRead>(value))
	{
		state = ParsingCellPointer;
	}
	else
	{
		visitor.onEventEnd();
		state = ParsingBuiltEvent;
	}
	return false;
}

bool BuiltEvent::Parser::parseHitcount(uint16_t value)
{
	if(!processPrefix(value, &Visitor::onHitChannelCount))
	{
		if(checkPrefix<fem::LastCellRead>(value))
		{
			state = ParsingCellPointer;
		}
		else if(checkPrefix<fem::ChannelHitIndex>(value))
		{
			state = ParsingChannelHit;
		}
		else
		{
			visitor.onEventEnd();
			state = ParsingBuiltEvent;
		}
		return false;
	}
	return true;
}

bool BuiltEvent::Parser::parseCellPointer(uint16_t value)
{
	if(!processPrefix(value, &Visitor::onLastCellRead))
	{
		if(checkPrefix<fem::ChannelHitIndex>(value))
		{
			state = ParsingChannelHit;
		}
		else
		{
			visitor.onEventEnd();
			state = ParsingBuiltEvent;
		}
		return false;
	}
	return true;
}

bool BuiltEvent::Parser::parseChannelHit(uint16_t value)
{
	if(processPrefix(value, &Visitor::onChannelHitBegin))
	{
		state = ParsingWaveform;
	}
	else
	{
		throw std::runtime_error("unexpected invalid channel hit");
	}
	return true;
}

bool BuiltEvent::Parser::parseWaveform(uint16_t value)
{
	if(checkPrefix<fem::TimeBinIndex>(value))
	{
		visitor.onWaveformBegin(reinterpret_cast<fem::TimeBinIndex*>(&value));
		state = ParsingSamples;
	}
	else if(value == 0)
	{
		visitor.onChannelHitEnd();
		state = ParsingEvent;
	}
	else
	{
		if(checkPrefix<fem::AdcSample>(value))
		{
			visitor.onWaveformBegin(0);
			state = ParsingSamples;
		}
		else
		{
			visitor.onChannelHitEnd();
			state = ParsingEvent;
		}
		return false;
	}
	return true;
}

bool BuiltEvent::Parser::parseSamples(uint16_t value)
{
	if(!processPrefix(value, &Visitor::onSample))
	{
		visitor.onWaveformEnd();
		state = ParsingWaveform;
		return false;
	}
	return true;
}

template<class T>
bool BuiltEvent::Parser::checkPrefix(uint16_t value)
{
	T data;
	uint16_t* rawData = reinterpret_cast<uint16_t*>(&data);
	*rawData = value;
	return data.prefix == data.PREFIX;
}

template<class T>
bool BuiltEvent::Parser::processPrefix(uint16_t alreadyRead, void (Visitor::*func)(const T&))
{
	T data;
	uint16_t* rawData = reinterpret_cast<uint16_t*>(&data);
	*rawData++ = alreadyRead;
	if(data.prefix == data.PREFIX)
	{
		if(sizeof(T) > sizeof(uint16_t))
		{
			stream.read(reinterpret_cast<char*>(rawData), sizeof(T) - sizeof(uint16_t));
		}
		(visitor.*func)(data);
		return true;
	}
	else
	{
		return false;
	}
}

// =============================================================================
// minos::fem::BuiltEvent::Visitor class methods body
// -----------------------------------------------------------------------------

BuiltEvent::Visitor::~Visitor()
{
}

// =============================================================================
// minos::fem::BuiltEvent::BinaryToTextVisitor class methods body
// -----------------------------------------------------------------------------

BuiltEvent::BinaryToTextVisitor::BinaryToTextVisitor(std::ostream& s)
	: stream(s), indentLevel(0), sampleCount(0), waveformCount(0), hitCount(0), eventCount(0)
{
}

void BuiltEvent::BinaryToTextVisitor::onBuiltEventBegin(const fem::BuiltEventHeader& header)
{
	++eventCount;
	indentLevel = 0;
	indent();
	stream << "BUILT EVENT(size=" << header.size << ')' << std::endl;
	++indentLevel;
}

void BuiltEvent::BinaryToTextVisitor::onBuiltEventEnd(const fem::BuiltEventFooter& /*footer*/)
{
	--indentLevel;
	indent();
	stream << "BUILT EVENT END (" << eventCount << " events)" << std::endl;
}

void BuiltEvent::BinaryToTextVisitor::onEventBegin(const fem::EventHeader& header)
{
	hitCount = 0;
	indent();
	stream << "EVENT BEGIN(id=" << header.eventId << ", ts=" << header.timestamp << ", type=" << header.type << ')' << std::endl;
	++indentLevel;
}

void BuiltEvent::BinaryToTextVisitor::onEventEnd()
{
	--indentLevel;
	indent();
	stream << "EVENT END (" << hitCount << " hits)" << std::endl;;
}

void BuiltEvent::BinaryToTextVisitor::onHitChannelCount(const fem::HitChannelCount& hitCount)
{
	indent();
	stream << "Hit count(card=" << hitCount.card << ", chip=" << hitCount.chip << ", count=" << hitCount.hitChannelCount << ')' << std::endl;
}

void BuiltEvent::BinaryToTextVisitor::onLastCellRead(const fem::LastCellRead& lastCellRead)
{
	indent();
	stream << "Last cell read(chip=" << lastCellRead.asic << ", cell=" << lastCellRead.readPointer << ')' << std::endl;
}

void BuiltEvent::BinaryToTextVisitor::onChannelHitBegin(const fem::ChannelHitIndex& hitIndex)
{
	waveformCount = 0;
	++hitCount;
	indent();
	stream << "CHANNEL HIT BEGIN(card=" << hitIndex.card << ", chip=" << hitIndex.chip << ", channel=" << hitIndex.channel << ')' << std::endl;
	++indentLevel;
}

void BuiltEvent::BinaryToTextVisitor::onChannelHitEnd()
{
	--indentLevel;
	indent();
	stream << "CHANNEL HIT END (" << waveformCount << " waveforms)" << std::endl;;
}

void BuiltEvent::BinaryToTextVisitor::onWaveformBegin(const fem::TimeBinIndex* binIndex)
{
	sampleCount = 0;
	++waveformCount;
	indent();
	stream << "WAVEFORM BEGIN(";
	if(binIndex != 0)
	{
		stream << "cell=" << binIndex->index;
	}
	stream << ')' << std::endl;
	++indentLevel;
	indent();
}

void BuiltEvent::BinaryToTextVisitor::onWaveformEnd()
{
	stream << std::endl;
	--indentLevel;
	indent();
	stream << "WAVEFORM END (" << sampleCount << " samples)" << std::endl;
}

void BuiltEvent::BinaryToTextVisitor::onSample(const fem::AdcSample& sample)
{
	stream << sample.value << ' ';
	++sampleCount;
}

void BuiltEvent::BinaryToTextVisitor::indent()
{
	for(int i = 0; i < indentLevel; ++i)
	{
		stream << "  ";
	}
}

// =============================================================================
// minos::fem::BuiltEvent::UntilSize class methods body
// -----------------------------------------------------------------------------

BuiltEvent::UntilSize::UntilSize(const ::utl::io::ConstBuffer& buf)
	: InplaceParser(buf)
{
	if(!parse_1(builtEvtHeader))
	{
		throw std::runtime_error("invalid built event");
	}
}

// =============================================================================
// minos::fem::BuiltEvent::UntilId class methods body
// -----------------------------------------------------------------------------

BuiltEvent::UntilId::UntilId(const ::utl::io::ConstBuffer& buf)
	: UntilSize(buf)
{
	if(!parse_1(header))
	{
		throw std::runtime_error("invalid event");
	}
}

// =============================================================================
// minos::fem::BuiltEvent class methods body
// -----------------------------------------------------------------------------

void BuiltEvent::read(std::istream& s, Visitor& v)
{
	Parser parser(v, s);
	parser.run();
}

} /* namespace fem */
} /* namespace minos */
