/**
 * @file DetectorMapping.cpp
 * @date 8 oct. 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the DaqMinos software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "DetectorMapping.h"
#include "FrameTypes.h"
#include <utl/Logging.h>

namespace minos
{
namespace fem
{

// =============================================================================
// minos::fem::ProjectiveTpcPad class methods body
// -----------------------------------------------------------------------------

bool ProjectiveTpcPad::isValid() const
{
	return ring >= 0 and ring < TpcRingCount and ray >= 0 and ray < TpcRayCount;
}

// =============================================================================
// minos::fem::ConstantTpcPad class methods body
// -----------------------------------------------------------------------------

const int ConstantTpcPad::lowerOffset[TpcRingCount] = {14, 13, 12, 11, 10, 10, 9, 8, 7, 6, 5, 5, 4, 3, 2, 1, 1, 0};
const int ConstantTpcPad::upperOffset[TpcRingCount] = {14, 13, 13, 12, 11, 10, 9, 8, 8, 7, 6, 5, 4, 4, 3, 2, 1, 0};

#define RING_SIZE(N) TpcRayCount - ConstantTpcPad::lowerOffset[N] - ConstantTpcPad::upperOffset[N]

const int ConstantTpcPad::ringSize[TpcRingCount] = {
	RING_SIZE(0),
	RING_SIZE(1),
	RING_SIZE(2),
	RING_SIZE(3),
	RING_SIZE(4),
	RING_SIZE(5),
	RING_SIZE(6),
	RING_SIZE(7),
	RING_SIZE(8),
	RING_SIZE(9),
	RING_SIZE(10),
	RING_SIZE(11),
	RING_SIZE(12),
	RING_SIZE(13),
	RING_SIZE(14),
	RING_SIZE(15),
	RING_SIZE(16),
	RING_SIZE(17)
};

// =============================================================================
// minos::fem::DetectorMapping class methods body
// -----------------------------------------------------------------------------

DetectorMapping::DetectorMapping()
{
	// use default wiring
	for(int i = 0; i < FeminosMaxCount; ++i)
	{
		for(int j = 0; j < 2; ++j)
		{
			wiringInfos_[i][j].board = i;
			wiringInfos_[i][j].isUpperChips = j;
			wiringInfos_[i][j].isDefault = 1;
		}
	}
}

void DetectorMapping::setWiring(fem::WiringInfo wiring)
{
	wiringInfos_[wiring.board][wiring.isUpperChips] = wiring;
}

fem::WiringInfo DetectorMapping::getWiring(const Channel& channel) const
{
	return wiringInfos_[channel.card][channel.chip / 2];
}

bool DetectorMapping::toAGetChannel(const Channel& daqChannel, Channel& agetChannel)
{
	agetChannel.chip = daqChannel.chip;
	agetChannel.card = daqChannel.card;
	int channel = daqChannel.channel;
	if(channel > 1 and channel < 13)
	{
		agetChannel.channel = channel - 1;
	}
	else if(channel > 13 and channel < 24)
	{
		agetChannel.channel = channel - 2;
	}
	else if(channel > 24 and channel < 47)
	{
		agetChannel.channel = channel - 3;
	}
	else if(channel > 47 and channel < 58)
	{
		agetChannel.channel = channel - 4;
	}
	else if(channel > 58 and channel < 70)
	{
		agetChannel.channel = channel - 5;
	}
	else
	{
//		LOG_DEBUG() << "unmapped channel " << daqChannel.channel;
		return false;
	}
	return true;
}

bool DetectorMapping::toAfterChannel(const Channel& daqChannel, Channel& afterChannel)
{
	afterChannel.chip = daqChannel.chip;
	afterChannel.card = daqChannel.card;
	int channel = daqChannel.channel;
	if(channel > 2 and channel < 15)
	{
		afterChannel.channel = channel - 2;
	}
	else if(channel > 15 and channel < 28)
	{
		afterChannel.channel = channel - 3;
	}
	else if(channel > 28 and channel < 53)
	{
		afterChannel.channel = channel - 4;
	}
	else if(channel > 53 and channel < 66)
	{
		afterChannel.channel = channel - 5;
	}
	else if(channel > 66 and channel < 79)
	{
		afterChannel.channel = channel - 6;
	}
	else
	{
//		LOG_DEBUG() << "unmapped channel " << daqChannel.channel;
		return false;
	}
	return true;
}

bool DetectorMapping::agetChannelToFecPin(const Channel& agetChannel, FecPin& fecPin)
{
	fecPin.card = agetChannel.card;
	fecPin.chip = agetChannel.chip;
	int pinSeq = agetChannel.channel;
	if(agetChannel.channel > 16) ++pinSeq;
	if(agetChannel.channel > 48) ++pinSeq;
	fecPin.sideB = pinSeq % 2;
	fecPin.pin   = (pinSeq / 2) + 4;
	return true;
}

bool DetectorMapping::afterChannelToFecPin(const Channel& afterChannel, FecPin& fecPin)
{
	fecPin.card = afterChannel.card;
	fecPin.chip = afterChannel.chip;
	fecPin.sideB = (afterChannel.channel - 1) % 2;
	fecPin.pin   = ((afterChannel.channel - 1) / 2) + 3;
	return true;
}

bool DetectorMapping::toCoaxPin(const FecPin& fecPin, CoaxPin& coaxPin)
{
	coaxPin.card = fecPin.card;
	coaxPin.baseChip = (fecPin.chip / 2) * 2;
	int indexForward = (fecPin.pin - 1) * 2 + fecPin.sideB;
	int indexBackward = 56 - ((fecPin.pin - 1) * 2 - fecPin.sideB);
	if(indexForward > 6 and indexForward < 23)
	{
		coaxPin.pin = indexForward - 6;
	}
	else if(indexForward > 23 and indexBackward > 0)
	{
		coaxPin.pin = fecPin.pin * 2 - fecPin.sideB - 8;
	}
	else if(indexForward > 57 and indexForward < 73)
	{
		coaxPin.pin = indexForward - 8;
	}
	else
	{
//		LOG_DEBUG() << "invalid fec pin: " << (char)('A' + fecPin.sideB) << fecPin.pin;
		return false;
	}
	coaxPin.pin += (1 - (fecPin.chip % 2)) * 64; // pins for pair chips are shifted by 64
	return true;
}

//void DetectorMapping::toConnectorPin(const CoaxPin& coaxPin, ConnectorPin& connPin)
//{
//	connPin.card = coaxPin.card;
//	if(coaxPin.pin > 0 and coaxPin.pin <= 32)
//	{
//		connPin.pin = (coaxPin.pin + 2) * 2;
//	}
//	else if(coaxPin.pin > 32 and coaxPin.pin <= 64)
//	{
//		connPin.pin = (coaxPin.pin + 4) * 2;
//	}
//	else if(coaxPin.pin > 64 and coaxPin.pin <= 96)
//	{
//		connPin.pin = (coaxPin.pin - 62) * 2 - 1;
//	}
//	else if(coaxPin.pin > 96 and coaxPin.pin <= 128)
//	{
//		connPin.pin = (coaxPin.pin - 60) * 2 - 1;
//	}
//	else
//	{
//		LOG_ERROR() << "invalid coax pin " << coaxPin.pin;
//	}
//}

bool DetectorMapping::daqChannelToCoaxPin(const Channel& daqChannel, CoaxPin& coaxPin, AsicType asicType)
{
	Channel asicChannel;
	if((asicType == AfterAsic and toAfterChannel(daqChannel, asicChannel)) or
	   (asicType == AGetAsic  and toAGetChannel(daqChannel, asicChannel)))
	{
		FecPin fecPin;
		if((asicType == AfterAsic and afterChannelToFecPin(asicChannel, fecPin)) or
		   (asicType == AGetAsic  and agetChannelToFecPin(asicChannel, fecPin)))
		{
			return toCoaxPin(fecPin, coaxPin);
		}
	}
	return false;
}

bool DetectorMapping::toTpcPad(const CoaxPin& coaxPin, ProjectiveTpcPad& pad) const
{
	Connector conn;
	cableToConnector(coaxPin, conn);
	if(conn.quadrant < 0 or conn.quadrant > 3)
	{
//		LOG_DEBUG() << "invalid quadrant " << conn.quadrant;
		return true;
	}
	if(conn.connector < 0 or conn.connector > 9)
	{
//		LOG_DEBUG() << "invalid connector " << conn.connector;
	}
	pad.ring = (conn.connector * 2) + (1 - (coaxPin.pin - 1) / 64);
	pad.ray = conn.quadrant * 64 + (coaxPin.pin - 1) % 64;
	return true;
}

bool DetectorMapping::toConstantPad(const ProjectiveTpcPad& projPad, ConstantTpcPad& pad) const
{
	pad.ring = projPad.ring;
	int pin = projPad.ray % 64;
	int firstPin = ConstantTpcPad::lowerOffset[pad.ring];
	int lastPin  = TpcRayCount/4 - 1 - ConstantTpcPad::upperOffset[pad.ring];

	if(pin < firstPin or pin > lastPin)
	{
		return false;
	}
	else
	{
		pad.sector = projPad.ray;
		return true;
	}
}

bool DetectorMapping::cableToConnector(const CoaxCable& cable, Connector& conn) const
{
	if(cable.card < 0 or cable.card > FeminosMaxCount or (cable.baseChip != 0 and cable.baseChip != 2))
	{
		return false;
	}

	WiringInfo wiring = wiringInfos_[cable.card][cable.baseChip / 2];
	if(wiring.device != ProjectiveTPC and wiring.device != ConstantTPC)
	{
		return false;
	}
	if(wiring.isDefault)
	{
		if(cable.card < 0 or cable.card > 20)
		{
//			LOG_ERROR() << "invalid card number " << cable.card;
			return false;
		}
		conn.connector = (cable.card % 5) * 2 + cable.baseChip / 2;
		conn.quadrant = cable.card / 5;
	}
	else
	{
		conn.connector = wiring.tpcConnector;
		conn.quadrant = wiring.tpcQuadrant;
	}
	return true;
}

bool DetectorMapping::daqChannelToTpcPad(const Channel& daqChannel, ProjectiveTpcPad& pad, AsicType asicType) const
{
	CoaxPin coaxPin;
	if(daqChannelToCoaxPin(daqChannel, coaxPin, asicType))
	{
		return toTpcPad(coaxPin, pad);
	}
	return false;
}

bool DetectorMapping::daqChannelToTpcPad(const Channel& daqChannel, ConstantTpcPad& pad, AsicType asicType) const
{
	ProjectiveTpcPad projPad;
	if(daqChannelToTpcPad(daqChannel, projPad, asicType))
	{
		return toConstantPad(projPad, pad);
	}
	return false;
}

} /* namespace fem */
} /* namespace minos */
