/* =============================================================================
 * FeminosController.cpp
 * -----------------------------------------------------------------------------
 * class minos::fem::FeminosController
 * Created on: 13 déc. 2012 at Irfu/Sedi/Lilas, CEA Saclay, F-91191, France
 * Author: fchateau
 * -----------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * -----------------------------------------------------------------------------
 * Contributors: fchateau
 * -----------------------------------------------------------------------------
 * This software is part of the DaqMinos project.
 * -----------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding
 * by the rules of distribution of free software. You can use, modify and/or
 * redistribute the software under the terms of the CeCILL license as circulated
 * by CEA, CNRS and INRIA at the following URL: "http://www.cecill.info".
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability. In this
 * respect, the user's attention is drawn to the risks associated with loading,
 * using, modifying and/or developing or reproducing the software by the user in
 * light of its specific status of free software, that may mean that it is
 * complicated to manipulate, and that also therefore means that it is reserved
 * for developers and experienced professionals having in-depth computer
 * knowledge. Users are therefore encouraged to load and test the software's
 * suitability as regards their requirements in conditions enabling the security
 * of their systems and/or data to be ensured and, more generally, to use and
 * operate it in the same conditions as regards security.
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * -----------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for
 * commercial purposes. For this you will need to negotiate a specific contract
 * with a legal representative of CEA.
 * =============================================================================
 */

#include "FeminosController.h"
#include <utl/compat/ArrayHelper.hpp>
#include <utl/Logging.h>
#include <iomanip>
#include <cstdio>

namespace minos
{
namespace fem
{
using namespace mdcs::utl::net;

// =============================================================================
// minos::fem::FeminosController class methods body
// -----------------------------------------------------------------------------

FeminosController::FeminosController(const utl::net::SocketAddress& endpoint)
	: BoardController(endpoint)
{
}

FeminosController::~FeminosController()
{
}

void FeminosController::setAsicType(const std::string& asicName)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "mode " << asicName;
	checkConfigFrame(cmd);
}

FemNum FeminosController::getFeminosNumber()
{
	FeminosCommand cmd;
	sendCommand(cmd) << "reg 0";
	checkConfigFrame(cmd);
	FemNum femNum = -1;
	if(sscanf(cmd.frame<ConfigFrame>().message().get(), "0 Fem(%u) Reg(0)", &femNum) != 1)
	{
		throw CommandBadReply(cmd, UnexpectedMessageFormat);
	}
	return femNum;
}

void FeminosController::testdata_fillPattern(TestPattern pattern, PatternLength length)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "tdata " << ('A' + pattern) << ' ' << length;
	checkConfigFrame(cmd);
}

AfterRegValue FeminosController::after_read(AsicID asic, RegisterID reg)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "after " << asic << " read " << reg;
	checkConfigFrame(cmd);
	FemNum femNum = -1;
	AsicNum asicNum = -1;
	RegisterID regNum = -1;
	AfterRegValue regValue = -1;
	if(sscanf(cmd.frame<ConfigFrame>().message().get(), "0 Fem(%u) After(%hu) Reg(%u): %hx",
		&femNum, &asicNum, &regNum, &regValue) != 4)
	{
		throw CommandBadReply(cmd, UnexpectedMessageFormat);
	}
	if(regValue == 0xFFFF)
	{
		throw CommandBadReply(cmd, InvalidResult);
	}
	return regValue;
}

void FeminosController::after_write(AsicRange asic, RegisterID reg, AfterRegValue value)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "after " << asic << " write " << reg << ' ' << value;
	checkConfigFrame(cmd);
}

AGetAccessor FeminosController::aget_read(AsicID asic)
{
	return AGetAccessor(*this, asic);
}

AGetMutator FeminosController::aget_write(AsicRange asics)
{
	return AGetMutator(*this, asics);
}

PedValue FeminosController::pedestal_get(AsicID asic, ChannelID channel)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "ped " << asic << ' ' << channel;
	checkConfigFrame(cmd);
	unsigned asicReply = -1, channelReply = -1, value = -1;
	if(sscanf(cmd.frame<ConfigFrame>().message().get(), "0 ped[%u][%u] = %x", &asicReply, &channelReply, &value) != 3)
	{
		throw CommandBadReply(cmd, UnexpectedMessageFormat);
	}
	return static_cast<PedValue>(value);
}

void FeminosController::pedestal_list(ListResult& result, AsicRange asic)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "list ped " << asic;
	checkMonitoringFrame(cmd);
	const MonitFrame& frame = cmd.frame<MonitFrame>();
	if(frame.pedestalList().size() == 0)
	{
		throw CommandBadReply(cmd, UnexpectedFrameContent);
	}
	result.setBuffer(cmd.reply());
	result.lists_ = frame.pedestalList();
}

void FeminosController::pedestal_set(AsicRange asic, ChannelRange channel, PedValue value)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "ped " << asic << ' ' << channel << ' ' << value;
	checkConfigFrame(cmd);
}

ThrValue FeminosController::threshold_get(AsicID asic, ChannelID channel)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "thr " << asic << ' ' << channel;
	checkConfigFrame(cmd);
	unsigned asicReply = -1, channelReply = -1, value = -1;
	if(sscanf(cmd.frame<ConfigFrame>().message().get(), "0 thr[%u][%u] = %x", &asicReply, &channelReply, &value) != 3)
	{
		throw CommandBadReply(cmd, UnexpectedMessageFormat);
	}
	return static_cast<ThrValue>(value);
}

void FeminosController::threshold_set(AsicRange asic, ChannelRange channel, ThrValue value)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "thr " << asic << ' ' << channel << ' ' << value;
	checkConfigFrame(cmd);
}

void FeminosController::threshold_list(ListResult& result, AsicRange asic)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "list thr " << asic;
	checkMonitoringFrame(cmd);
	const MonitFrame& frame = cmd.frame<MonitFrame>();
	if(frame.pedestalList().size() == 0)
	{
		throw CommandBadReply(cmd, UnexpectedFrameContent);
	}
	result.setBuffer(cmd.reply());
	result.lists_ = frame.pedestalList();
}

DaqTarget FeminosController::daq_getTarget()
{
	FeminosCommand cmd;
	sendCommand(cmd) << "serve_target";
	checkConfigFrame(cmd);
	FemNum femNum, target;
	if(sscanf(cmd.frame<ConfigFrame>().message().get(), "0 Fem(%u) Serve_Target: %u", &femNum, &target) != 2)
	{
		throw CommandBadReply(cmd, UnexpectedMessageFormat);
	}
	return static_cast<DaqTarget>(target);
}

void FeminosController::daq_setTarget(DaqTarget target)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "serve_target " << target;
	checkConfigFrame(cmd);
}

void FeminosController::daq_getCredits(DaqCredit& credits, DaqCreditUnit& unit)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "daq";
	checkConfigFrame(cmd);

	FemNum femNum;
	if(sscanf(cmd.frame<ConfigFrame>().message().get(), "0 Fem(%u) send_credits: %d B", &femNum, &credits) != 2)
	{
		throw CommandBadReply(cmd, UnexpectedMessageFormat);
	}
	unit = ByteCredits;
}

void FeminosController::daq_addCredits(DaqCredit credits, DaqCreditUnit unit, bool resetSeq)
{
	static uint8_t sequenceNumber = 0;
	std::ostringstream cmd;
	if(resetSeq)
	{
		sequenceNumber = 0;
	}
	else
	{
		++sequenceNumber;
	}
	cmd << "daq 0x"
		<< std::hex << std::setfill('0') << std::setw(6) << credits
		<< ' ' << (unit == ByteCredits ? 'B' : 'F')
		<< " 0x"<< std::setw(2) << static_cast<unsigned>(sequenceNumber);
	LOG_DEBUG() << cmd.str();
	socket.doSend(endpoint, buffer(cmd.str()), sendTimeo);
}

void FeminosController::daq_clearCredits()
{
	FeminosCommand cmd;
	sendCommand(cmd) << "daq 0xFFFFFF B";
	checkConfigFrame(cmd);
}

DaqLossPolicy FeminosController::daq_getLossPolicy()
{
	FeminosCommand cmd;
	sendCommand(cmd) << "loss_policy";
	checkConfigFrame(cmd);

	FemNum femNum;
	unsigned policy;
	if(sscanf(cmd.frame<ConfigFrame>().message().get(), "0 Fem(%u): Loss_Policy : %u", &femNum, &policy) != 2)
	{
		throw CommandBadReply(cmd, UnexpectedMessageFormat);
	}
	return static_cast<DaqLossPolicy>(policy);
}

void FeminosController::daq_setLossPolicy(DaqLossPolicy policy)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "loss_policy " << policy;
	checkConfigFrame(cmd);
}

uint16_t FeminosController::daq_getCreditWaitTime()
{
	FeminosCommand cmd;
	sendCommand(cmd) << "cred_wait_time";
	checkConfigFrame(cmd);

	FemNum femNum;
	int waitTime = -1;
	if(sscanf(cmd.frame<ConfigFrame>().message().get(), "0 Fem(%u): Credit_Wait_Time : %d ms", &femNum, &waitTime) != 2)
	{
		throw CommandBadReply(cmd, UnexpectedMessageFormat);
	}
	if(waitTime < 0 or waitTime > 0xFFFF)
	{
		throw CommandBadReply(cmd, InvalidResult);
	}
	return waitTime;
}

void FeminosController::daq_setCreditWaitTime(uint16_t waitTime)
{
	//range is [0;42949] ms
	FeminosCommand cmd;
	sendCommand(cmd) << "cred_wait_time " << waitTime;
	checkConfigFrame(cmd);
}

void FeminosController::buffer_getInfo(BufferInfo& info)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "rbf config";
	checkConfigFrame(cmd);

	FemNum femNum;
	unsigned isRunning;
	char mode[16];
	char timeout[8];
	if(sscanf(cmd.frame<ConfigFrame>().message().get(),
		"0 Fem(%u) AXI Ring Buffer Base:%x Buf_Capacity:%u bytes Run:%u RetPnd:%u Mode:%13s Timeout:%5s",
		&femNum, &info.baseAddress, &info.capacity, &isRunning, &info.pndCount, mode, timeout) != 7)
	{
		throw CommandBadReply(cmd, UnexpectedMessageFormat);
	}
	info.isRunActive = isRunning;
	info.hasTimeout = buffer_isTimeoutEnabled(cmd, mode);
	info.timeout = buffer_getTimeout(cmd, timeout);
}

void FeminosController::buffer_setActive(bool isActive)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "rbf " << (isActive ? "resume" : "suspend");
	checkConfigFrame(cmd);
}

bool FeminosController::buffer_isTimeoutEnabled()
{
	FeminosCommand cmd;
	sendCommand(cmd) << "rbf timed";
	checkConfigFrame(cmd);
	FemNum femNum;
	char mode[16];
	if(sscanf(cmd.frame<ConfigFrame>().message().get(), "0 Fem(%u) AXI Ring Buffer Mode: %13s", &femNum, mode) != 2)
	{
		throw CommandBadReply(cmd, UnexpectedMessageFormat);
	}
	return buffer_isTimeoutEnabled(cmd, mode);
}

void FeminosController::buffer_enableTimeout(bool enabled)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "rbf timed " << (enabled ? "1" : "0");
	checkConfigFrame(cmd);
}

BufferTimeout FeminosController::buffer_getTimeout()
{
	FeminosCommand cmd;
	sendCommand(cmd) << "rbf timeval";
	checkConfigFrame(cmd);
	FemNum femNum;
	unsigned timeout;
	if(sscanf(cmd.frame<ConfigFrame>().message().get(), "0 Fem(%u) AXI Ring Buffer Timeout: %u", &femNum, &timeout) != 2)
	{
		throw CommandBadReply(cmd, UnexpectedMessageFormat);
	}
	if(timeout > BufferTimeout_1000ms)
	{
		throw CommandBadReply(cmd, InvalidResult);
	}
	return static_cast<BufferTimeout>(timeout);
}

void FeminosController::buffer_setTimeout(BufferTimeout timeout)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "rbf timeval " << timeout;
	checkConfigFrame(cmd);
}

void FeminosController::buffer_reset()
{
	FeminosCommand cmd;
	sendCommand(cmd) << "rbf reset";
	checkConfigFrame(cmd);
}

void FeminosController::buffer_sendPendingData()
{
	FeminosCommand cmd;
	sendCommand(cmd) << "rbf getpnd";
	checkConfigFrame(cmd);
}

void FeminosController::tcmStats_clear()
{
	FeminosCommand cmd;
	sendCommand(cmd) << "tcm clr";
	checkConfigFrame(cmd);
}

void FeminosController::hbusy_clear()
{
	FeminosCommand cmd;
	sendCommand(cmd) << "hbusy clr";
	checkConfigFrame(cmd);
}

void FeminosController::hbusy_getHisto(DeadtimeHistoResult& result)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "hbusy get";
	checkMonitoringFrame(cmd);
	const MonitFrame& frame = cmd.frame<MonitFrame>();
	if(!frame.deadtimeStats().isSet() or
	   (!frame.deadtimeHisto().isSet() and
	      frame.frameStart().payloadSize > (sizeof(DeadtimeStats) + sizeof(MonitFrameStart) + 2)))
	{
		throw CommandBadReply(cmd, UnexpectedFrameContent);
	}
	result.setBuffer(cmd.reply());
	result.deadtimeStats_ = frame.deadtimeStats();
	result.deadtimeHisto_ = frame.deadtimeHisto();
}

void FeminosController::hped_clear(AsicRange asic, ChannelRange channel)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "hped clr " << asic << ' ' << channel;
	checkConfigFrame(cmd);
}

void FeminosController::hped_getSummary(PedestalSummaryResult& result, AsicID asic, ChannelRange channel)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "hped getsummary " << asic << ' ' << channel;
	checkMonitoringFrame(cmd);
	const MonitFrame& frame = cmd.frame<MonitFrame>();
	if(!frame.pedestalSummary().isSet())
	{
		throw CommandBadReply(cmd, UnexpectedFrameContent);
	}
	result.setBuffer(cmd.reply());
	result.pedestalSummary_ = frame.pedestalSummary();
}

void FeminosController::hped_getStats(PedestalStatsResult& result, AsicID asic, ChannelID channel)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "hped getmath " << asic << ' ' << channel;
	checkMonitoringFrame(cmd);
	const MonitFrame& frame = cmd.frame<MonitFrame>();
	if(!frame.histoChannel().isSet() or !frame.pedestalStats().isSet())
	{
		throw CommandBadReply(cmd, UnexpectedFrameContent);
	}
	result.setBuffer(cmd.reply());
	result.histoChannel_  = frame.histoChannel();
	result.pedestalStats_ = frame.pedestalStats();
}

void FeminosController::hped_getHisto(PedestalHistoResult& result, AsicID asic, ChannelID channel)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "hped getbins " << asic << ' ' << channel;
	checkMonitoringFrame(cmd);
	const MonitFrame& frame = cmd.frame<MonitFrame>();
	if(!frame.histoChannel().isSet() or
	  (!frame.pedestalHisto().isSet() and frame.frameStart().payloadSize > 8))
	{
		throw CommandBadReply(cmd, UnexpectedFrameContent);
	}
	result.setBuffer(cmd.reply());
	result.histoChannel_  = frame.histoChannel();
	result.pedestalHisto_ = frame.pedestalHisto();
}

void FeminosController::hped_setOffset(AsicRange asic, ChannelRange channel, OffsetValue offset)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "hped offset " << asic << ' ' << channel << ' ' << offset;
	checkConfigFrame(cmd);
}

void FeminosController::hped_setThreshold(AsicRange asic, ChannelRange channel, ThrValue threshold, float sigma)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "hped setthr " << asic << ' ' << channel << ' ' << threshold << ' ' << sigma;
	checkConfigFrame(cmd);
}

void FeminosController::hped_center(AsicRange asic, ChannelRange channel, PedValue value)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "hped centermean " << asic << ' ' << channel << ' ' << value;
	checkConfigFrame(cmd);
}

void FeminosController::shisto_clear(AsicRange asic, ChannelRange channel)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "shisto clr " << asic << ' ' << channel;
	checkConfigFrame(cmd);
}

void FeminosController::shisto_getHisto(SHistoResult& result, AsicID asic, ChannelID channel)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "shisto getbins " << asic << ' ' << channel;
	checkMonitoringFrame(cmd);
	const MonitFrame& frame = cmd.frame<MonitFrame>();
	if(!frame.sHisto().isSet())
	{
		throw CommandBadReply(cmd, UnexpectedFrameContent);
	}
	result.setBuffer(cmd.reply());
	result.sHisto_ = frame.sHisto();
}

void FeminosController::shisto_setThreshold(unsigned threshold)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "shisto thr * * " << std::hex << std::showbase << threshold;
	checkConfigFrame(cmd);
}

void FeminosController::hhit_clear(AsicRange asic)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "hhit clr " << asic;
	checkConfigFrame(cmd);
}

void FeminosController::hhit_getHisto(HitCountHistoResult& result, AsicID asic)
{
	FeminosCommand cmd;
	sendCommand(cmd) << "hhit get " << asic;
	checkMonitoringFrame(cmd);
	const MonitFrame& frame = cmd.frame<MonitFrame>();
	if(!frame.hitCountHisto().isSet())
	{
		throw CommandBadReply(cmd, UnexpectedFrameContent);
	}
	result.setBuffer(cmd.reply());
	result.hitCountHisto_ = frame.hitCountHisto();
}

BufferTimeout FeminosController::buffer_getTimeout(FeminosCommand& cmd, const char* text)
{
	static const char* const strTimeouts[] = {"1ms", "10ms", "100ms", "1s"};
	for(unsigned i = 0; i < utl::array::size(strTimeouts); ++i)
	{
		if(!strcmp(text, strTimeouts[i]))
		{
			return static_cast<BufferTimeout>(i);
		}
	}
	throw CommandBadReply(cmd, InvalidResult);
}

bool FeminosController::buffer_isTimeoutEnabled(FeminosCommand& cmd, const char* text)
{
	if(!strcmp(text, "infinite_wait"))
	{
		return false;
	}
	else if(!strcmp(text, "timed_wait"))
	{
		return true;
	}
	else
	{
		throw CommandBadReply(cmd, InvalidResult);
	}
}
// =============================================================================
// minos::fem::AGetAccessor class methods body
// -----------------------------------------------------------------------------

AGetAccessor::AGetAccessor(FeminosController& controller, AsicID asic)
	: ctrl_(controller), asic_(asic)
{
}

// =============================================================================
// minos::fem::AGetMutator class methods body
// -----------------------------------------------------------------------------

AGetMutator::AGetMutator(FeminosController& controller, AsicRange asics)
	: ctrl_(controller), asics_(asics)
{
}

AGetMutator& AGetMutator::setGain(ChannelRange channels, short gain)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " gain " << channels << ' ' << gain;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setShapingTime(ChannelRange channels, short shapingTime)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " time " << channels << ' ' << shapingTime;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setThreshold(ChannelRange channels, short threshold)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " threshold " << channels << ' ' << threshold;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setInhibit(ChannelRange channels, bool inhibit)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " inhibit " << channels << ' ' << inhibit;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setEnableMarkerReset(bool enable)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " en_mkr_rst " << enable;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setResetLevel(bool level)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " rst_level " << level;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setReadFromZero(bool readFromZero)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " rd_from_0 " << readFromZero;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setTestDigOut(bool testDigOut)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " tst_digout " << testDigOut;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setModeReadAll(bool readAll)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " mode " << readAll;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setPolarityPositive(bool positive)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " polarity " << positive;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setReadFpn(bool readFpn)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " fpn " << readFpn;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setIcmVoltage(short icmVoltage)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " vicm " << icmVoltage;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setDacThreshold(short dacThreshold)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " dac " << dacThreshold;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setTriggerVeto(short triggerVeto)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " trigger_veto " << triggerVeto;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setSynchroDiscri(bool synchroDiscri)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " synchro_discri " << synchroDiscri;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setTimeOverThreshold(bool timeOverThreshold)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " tot " << timeOverThreshold;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setRangeTriggerWidth(bool rangeTriggerWidth)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " range_tw " << rangeTriggerWidth;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setTriggerWidth(short triggerWidth)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " trig_width " << triggerWidth;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setShortRead(bool shortRead)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " short_read " << shortRead;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setDisableMultiplicityOut(bool disable)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " dis_multiplicity_out " << disable;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

AGetMutator& AGetMutator::setAutoresetBank(bool autoreset)
{
	FeminosCommand cmd;
	ctrl_.sendCommand(cmd) << "aget " << asics_ << " autoreset_bank " << autoreset;
	ctrl_.checkConfigFrame(cmd);
	return *this;
}

} /* namespace fem */
} /* namespace minos */
