/* =============================================================================
 * FeminosController.h
 * -----------------------------------------------------------------------------
 * class minos::fem::FeminosController
 * Created on: 13 déc. 2012 at Irfu/Sedi/Lilas, CEA Saclay, F-91191, France
 * Author: fchateau
 * -----------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * -----------------------------------------------------------------------------
 * Contributors: fchateau
 * -----------------------------------------------------------------------------
 * This software is part of the DaqMinos project.
 * -----------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding
 * by the rules of distribution of free software. You can use, modify and/or
 * redistribute the software under the terms of the CeCILL license as circulated
 * by CEA, CNRS and INRIA at the following URL: "http://www.cecill.info".
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability. In this
 * respect, the user's attention is drawn to the risks associated with loading,
 * using, modifying and/or developing or reproducing the software by the user in
 * light of its specific status of free software, that may mean that it is
 * complicated to manipulate, and that also therefore means that it is reserved
 * for developers and experienced professionals having in-depth computer
 * knowledge. Users are therefore encouraged to load and test the software's
 * suitability as regards their requirements in conditions enabling the security
 * of their systems and/or data to be ensured and, more generally, to use and
 * operate it in the same conditions as regards security.
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * -----------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for
 * commercial purposes. For this you will need to negotiate a specific contract
 * with a legal representative of CEA.
 * =============================================================================
 */

#ifndef minos_fem_FeminosController_h_INCLUDED
#define minos_fem_FeminosController_h_INCLUDED

#include "BoardController.h"
#include "FeminosResult.h"
#include "AsicChannelID.h"
#include "Frame.h"

namespace minos
{
namespace fem
{

enum TestPattern
{
	Pattern_RaisingRamp = 0,
	Pattern_FallingRamp,
	Pattern_Pulse
};

enum DaqTarget
{
	DataDiscard = 0,
	DataToDaq,
	DataToPedestal,
	DataToChannelHit
};

enum DaqLossPolicy
{
	DaqLossIgnore = 0,
	DaqLossRecredit,
	DaqLossResend
};

enum DaqCreditUnit
{
	ByteCredits,
	FrameCredits
};

enum BufferTimeout
{
	BufferTimeout_1ms = 0,
	BufferTimeout_10ms,
	BufferTimeout_100ms,
	BufferTimeout_1000ms
};

enum TriggerType
{
	TriggerAuto = 0,
	TriggerExternal,
	TriggerPulser,
	TriggerTCM
};

struct BufferInfo
{
	uint32_t baseAddress;
	uint32_t capacity;
	unsigned pndCount;
	bool isRunActive;
	bool hasTimeout;
	BufferTimeout timeout;
};

typedef int DaqCredit;
typedef unsigned FemNum;
typedef uint16_t PedValue;
typedef uint16_t ThrValue;
typedef uint16_t OffsetValue;
typedef uint16_t AfterRegValue;
typedef uint16_t PatternLength;

class AGetAccessor
{
public:
	AGetAccessor(FeminosController& controller, AsicID asic);

	bool enableMarkerReset() const;
	bool resetLevel() const;
	bool readFromZero() const;
	bool testDigOut() const;
	bool modeReadAll() const;
	bool polarityPositive() const;
	bool readFpn() const;
	short icmVoltage() const;
	short dacThreshold() const;
	short triggerVeto() const;
	bool synchroDiscri() const;
	bool timeOverThreshold() const;
	short rangeTriggerWidth() const;
	short triggerWidth() const;
	bool shortRead() const;
	bool disableMultiplicityOut() const;
	bool autoresetBank() const;

private:
	FeminosController& ctrl_;
	AsicID asic_;
};

class AGetMutator
{
public:
	AGetMutator(FeminosController& controller, AsicRange asics);

	AGetMutator& setGain(ChannelRange channels, short gain);
	AGetMutator& setShapingTime(ChannelRange channels, short shapingTime);
	AGetMutator& setThreshold(ChannelRange channels, short threshold);
	AGetMutator& setInhibit(ChannelRange channels, bool inhibit);

	AGetMutator& setEnableMarkerReset(bool enable);
	AGetMutator& setResetLevel(bool level);
	AGetMutator& setReadFromZero(bool readFromZero);
	AGetMutator& setTestDigOut(bool testDigOut);
	AGetMutator& setModeReadAll(bool readAll);
	AGetMutator& setPolarityPositive(bool positive);
	AGetMutator& setReadFpn(bool readFpn);
	AGetMutator& setIcmVoltage(short icmVoltage);
	AGetMutator& setDacThreshold(short dacThreshold);
	AGetMutator& setTriggerVeto(short triggerVeto);
	AGetMutator& setSynchroDiscri(bool synchroDiscri);
	AGetMutator& setTimeOverThreshold(bool timeOverThreshold);
	AGetMutator& setRangeTriggerWidth(bool rangeTriggerWidth);
	AGetMutator& setTriggerWidth(short triggerWidth);
	AGetMutator& setShortRead(bool shortRead);
	AGetMutator& setDisableMultiplicityOut(bool disable);
	AGetMutator& setAutoresetBank(bool autoreset);
private:
	FeminosController& ctrl_;
	AsicRange asics_;
};

class FeminosController : public BoardController
{
public:
	FeminosController(const utl::net::SocketAddress& endpoint);
	~FeminosController();

	void setAsicType(const std::string& asicName);
	FemNum getFeminosNumber();

	AfterRegValue after_read(AsicID asic, RegisterID reg);
	void after_write(AsicRange asic, RegisterID reg, AfterRegValue value);
	template<class Reg>	Reg after_read(AsicID asic);
	template<class Reg>	void after_write(AsicRange asic, Reg value);

	AGetAccessor aget_read(AsicID asic);
	AGetMutator aget_write(AsicRange asics);

	void testdata_fillPattern(TestPattern pattern, PatternLength length);

	PedValue pedestal_get(AsicID asic, ChannelID channel);
	void pedestal_set(AsicRange asic, ChannelRange channel, PedValue value);
	void pedestal_list(ListResult& result, AsicRange asic);

	ThrValue threshold_get(AsicID asic, ChannelID channel);
	void threshold_set(AsicRange asic, ChannelRange channel, ThrValue value);
	void threshold_list(ListResult& result, AsicRange asic);

	DaqTarget daq_getTarget();
	void daq_setTarget(DaqTarget target);
	void daq_getCredits(DaqCredit& credits, DaqCreditUnit& unit);
	void daq_addCredits(DaqCredit credits, DaqCreditUnit unit, bool resetSeq = false);
	void daq_clearCredits();
	DaqLossPolicy daq_getLossPolicy();
	void daq_setLossPolicy(DaqLossPolicy policy);
	uint16_t daq_getCreditWaitTime();
	void daq_setCreditWaitTime(uint16_t waitTime);

	void buffer_getInfo(BufferInfo& info);
	void buffer_setActive(bool isActive);
	bool buffer_isTimeoutEnabled();
	void buffer_enableTimeout(bool enabled);
	BufferTimeout buffer_getTimeout();
	void buffer_setTimeout(BufferTimeout timeout);
	void buffer_reset();
	void buffer_sendPendingData();

	void tcmStats_clear();

	void hbusy_clear();
	void hbusy_getHisto(DeadtimeHistoResult& result);

	void hped_clear(AsicRange asic, ChannelRange channel);
	void hped_getSummary(PedestalSummaryResult& result, AsicID asic, ChannelRange channel);
	void hped_getStats(PedestalStatsResult& result, AsicID asic, ChannelID channel);
	void hped_getHisto(PedestalHistoResult& result, AsicID asic, ChannelID channel);
	void hped_setOffset(AsicRange asic, ChannelRange channel, OffsetValue offset);
	void hped_setThreshold(AsicRange asic, ChannelRange channel, ThrValue threshold, float sigma);
	void hped_center(AsicRange asic, ChannelRange channel, PedValue value);

	void shisto_clear(AsicRange asic, ChannelRange channel);
	void shisto_getHisto(SHistoResult& result, AsicID asic, ChannelID channel);
	void shisto_setThreshold(unsigned threshold);

	void hhit_clear(AsicRange asic);
	void hhit_getHisto(HitCountHistoResult& result, AsicID asic);

private:
	static BufferTimeout buffer_getTimeout(FeminosCommand& cmd, const char* text);
	static bool buffer_isTimeoutEnabled(FeminosCommand& cmd, const char* text);
};

// =============================================================================
// minos::fem::FeminosController inline methods body
// -----------------------------------------------------------------------------

template<class Reg>
Reg FeminosController::after_read(AsicID asic)
{
	Reg reg = after_read(asic, Reg::ID);
	return reg;
}

template<class Reg>
void FeminosController::after_write(AsicRange asic, Reg value)
{
	after_write(asic, Reg::ID, value.value);
}

// =============================================================================
// Registers definition
// -----------------------------------------------------------------------------

struct Reg0
{
	static const int ID = 0;

	Reg0() {}
	Reg0(uint32_t r) : value(r) {}

	union
	{
		uint32_t value;
		struct
		{
			uint32_t sca_cnt         : 10;
			uint32_t mode_after      : 1;
			uint32_t rst_chan_cnt    : 1;
			uint32_t force_on_all    : 1;
			uint32_t zero_suppress   : 1;
			uint32_t ped_subtract    : 1;
			uint32_t emit_ch_hit_cnt : 1;
			uint32_t keep_rst_chan   : 1;
			uint32_t skip_rst_until  : 2;
			uint32_t modify_hit_reg  : 1;
			uint32_t sca_enable      : 1;
			uint32_t sca_auto_start  : 1;
			uint32_t sca_start       : 1;
			uint32_t sca_stop        : 1;
			uint32_t clr_tstamp      : 1;
			uint32_t clr_evcnt       : 1;
			uint32_t bios_B          : 1;
			uint32_t card_idx        : 5;
		};
	};
};

struct Reg1
{
	static const int ID = 1;

	Reg1() {}
	Reg1(uint32_t r) : value(r) {}

	union
	{
		uint32_t value;
		struct
		{
			uint32_t private1   : 12;
			uint32_t pul_load   : 1;
			uint32_t private2   : 4;
			uint32_t keep_fco   : 1;
			uint32_t fec_enable : 1;
			uint32_t power_inv  : 1;
			uint32_t asic_mask  : 4;
			uint32_t private3   : 9;
		};
	};
};

struct Reg2
{
	static const int ID = 2;

	Reg2() {}
	Reg2(uint32_t r) : value(r) {}

	union
	{
		uint32_t value;
		struct
		{
			uint32_t trig_delay       : 16;
			uint32_t trig_rate        : 7;
			uint32_t trig_rate_range  : 2;
			uint32_t event_limit      : 3;
			uint32_t enable_auto_trig : 1;
			uint32_t enable_ext_trig  : 1;
			uint32_t enable_mult_trig : 1;
			uint32_t enable_tcm_trig  : 1;
		};
	};
};

struct Reg3
{
	static const int ID = 3;
	Reg3() {}
	Reg3(uint32_t r) : value(r) {}

	union
	{
		uint32_t value;
		struct
		{
			uint32_t pul_delay : 12;
			uint32_t reserved  : 20;
		};
	};
};

struct Reg5
{
	static const int ID = 5;

	Reg5() {}
	Reg5(uint32_t r) : value(r) {}

	union
	{
		uint32_t value;
		struct
		{
			uint32_t zs_pre           : 4;
			uint32_t zs_post          : 5;
			uint32_t test_enable      : 1;
			uint32_t test_mode        : 1;
			uint32_t test_zbt         : 1;
			uint32_t gen_use_limit    : 1;
			uint32_t emit_lst_cell_rd : 1;
			uint32_t eof_on_eoe       : 1;
			uint32_t emit_empty_ch    : 1;
			uint32_t sca_wckdiv       : 6;
			uint32_t polarity         : 1;
			uint32_t private1         : 9;
		};
	};
};

struct Reg6
{
	static const int ID = 6;

	Reg6() {}
	Reg6(uint32_t r) : value(r) {}

	union
	{
		uint32_t value;
		struct
		{
			uint32_t rx_cnt   : 8;
			uint32_t rx_perr  : 8;
			uint32_t tx_cnt   : 8;
			uint32_t reserved : 7;
			uint32_t cnt_clr  : 1;
		};
	};
};

struct AfterCfgReg1
{
	static const int ID = 1;

	AfterCfgReg1() {}
	AfterCfgReg1(uint16_t r) : value(r) {}
	union
	{
		uint16_t value;
		struct
		{
			uint16_t lcsa : 1;
			uint16_t gain : 2;
			uint16_t time : 4;
			uint16_t test : 2;
			uint16_t unused : 4;
			uint16_t power_down_write : 1;
			uint16_t power_down_read  : 1;
			uint16_t alternate_power  : 1;
		};
	};
};

struct AfterCfgReg2
{
	static const int ID = 2;

	AfterCfgReg2() {}
	AfterCfgReg2(uint16_t r) : value(r) {}
	union
	{
		uint16_t value;
		struct
		{
			uint16_t debug       : 2;
			uint16_t unused1     : 1;
			uint16_t read_from_0 : 1;
			uint16_t test_digout : 1;
			uint16_t unused2     : 1;
			uint16_t reset_marker_enabled : 1;
			uint16_t reset_marker_level   : 1;
		};
	};
};

} /* namespace fem */
} /* namespace minos */

#endif /* minos_fem_FeminosController_h_INCLUDED */
