/* =============================================================================
 * Frame.cpp
 * -----------------------------------------------------------------------------
 * class minos::fem::Frame
 * Created on: 13 déc. 2012 at Irfu/Sedi/Lilas, CEA Saclay, F-91191, France
 * Author: fchateau
 * -----------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * -----------------------------------------------------------------------------
 * Contributors: fchateau
 * -----------------------------------------------------------------------------
 * This software is part of the DaqMinos project.
 * -----------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding
 * by the rules of distribution of free software. You can use, modify and/or
 * redistribute the software under the terms of the CeCILL license as circulated
 * by CEA, CNRS and INRIA at the following URL: "http://www.cecill.info".
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability. In this
 * respect, the user's attention is drawn to the risks associated with loading,
 * using, modifying and/or developing or reproducing the software by the user in
 * light of its specific status of free software, that may mean that it is
 * complicated to manipulate, and that also therefore means that it is reserved
 * for developers and experienced professionals having in-depth computer
 * knowledge. Users are therefore encouraged to load and test the software's
 * suitability as regards their requirements in conditions enabling the security
 * of their systems and/or data to be ensured and, more generally, to use and
 * operate it in the same conditions as regards security.
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * -----------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for
 * commercial purposes. For this you will need to negotiate a specific contract
 * with a legal representative of CEA.
 * =============================================================================
 */

#include "Frame.h"
#include <utl/io/InplaceParser.ipp>
#include <utl/Logging.h>

namespace utl
{
namespace io
{
using namespace minos::fem;

// =============================================================================
// specializations of utl::io::InplaceParser methods for Minos frames
// -----------------------------------------------------------------------------

template<>
bool InplaceParser::parse_1(PedThrList& dest)
{
	return
	   parse_1(dest.header)
	&& parse_N(dest.values, dest.header->listLength());
}

template<>
bool InplaceParser::parse_1(Event& event)
{
	return
	   parse_0_1(event.header)
	&& parse_0_N(event.hitInfo)
	&& parse_0_N(event.lastCell)
	&& parse_1_N(event.hits)
	&& parse_0_1(event.epilogue);
}

template<>
bool InplaceParser::parse_1(ChannelHit& hit)
{
	ValueRef<ChannelHitEpilogue> epilogue;
	return
       parse_1(hit.channelHitIndex)
	&&
	(
		   parse_1_N(hit.waveform)
		|| parse_1_N(hit.waveformsOverThreshold)
	)
	&& parse_0_1(epilogue);
}

template<>
bool InplaceParser::parse_1(PartialWaveform& wf)
{
	return
	   parse_1  (wf.timeBinIndex)
	&& parse_1_N(wf.samples);
}

template<>
bool InplaceParser::parse_1(ConfigFrame& frame)
{
	InplaceParser buf = *this;
	ValueRef<Preamble> preamble;
	ValueRef<FrameEnd> frameEnd;
	return
	   buf.parse_1(preamble)
	&& buf.parse_1(frame.frameStart_)
	&& buf.parse_1(frame.msgInfo_)
	&& buf.parse_N(frame.message_, frame.msgInfo_->totalLength())
	&& buf.parse_0_1(frameEnd)
	&& buf.commit(*this);
}

template<>
bool InplaceParser::parse_1(MonitFrame& frame)
{
	InplaceParser buf = *this;
	ValueRef<Preamble> preamble;
	ValueRef<FrameEnd> frameEnd;
	return
	   buf.parse_1(preamble)
	&& buf.parse_1(frame.frameStart_)
	&&
	(
		buf.parse_1_N(frame.pedestalSummary_)
		||
		(
			buf.parse_1(frame.histoChannel_)
			&&
			(
				   buf.parse_1  (frame.pedestalStats_)
				|| buf.parse_1  (frame.sHisto_)
				|| buf.parse_0_N(frame.pedestalHisto_)
			)
		)
		||
		(
			   buf.parse_1  (frame.deadtimeStats_)
			&& buf.parse_0_N(frame.deadtimeHisto_)
		)
		|| buf.parse_1_N(frame.lists_)
		|| buf.parse_1(frame.hitCountHisto_)
		|| buf.parse_1(frame.cmdStats_)
	)
	&& buf.parse_0_1(frameEnd)
	&& buf.commit(*this);
}

template<>
bool InplaceParser::parse_1(DataFrame& frame)
{
	InplaceParser buf = *this;
	ValueRef<FrameEnd> frameEnd;
	return
	   buf.parse_1  (frame.dataPreamble_)
	&& buf.parse_1  (frame.frameStart_)
	&& buf.parse_1_N(frame.events_)
	&& buf.parse_1  (frameEnd)
	&& buf.commit(*this);
}

} /* namespace io */
} /* namespace utl */

// =============================================================================
// minos::fem::Frame class methods body
// -----------------------------------------------------------------------------

namespace minos
{
namespace fem
{

Frame::Frame()
	: frame_(reinterpret_cast<MinosFrame*>(frameData_))
{
	new(frame_) MinosFrame;
}

Frame::~Frame()
{
	frame().~MinosFrame();
}

template<class F>
F& Frame::createFrame()
{
	frame().~MinosFrame();
	return *new(frame_) F;
}

bool Frame::parse(const ::utl::io::ConstBuffer& cbuf)
{
	::utl::io::InplaceParser p = cbuf;

	ValueRef<Preamble> preamble;
	bool result =
		(
			   p.parse_1(createFrame<DataFrame>())
			|| p.parse_1(createFrame<ConfigFrame>())
			|| p.parse_1(createFrame<MonitFrame>())
		)
		&& p.isEndOfBuffer();
	return result;
}

} /* namespace fem */
} /* namespace minos */
