/* =============================================================================
 * FrameTypes.h
 * -----------------------------------------------------------------------------
 * Created on: 21 déc. 2012 at Irfu/Sedi/Lilas, CEA Saclay, F-91191, France
 * Author: fchateau
 * -----------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * -----------------------------------------------------------------------------
 * Contributors: fchateau
 * -----------------------------------------------------------------------------
 * This software is part of the DaqMinos project.
 * -----------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding
 * by the rules of distribution of free software. You can use, modify and/or
 * redistribute the software under the terms of the CeCILL license as circulated
 * by CEA, CNRS and INRIA at the following URL: "http://www.cecill.info".
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability. In this
 * respect, the user's attention is drawn to the risks associated with loading,
 * using, modifying and/or developing or reproducing the software by the user in
 * light of its specific status of free software, that may mean that it is
 * complicated to manipulate, and that also therefore means that it is reserved
 * for developers and experienced professionals having in-depth computer
 * knowledge. Users are therefore encouraged to load and test the software's
 * suitability as regards their requirements in conditions enabling the security
 * of their systems and/or data to be ensured and, more generally, to use and
 * operate it in the same conditions as regards security.
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * -----------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for
 * commercial purposes. For this you will need to negotiate a specific contract
 * with a legal representative of CEA.
 * =============================================================================
 */

#ifndef minos_fem_FrameTypes_h_INCLUDED
#define minos_fem_FrameTypes_h_INCLUDED

#include <utl/io/InplaceParser.hpp>
#include <boost/utility/enable_if.hpp>
#include <boost/static_assert.hpp>
#include <boost/cstdint.hpp>
#include <vector>

#define SHIFT_PFX(pfx) (pfx >> (16-PREFIX_WIDTH))

namespace minos
{
namespace fem
{
using ::utl::io::ArrayRef;
using ::utl::io::ValueRef;

enum AsicType
{
	AfterAsic = 0,
	AGetAsic
};

enum ChannelCount
{
	AfterChannelCount = 79,
	AgetChannelCount = 72,
	AgetConfigurableChannelCount = 64
};

enum BoardCount
{
	FeminosMaxCount = 32
};

enum TpcDimensions
{
	TpcRingCount = 18,
	TpcRayCount = 256
};

enum DataPrefix
{
	PFX_CARD_CHIP_CHAN_HIT_IX  = 0x3 << 14,
	PFX_CARD_CHIP_CHAN_HIT_CNT = 0x2 << 14,
	PFX_CARD_CHIP_CHAN_HISTO   = 0x1 << 14,
	PFX_ADC_SAMPLE             = 0x3 << 12,
	PFX_LAT_HISTO_BIN          = 0x2 << 12,
	PFX_CHIP_LAST_CELL_READ    = 0x1 << 12,
	PFX_TIME_BIN_IX            = 0x7 << 9,
	PFX_HISTO_BIN_IX           = 0x6 << 9,
	PFX_PEDTHR_LIST            = 0x5 << 9,
	PFX_START_OF_DFRAME        = 0x4 << 9,
	PFX_START_OF_MFRAME        = 0x3 << 9,
	PFX_START_OF_CFRAME        = 0x2 << 9,
	PFX_ASCII_MSG_LEN          = 0x1 << 8,
	PFX_START_OF_EVENT         = 0xF << 4,
	PFX_END_OF_EVENT           = 0xE << 4,
	PFX_CH_HIT_CNT_HISTO       = 0x1F << 2,
	PFX_END_OF_FRAME           = 0xF,
	PFX_DEADTIME_HSTAT_BINS    = 0xE,
	PFX_PEDESTAL_HSTAT         = 0xD,
	PFX_PEDESTAL_H_MD          = 0xC,
	PFX_SHISTO_BINS            = 0xB,
	PFX_CMD_STATISTICS         = 0xA,
	PFX_END_OF_BUILT_EVENT     = 0x8,
	PFX_SOBE_SIZE              = 0x6,
	PFX_NULL_CONTENT           = 0x0
};

#pragma pack(push, 2)
struct Preamble
{
	static const unsigned PREFIX_WIDTH = 16;
	static const unsigned PREFIX = SHIFT_PFX(PFX_NULL_CONTENT);

	uint16_t prefix;
};
struct DataPreamble
{
	struct
	{
		uint16_t sendCount : 8;
		uint16_t recvCount : 8;
	};
};
struct FrameStart
{
	static const unsigned PREFIX_WIDTH = 7;
	struct
	{
		uint16_t femIndex : 5; // lower bits
		uint16_t version : 4;
		uint16_t prefix : PREFIX_WIDTH;
	};
};
struct FrameEnd
{
	static const unsigned PREFIX_WIDTH = 16;
	static const unsigned PREFIX = SHIFT_PFX(PFX_END_OF_FRAME);

	uint16_t prefix;
};
struct ConfigFrameStart : FrameStart
{
	static const unsigned PREFIX = SHIFT_PFX(PFX_START_OF_CFRAME);

	int16_t errorCode;
};
struct DataFrameStart : FrameStart
{
	static const unsigned PREFIX = SHIFT_PFX(PFX_START_OF_DFRAME);

	uint16_t payloadSize;
};
struct MonitFrameStart : FrameStart
{
	static const unsigned PREFIX = SHIFT_PFX(PFX_START_OF_MFRAME);

	uint16_t payloadSize;
};
struct AsciiMessage
{
	static const unsigned PREFIX_WIDTH = 8;
	static const unsigned PREFIX = SHIFT_PFX(PFX_ASCII_MSG_LEN);

	uint16_t messageLength : 8;
	uint16_t prefix : PREFIX_WIDTH;

	unsigned totalLength() const;
};

struct HistoChannel
{
	static const unsigned PREFIX_WIDTH = 2;
	static const unsigned PREFIX = SHIFT_PFX(PFX_CARD_CHIP_CHAN_HISTO);
	struct
	{
		uint16_t channel : 7;
		uint16_t chip : 2;
		uint16_t card : 5;
		uint16_t prefix : PREFIX_WIDTH;
	};
	HistoChannel();
};
struct PedestalHistoBin
{
	static const unsigned PREFIX_WIDTH = 7;
	static const unsigned PREFIX = SHIFT_PFX(PFX_HISTO_BIN_IX);
	struct
	{
		uint16_t binIndex : 9;
		uint16_t prefix : PREFIX_WIDTH;
	};
	uint16_t binValue;
};
struct PedestalStats
{
	static const unsigned PREFIX_WIDTH = 16;
	static const unsigned PREFIX = SHIFT_PFX(PFX_PEDESTAL_HSTAT);
	uint16_t prefix;
	uint32_t minBin;
	uint32_t maxBin;
	uint32_t binWidth;
	uint32_t binCount;
	uint32_t minNonNullBin;
	uint32_t maxNonNullBin;
	uint32_t mean;
	uint32_t stdev;
	uint32_t entryCount;
};
struct PedestalSummary
{
	static const unsigned PREFIX_WIDTH = 16;
	static const unsigned PREFIX = SHIFT_PFX(PFX_PEDESTAL_H_MD);

	HistoChannel channel;
	uint16_t prefix;
	uint32_t mean;
	uint32_t stdev;

	PedestalSummary();
	float pedestalMean() const { return static_cast<float>(mean) / 100; }
	float pedestalStdev() const { return static_cast<float>(stdev) / 100; }
};
struct HistoStats
{
	static const unsigned PREFIX_WIDTH = 16;
	uint16_t prefix;
	uint16_t padding;
	uint32_t minBin;
	uint32_t maxBin;
	uint32_t binWidth;
	uint32_t binCount;
	uint32_t minNonNullBin;
	uint32_t maxNonNullBin;
	uint32_t mean;
	uint32_t stdev;
	uint32_t entryCount;
};
struct DeadtimeStats : HistoStats
{
	static const unsigned PREFIX = SHIFT_PFX(PFX_DEADTIME_HSTAT_BINS);
};
struct HitCountStats : HistoStats
{
	static const unsigned PREFIX = SHIFT_PFX(PFX_CH_HIT_CNT_HISTO);
};
struct HitCountHisto : HitCountStats
{
	static const unsigned BIN_COUNT = 80;
	uint32_t bin[BIN_COUNT];
};
struct DeadtimeHistoBin
{
	static const unsigned PREFIX_WIDTH = 4;
	static const unsigned PREFIX = SHIFT_PFX(PFX_LAT_HISTO_BIN);
	struct
	{
		uint16_t binIndex : 12;
		uint16_t prefix : PREFIX_WIDTH;
	};
	uint32_t binValue;
};
struct CommandStats
{
	static const unsigned PREFIX_WIDTH = 16;
	static const unsigned PREFIX = SHIFT_PFX(PFX_CMD_STATISTICS);
	uint16_t prefix;
	uint32_t cmdCount;
	uint32_t daqReqCount;
	uint32_t daqTimeoutCount;
	uint32_t cmdErrCount;
	uint32_t cmdReplyCount;
	uint32_t daqReplyCount;
};
struct SHisto
{
	static const unsigned PREFIX_WIDTH = 16;
	static const unsigned PREFIX = SHIFT_PFX(PFX_SHISTO_BINS);
	static const unsigned BIN_COUNT = 16;
	uint16_t prefix;
	uint16_t bin[BIN_COUNT];
};
struct ListHeader
{
	static const unsigned PREFIX_WIDTH = 7;
	static const unsigned PREFIX = SHIFT_PFX(PFX_PEDTHR_LIST);
	uint16_t isThreshold : 1;
	uint16_t isAfter : 1;
	uint16_t asic : 2;
	uint16_t card : 5;
	uint16_t prefix : PREFIX_WIDTH;

	unsigned listLength() const;
};

typedef uint16_t ListValue;

struct EventHeader
{
	static const unsigned PREFIX_WIDTH = 12;
	static const unsigned PREFIX = SHIFT_PFX(PFX_START_OF_EVENT);
	struct
	{
		uint64_t type : 4;
		uint64_t prefix : PREFIX_WIDTH;
		uint64_t timestamp : 48;
	};
	uint32_t eventId;
};
struct HitChannelCount
{
	static const unsigned PREFIX_WIDTH = 2;
	static const unsigned PREFIX = SHIFT_PFX(PFX_CARD_CHIP_CHAN_HIT_CNT);
	struct
	{
		uint16_t hitChannelCount : 7;
		uint16_t chip : 2;
		uint16_t card : 5;
		uint16_t prefix : PREFIX_WIDTH;
	};
};
struct LastCellRead
{
	static const unsigned PREFIX_WIDTH = 4;
	static const unsigned PREFIX = SHIFT_PFX(PFX_CHIP_LAST_CELL_READ);
	uint16_t readPointer : 10;
	uint16_t asic : 2;
	uint16_t prefix : PREFIX_WIDTH;
};
struct EventEpilogue
{
	static const unsigned PREFIX_WIDTH = 12;
	static const unsigned PREFIX = SHIFT_PFX(PFX_END_OF_EVENT);
	struct
	{
		uint16_t eventSizeMsb : 4;
		uint16_t prefix : PREFIX_WIDTH;
	};
	uint16_t eventSizeLsw;
	uint32_t eventSize() const;
	void setEventSize(uint32_t size);
};
struct ChannelHitIndex
{
	static const unsigned PREFIX_WIDTH = 2;
	static const unsigned PREFIX = SHIFT_PFX(PFX_CARD_CHIP_CHAN_HIT_IX);
	struct
	{
		uint16_t channel : 7;
		uint16_t chip    : 2;
		uint16_t card    : 5;
		uint16_t prefix  : PREFIX_WIDTH;
	};
};
struct TimeBinIndex
{
	static const unsigned PREFIX_WIDTH = 7;
	static const unsigned PREFIX = SHIFT_PFX(PFX_TIME_BIN_IX);
	struct
	{
		uint16_t index  : 9;
		uint16_t prefix : PREFIX_WIDTH;
	};
};
struct AdcSample
{
	static const unsigned PREFIX_WIDTH = 4;
	static const unsigned PREFIX = SHIFT_PFX(PFX_ADC_SAMPLE);
	struct
	{
		uint16_t value  : 12;
		uint16_t prefix : PREFIX_WIDTH;
	};
};
struct ChannelHitEpilogue
{
	static const unsigned PREFIX_WIDTH = 16;
	static const unsigned PREFIX = SHIFT_PFX(PFX_NULL_CONTENT);

	uint16_t prefix;
};

struct BuiltEventHeader
{
	static const unsigned PREFIX_WIDTH = 16;
	static const unsigned PREFIX = SHIFT_PFX(PFX_SOBE_SIZE);
	uint16_t prefix;
	uint32_t size;
};
struct BuiltEventFooter
{
	static const unsigned PREFIX_WIDTH = 16;
	static const unsigned PREFIX = SHIFT_PFX(PFX_END_OF_BUILT_EVENT);
	uint16_t prefix;
};
#pragma pack(pop)

BOOST_STATIC_ASSERT(sizeof(HistoChannel) == 2);
BOOST_STATIC_ASSERT(sizeof(PedestalStats) == 38);
BOOST_STATIC_ASSERT(sizeof(PedestalSummary) == 12);
BOOST_STATIC_ASSERT(sizeof(DeadtimeHistoBin) == 6);
BOOST_STATIC_ASSERT(sizeof(CommandStats) == 26);


struct PartialWaveform
{
	ValueRef<TimeBinIndex> timeBinIndex; ///< index of the first time bin over the threshold
	ArrayRef<AdcSample>    samples;
};

struct ChannelHit
{
	ValueRef<ChannelHitIndex>    channelHitIndex; ///< index of the channel being hit
	ArrayRef<AdcSample>          waveform; ///< if zero-suppression disabled, single complete waveform
	std::vector<PartialWaveform> waveformsOverThreshold; ///< if zero-suppression enabled, we can have multiple waveforms
};

struct Event
{
	ValueRef<EventHeader>     header;
	ArrayRef<HitChannelCount> hitInfo;
	ArrayRef<LastCellRead>    lastCell;
	std::vector<ChannelHit>   hits;
	ValueRef<EventEpilogue>   epilogue;
};

struct PedThrList
{
	ValueRef<ListHeader> header;
	ArrayRef<ListValue>  values;
};

} /* namespace fem */
} /* namespace minos */

namespace utl
{
namespace io
{

template<class T>
struct has_prefix
{
	typedef char Yes[2];
	typedef char No[1];

	template<class C, C> class Test;
	template<class U> static Yes& check(Test<const unsigned, U::PREFIX>*);
	template<class U> static No& check(...);

	const static bool value = sizeof(check<T>(0)) == sizeof(Yes);
};


template<class DataField>
bool isDataField(
	typename boost::enable_if_c<
		has_prefix<DataField>::value,
		const InplaceParser&
	>::type buf,
	const DataField* f)
{
	return
	   buf.tryBind(f)
	&& f->prefix == DataField::PREFIX;
}

inline bool isDataField(const InplaceParser& buf, const minos::fem::PedestalSummary* f)
{
	return
		buf.tryBind(f)
	&&  f->channel.prefix == minos::fem::HistoChannel::PREFIX
	&&  f->prefix == minos::fem::PedestalSummary::PREFIX;
}

}
}

#undef SHIFT_PFX

#endif /* minos_fem_FrameTypes_h_INCLUDED */
