/**
 * @file TriggerRate.cpp
 * @date 8 août 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the DaqMinos software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "TriggerRate.h"
#include <cmath>

namespace minos
{
namespace fem
{

TriggerRate TriggerRate::createFromFreq(double freq)
{
	return TriggerRate(freq);
}

TriggerRate::TriggerRate()
	: range_(0), rate_(0)
{
}

/**
 * @param freq trigger frequency in Hz [0,1; 127k], 1-3 significant digits
 */
TriggerRate::TriggerRate(double freq)
	: range_(computeTriggerRange(freq)),
	  rate_(computeTriggerRate(freq, range_))
{
}

TriggerRate::TriggerRate(int rate, int range)
	: range_(range), rate_(rate)
{
}

double TriggerRate::computeFreq() const
{
	return rate_ * computeRangeBaseFreq(range_);
}

double TriggerRate::computeRangeBaseFreq(int range)
{
	return pow(10.0, (range == 0) ? -1 : range);
}

int TriggerRate::computeTriggerRange(double freq)
{
	if(freq < 0.1)
	{
		return 0;
	}
	int logFreq = static_cast<int>(floor(log10(freq / 128.0)));
	if(logFreq < 2)
	{
		return (logFreq + 3) / 2;
	}
	else
	{
		return 3;
	}
}

int TriggerRate::computeTriggerRate(double freq, int range)
{
	int rate = static_cast<int>(0.5 + freq / computeRangeBaseFreq(range));
	if(rate < 0)
	{
		return 0;
	}
	else if(rate > 127)
	{
		return 127;
	}
	else
	{
		return rate;
	}
}

} /* namespace fem */
} /* namespace minos */
