/**
 * @file PedestalsWidget.cpp
 * @date 8 oct. 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the DaqMinos software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "PedestalsWidget.h"
#include <minos/fem/DetectorMapping.h>
#include <minos/fem/Frame.h>
#include <minos/gui/PedestalsWidget.ui.h>
#include <mdcs/ctrl/mon/MetricsMonitor.h>
#include <qwt_plot_spectrogram.h>
#include <qwt_plot_layout.h>
#include <qwt_scale_widget.h>
#include <qwt_color_map.h>
#include <qwt_plot_zoomer.h>
#include <qwt_plot_layout.h>
#include <QTextStream>
#include <QByteArray>

namespace minos
{
namespace gui
{
using namespace ::utl::io;
using namespace mdcs::ctrl::mon;
using namespace fem;

// =============================================================================
// minos::gui::PedestalsWidget::SpectrogramData private classes definition
// -----------------------------------------------------------------------------

class PedestalsWidget::SpectrogramData : public QwtRasterData
{
public:
    SpectrogramData()
		: QwtRasterData(QwtDoubleRect(0, 0, TpcRayCount, TpcRingCount))
    {
    }
    QwtRasterData *copy() const
    {
        return new SpectrogramData(*this);
    }
    QwtDoubleInterval range() const
    {
        return QwtDoubleInterval(0.0, 40.0);
    }
	double value(double x, double y) const
	{
		ProjectiveTpcPad pad(y, x);
		if(pad.isValid())
		{
			return static_cast<double>(pedestal_[pad.ring][pad.ray].pedestalStdev());
		}
		else
		{
			return 0.0;
		}
	}
	bool values(double x, double y, ProjectiveTpcPad& pad, PedestalSummary& pedestal) const
	{
		pad.ring = y;
		pad.ray = x;
		if(pad.isValid())
		{
			pedestal = pedestal_[pad.ring][pad.ray];
			return true;
		}
		else
		{
			return false;
		}
	}
	void setValues(const ProjectiveTpcPad& pad, const PedestalSummary& pedestal)
	{
		if(pad.isValid())
		{
			pedestal_[pad.ring][pad.ray] = pedestal;
		}
		else
		{
			LOG_WARN() << "SpectrogramData::setValue: invalid ring or ray " << pad.ring << ',' << pad.ray;
		}
	}

private:
	PedestalSummary pedestal_[TpcRingCount][TpcRayCount];
};

// =============================================================================
// minos::gui::PedestalsWidget::Zoomer private classes definition
// -----------------------------------------------------------------------------

class PedestalsWidget::Zoomer : public QwtPlotZoomer
{
public:
	Zoomer(QwtPlotCanvas* canvas)
    	: QwtPlotZoomer(canvas),
    	  font("monospace")
    {
    }
    QwtText trackerText(const QwtDoublePoint &pos) const
    {
    	QwtText qwtText;
    	const SpectrogramData& data = dynamic_cast<const SpectrogramData&>(
    			dynamic_cast<const QwtPlotSpectrogram&>(*plot()->itemList()[0]).data()
    		);
        ProjectiveTpcPad pad;
        PedestalSummary ped;
        if(data.values(pos.x(), pos.y(), pad, ped))
        {
            QString text;
            QTextStream stream(&text, QIODevice::WriteOnly);
        	stream
        		<< qSetRealNumberPrecision(2) << forcepoint << fixed
        		<< left << qSetFieldWidth(7)
        		<< "mean" << ped.pedestalMean() << "stdev" << ped.pedestalStdev()
        		<< right << endl << left
        		<< "ring" << pad.ring << "ray" << pad.ray;

        	if(ped.prefix != 0)
        	{
        		stream << right << endl << left
        			<< "card" << ped.channel.card
        			<< "chip" << ped.channel.chip
        			<< "chan" << ped.channel.channel;
        	}
        	qwtText.setText(text);
        	qwtText.setRenderFlags(Qt::AlignLeft | Qt::AlignTop);
        	qwtText.setFont(font);
        }
        return qwtText;
    }
private:
    QFont font;
};

// =============================================================================
// minos::gui::PedestalsWidget class methods body
// -----------------------------------------------------------------------------

PedestalsWidget::PedestalsWidget(QWidget* parent)
	: Utils::Ui::DesignedWidget<Ui_PedestalsWidget, QWidget>(parent),
	  wiringSignal_(this, SLOT(updateWiring(qint64))),
	  pedestalSignal_(this, SLOT(updatePedestal(const QByteArray&))),
	  spectrogram_(new QwtPlotSpectrogram),
	  spectrogramData_(0)
{
	spectrogram_->attach(ui_->spectroPlot); // the QwtPlot takes ownership of the QwtPlotSpectrogram

	QwtLinearColorMap colorMap(Qt::black, Qt::red);
	colorMap.addColorStop(0.01, Qt::darkBlue);
	colorMap.addColorStop(0.3, Qt::cyan);
	colorMap.addColorStop(0.4, Qt::green);
	colorMap.addColorStop(0.5, Qt::yellow);

	spectrogram_->setColorMap(colorMap);
//	spectrogram_->setCachePolicy(QwtPlotRasterItem::PaintCache);
	spectrogram_->setDisplayMode(QwtPlotSpectrogram::ImageMode, true);
	spectrogram_->setDisplayMode(QwtPlotSpectrogram::ContourMode, false);
	spectrogram_->setData(SpectrogramData());

	// configure the color bar on the right axis
	const QwtDoubleInterval& zRange = spectrogram_->data().range();
	QwtScaleWidget* rightAxis = ui_->spectroPlot->axisWidget(QwtPlot::yRight);
	rightAxis->setTitle("Standard Deviation (Sigma)");
	rightAxis->setColorBarEnabled(true);
	rightAxis->setColorMap(zRange, colorMap);
	ui_->spectroPlot->setAxisScale(QwtPlot::yRight, zRange.minValue(), zRange.maxValue());
	ui_->spectroPlot->enableAxis(QwtPlot::yRight);

	// configure other axis
	ui_->spectroPlot->axisWidget(QwtPlot::yLeft)->setTitle("Ring");
	ui_->spectroPlot->axisWidget(QwtPlot::xBottom)->setTitle("Ray");

	ui_->spectroPlot->plotLayout()->setAlignCanvasToScales(true);

	// configure the zoomer
	const QColor whiteColor(Qt::white);
	QwtPlotZoomer* zoomer = new Zoomer(ui_->spectroPlot->canvas()); // canvas takes ownership...
    zoomer->setTrackerMode(QwtPlotZoomer::AlwaysOn);
    zoomer->setRubberBandPen(whiteColor);
    zoomer->setTrackerPen(whiteColor);

    spectrogramData_ = dynamic_cast<SpectrogramData*>(const_cast<QwtRasterData*>(&spectrogram_->data()));
}

PedestalsWidget::~PedestalsWidget()
{
}

void PedestalsWidget::setMonitor(MetricsMonitor& monitor)
{
	wiringMetric_   = monitor.subscribeMetric("Feminos.wiring", wiringSignal_);
	pedestalMetric_ = monitor.subscribeMetric("Feminos.pedestals", pedestalSignal_);
}

void PedestalsWidget::updateWiring(qint64 wiringData)
{
	mapping_.setWiring(fem::WiringInfo(static_cast<uint16_t>(wiringData)));
}

void PedestalsWidget::updatePedestal(const QByteArray& pedestalData)
{
	LOG_DEBUG() << "received pedestal";

	Frame frame;
	frame.parse(ConstBuffer(pedestalData.data(), pedestalData.size()));
	const ArrayRef<PedestalSummary>& pedestals = frame.frame<MonitFrame>().pedestalSummary();
	AsicType asicType = (pedestals.count() > AgetChannelCount) ? AfterAsic : AGetAsic;
	for(unsigned i = 0; i < pedestals.count(); ++i)
	{
		Channel chan;
		chan.card = pedestals[i].channel.card;
		chan.chip = pedestals[i].channel.chip;
		chan.channel = pedestals[i].channel.channel;
		ProjectiveTpcPad pad;
		if(mapping_.daqChannelToTpcPad(chan, pad, asicType))
		{
			spectrogramData_->setValues(pad, pedestals[i]);
		}
	}
	ui_->spectroPlot->replot();
}

} /* namespace gui */
} /* namespace minos */
