/**
 * @file DataThread.cpp
 * @date 31 mai 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the minos_babies software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "common.h"
extern "C" {
#include "babies.h"
#include "ridf.h"
}
#include <minos/fem/BuiltEvent.h>
#include <mdcs/utl/net/TcpAcceptor.h>
#include <utl/Logging.h>
#include ICEUTIL_HEADER(Thread.h)
#include <boost/scoped_array.hpp>
#include <memory>

using namespace mdcs::utl::net;
using namespace ::utl::net;

class DataThread : public IceUtil::Thread
{
public:
	DataThread();
	void run();

	void askStop();

private:
	bool shouldStop;
	std::auto_ptr<TcpAcceptor> acceptor;
	std::auto_ptr<TcpConnection> inbound;
};

struct minos_data_thread
{
	IceUtil::Handle<DataThread> thread;
	IceUtil::ThreadControl threadCtrl;
};

DataThread::DataThread()
	: shouldStop(false)
{
}

void DataThread::run()
{
	try
	{
//		if(!efrc.connect)
//		{
//			efr_connect();
//		}

		acceptor.reset(new TcpAcceptor(SocketAddress(IpAddress(127, 0, 0, 1), 1123)));
		LOG_DEBUG() << "waiting for inbound connection";
		do
		{
			try
			{
				inbound = acceptor->acceptConnection(2000);
			}
			catch(const Socket::TimeoutException& e)
			{
//				LOG_DEBUG() << "Accept timeout";
			}
		}
		while(!shouldStop and inbound.get() == 0);
		if(shouldStop)
		{
			return;
		}
		LOG_DEBUG() << "accepting connection";
		TcpConnection outbound(efsock);
		LOG_DEBUG() << "outbound connection established";

		uint32_t firstEventId = 0xFFFFFFFF;
		boost::scoped_array<uint8_t> dataBuffer;
		uint32_t bufferSize = 0;
		Milliseconds timeout = 2000;
		while(!shouldStop)
		{
			try
			{
				uint32_t frameSize = 0;
				inbound->doRcve(buffer(frameSize), timeout);
				if(frameSize > bufferSize)
				{
					dataBuffer.reset();
					dataBuffer.reset(new uint8_t[frameSize]);
				}
				inbound->doRcve(buffer(dataBuffer.get(), frameSize), timeout);

				minos::fem::BuiltEvent::UntilId event(buffer(dataBuffer.get(), frameSize));

				LOG_DEBUG() << "event size=" << event.size() << " id=" << event.eventId() << " ts=" << event.timestamp();

				firstEventId = std::min(firstEventId, event.eventId());

				uint32_t packetSize = frameSize + sizeof(RIDFHDSEG);
				RIDFHDSEG seghd = ridf_mkhd_seg(RIDF_LY2, RIDF_SEGMENT, packetSize/sizeof(uint16_t), efrc.efid, 0x00123535);
				packetSize += sizeof(RIDFHDEVTTS);
				RIDFHDEVTTS evthd = ridf_mkhd_evtts(RIDF_LY1, RIDF_EVENT_TS, packetSize/sizeof(uint16_t), efrc.efid, event.eventId() - firstEventId + 1, event.timestamp());
				packetSize += sizeof(RIDFHD);
				RIDFHD hd = ridf_mkhd(RIDF_LY0, RIDF_EF_BLOCK, packetSize/sizeof(uint16_t), efrc.efid);

				outbound.doSend(buffer(packetSize));
				outbound.doSend(buffer(hd));
				outbound.doSend(buffer(evthd));
				outbound.doSend(buffer(seghd));
				outbound.doSend(buffer(dataBuffer.get(), frameSize));
				++evtn;
			}
			catch(const Socket::TimeoutException& e)
			{
//				LOG_DEBUG() << "Recv timeout";
			}
		}
	}
	HANDLE_EXCEPTIONS
}

void DataThread::askStop()
{
	shouldStop = true;
}

extern "C"
{

void minos_data_thread_start(minos_data_thread** th)
{
	try
	{
		if(*th == 0)
		{
			std::auto_ptr<minos_data_thread> mth(new minos_data_thread);
			LOG_DEBUG() << "creating data thread";
			mth->thread = new DataThread;
			LOG_DEBUG() << "done (" << mth->thread.get() << "starting data thread";
			mth->threadCtrl = mth->thread->start();
			*th = mth.release();
		}
	}
	HANDLE_EXCEPTIONS
}

void minos_data_thread_stop(minos_data_thread** th)
{
	try
	{
		if(!th and !*th) throw std::runtime_error("invalid data thread!");
		LOG_DEBUG() << "ask data thread (" << (*th)->thread.get() << ") to stop";
		(*th)->thread->askStop();
		LOG_DEBUG() << "waiting for data thread to stop";
		(*th)->threadCtrl.join();
		LOG_DEBUG() << "destroying data thread";
		delete *th;
		*th = 0;
		LOG_DEBUG() << "data thread destroyed";
	}
	HANDLE_EXCEPTIONS
}

}
