/**
 * @file RunControlClient.cpp
 * @date 29 mai 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the minos_babies software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

extern "C" {
#include "RunControlClient.h"
}
#include "common.h"
#include <mdcs/acq/StateMachineFactory.h>
#include <mdcs/ctrl/IRunConfigQueue.h>
#include <mdcs/cfg/Constants.h>
#include <mdcs/utl/ConsoleLoggingBackend.h>
#include <dhsm/MasterEngine.h>
#include <dhsm/StateMachine.h>
#include <dhsm/IEngineServant.h>
#include <Ice/Communicator.h>
#include <Ice/Initialize.h>
#include <utl/Logging.h>
#include <memory>

extern "C"
{
using namespace dhsm;
using namespace mdcs::acq;
using namespace mdcs::ctrl;
using namespace mdcs::utl;
using namespace utl;

struct minos_connection
{
	Ice::CommunicatorPtr ic;
	MasterEngine engine;
	IRunConfigQueuePrx configQueue;
};

minos_connection* minos_create(int* argc, char** argv)
{
	try
	{
		std::auto_ptr<ConsoleLoggingBackend> logging(new ConsoleLoggingBackend);
		logging->setIdentity("B");
		BackendLogger::setBackend(LoggingBackendPtr(logging));

		LOG_INFO() << "creating minos_connection";
		std::auto_ptr<minos_connection> conn(new minos_connection);
		LOG_INFO() << "initializing communicator";
		conn->ic = Ice::initialize(*argc, argv);
		LOG_INFO() << "initializing state-machine engine";
		conn->engine.init(conn->ic);
		LOG_INFO() << "creating state machine";
		conn->engine.setStateMachine(StateMachineFactory::instance()["MDaqCtrlStateMachine"].create());
		LOG_INFO() << "starting state machine";
		conn->engine.start();
		LOG_INFO() << "minos_create ok";
		return conn.release();
	}
	HANDLE_EXCEPTIONS
	return 0;
}
minos_error minos_destroy(minos_connection** conn)
{
	try
	{
		if(!conn or !*conn) throw std::runtime_error("invalid minos connection!");
		delete *conn;
		*conn = 0;
		return minos_ok;
	}
	HANDLE_EXCEPTIONS
	return minos_failed;
}

minos_error minos_connect(minos_connection* conn)
{
	try
	{
		if(!conn) throw std::runtime_error("invalid minos connection!");
		if(conn->engine.servantCount() == 0)
		{
			LOG_INFO() << "connecting to run control";
			conn->engine.connectServant(mdcs::cfg::RUNCONTROL_CTRL_SERVANT);
		}
		if(!conn->configQueue)
		{
			LOG_INFO() << "connecting to config queue";
			conn->configQueue = IRunConfigQueuePrx::checkedCast(conn->ic->stringToProxy("ConfigQueue"));
		}
		LOG_INFO() << "Minos connection established";
		return minos_ok;
	}
	HANDLE_EXCEPTIONS
	LOG_WARN() << "Minos connection failed!";
	return minos_failed;
}

minos_error minos_disconnect(minos_connection* conn)
{
	try
	{
		if(!conn) throw std::runtime_error("invalid minos connection!");
		LOG_INFO() << "Minos disconnect:";
		LOG_INFO() << "- reseting DAQ workers";
		conn->engine.step(Event("reset"));
		LOG_INFO() << "- disconnecting run control from DAQ workers";
		conn->engine.step(Event("disconnect"));
		LOG_INFO() << "- disconnecting from run control";
		conn->engine.disconnectAllServants();
		LOG_INFO() << "- disconnecting from config Queue";
		conn->configQueue = 0;
		LOG_INFO() << "disconnection from Minos complete!";
		return minos_ok;
	}
	HANDLE_EXCEPTIONS
	LOG_WARN() << "disconnection from Minos failed!";
	return minos_failed;
}

minos_error minos_start(minos_connection* conn)
{
	try
	{
		try
		{
			if(!conn) throw std::runtime_error("invalid minos connection!");
			bool doReset = false;
			bool doDisconnect = false;

			if(conn->engine.currentState().label() != "READY")
			{
				LOG_INFO() << "Not in ready state: reconnecting and reconfiguring";
				doReset = true;
				doDisconnect = true;
			}
			else
			{
				if(conn->configQueue->isChanged("connect"))
				{
					LOG_INFO() << "Connection config changed: reconnecting and reconfiguring";
					doReset = true;
					doDisconnect = true;
				}
				else if(conn->configQueue->isChanged("configure"))
				{
					LOG_INFO() << "DAQ config changed: reconfiguring";
					doReset = true;
				}
			}
			if(doReset)
			{
				conn->engine.step(Event("reset"));
				if(doDisconnect)
				{
					conn->engine.step(Event("disconnect"));
					conn->engine.step(Event("connect"));
				}
				conn->engine.step(Event("configure"));
			}

			LOG_INFO() << "Starting run...";
			if(conn->engine.step(Event("start")) == Engine::Processed)
			{
				LOG_INFO() << "Minos run start done";
				return minos_ok;
			}
		}
		catch(const dhsm::ServantError& e)
		{
			conn->engine.reinitiate(conn->engine.currentStateMachine());
			throw;
		}
	}
	HANDLE_EXCEPTIONS
	LOG_WARN() << "Minos run start failed !";
	return minos_failed;
}

minos_error minos_stop(minos_connection* conn)
{
	try
	{
		try
		{
			if(!conn) throw std::runtime_error("invalid minos connection!");
			LOG_INFO() << "Stopping run...";
			if(conn->engine.step(Event("stop")) == Engine::Processed)
			{
				LOG_INFO() << "Minos run stop done";
				return minos_ok;
			}
		}
		catch(const dhsm::ServantError& e)
		{
			conn->engine.reinitiate(conn->engine.currentStateMachine());
			throw;
		}
	}
	HANDLE_EXCEPTIONS
	LOG_WARN() << "Minos run stop failed !";
	return minos_failed;
}

}
