/* =============================================================================
 * FeminosControllerTest.cpp
 * -----------------------------------------------------------------------------
 * Created on: 21 janv. 2013 at Irfu/Sedi/Lilas, CEA Saclay, F-91191, France
 * Author: fchateau
 * -----------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * -----------------------------------------------------------------------------
 * Contributors: fchateau
 * -----------------------------------------------------------------------------
 * This software is part of the DaqMinos project.
 * -----------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding
 * by the rules of distribution of free software. You can use, modify and/or
 * redistribute the software under the terms of the CeCILL license as circulated
 * by CEA, CNRS and INRIA at the following URL: "http://www.cecill.info".
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability. In this
 * respect, the user's attention is drawn to the risks associated with loading,
 * using, modifying and/or developing or reproducing the software by the user in
 * light of its specific status of free software, that may mean that it is
 * complicated to manipulate, and that also therefore means that it is reserved
 * for developers and experienced professionals having in-depth computer
 * knowledge. Users are therefore encouraged to load and test the software's
 * suitability as regards their requirements in conditions enabling the security
 * of their systems and/or data to be ensured and, more generally, to use and
 * operate it in the same conditions as regards security.
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * -----------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for
 * commercial purposes. For this you will need to negotiate a specific contract
 * with a legal representative of CEA.
 * =============================================================================
 */

#include <minos/fem/FeminosController.h>
#include <boost/test/auto_unit_test.hpp>

using namespace minos::fem;
using namespace utl::net;

struct ControllerFixture
{
	ControllerFixture()
		: feminos(SocketAddress(IpAddress(192, 168, 1, 10), 1122))
	{
	}

	void checkPedThrList(ListResult result, unsigned asicCount, bool isThreshold, PedValue refValue);

	FeminosController feminos;
};

void ControllerFixture::checkPedThrList(ListResult result, unsigned asicCount, bool isThreshold, PedValue refValue)
{
	BOOST_CHECK_EQUAL(result.pedestalList().size(), asicCount);
	for(unsigned asic = 0; asic < asicCount; ++asic)
	{
		const PedThrList& list = result.pedestalList()[asic];
		unsigned channels = list.header->listLength();
		BOOST_CHECK((list.header->isAfter == 1 and channels == AfterChannelCount) or
				    (list.header->isAfter == 0 and channels == AgetChannelCount));
		BOOST_CHECK_EQUAL(list.header->asic, asic);
		BOOST_CHECK_EQUAL(list.header->isThreshold, isThreshold);
		for(unsigned i = 0; i < channels; ++i)
		{
			BOOST_CHECK_EQUAL(list.values[i], refValue);
		}
	}
}


BOOST_FIXTURE_TEST_SUITE(minos_fem_FeminosController, ControllerFixture)

// =============================================================================
// Pedestals
// -----------------------------------------------------------------------------

BOOST_AUTO_TEST_CASE(test_pedestal_get_set_one)
{
	feminos.pedestal_set(0, 0, 42);
	PedValue ped = feminos.pedestal_get(0, 0);
	BOOST_CHECK_EQUAL(ped, 42);
}

BOOST_AUTO_TEST_CASE(test_pedestal_get_set_channel)
{
	feminos.pedestal_set(AllAsics, 0, 129);
	for(int i = 0; i < 4; ++i)
	{
		PedValue ped = feminos.pedestal_get(i, 0);
		BOOST_CHECK_EQUAL(ped, 129);
	}
}

BOOST_AUTO_TEST_CASE(test_pedestal_list_set_asic)
{
	feminos.pedestal_set(0, AllChannels, 17);
	ListResult result;
	feminos.pedestal_list(result, 0);
	checkPedThrList(result, 1, false, 17);
}

BOOST_AUTO_TEST_CASE(test_pedestal_list_set_all)
{
	feminos.pedestal_set(AllAsics, AllChannels, 219);
	ListResult result;
	feminos.pedestal_list(result, AllAsics);
	checkPedThrList(result, 4, false, 219);
}

BOOST_AUTO_TEST_CASE(test_pedestal_get_set_range)
{
	feminos.pedestal_set(AllAsics, AllChannels, 0);
	feminos.pedestal_set(AsicRange(1, 2), ChannelRange(10, 20), 87);
	ListResult result;
	feminos.pedestal_list(result, AllAsics);
	for(unsigned asic = 0; asic < 4; ++asic)
	{
		unsigned channelCount = result.pedestalList()[asic].header->listLength();
		for(unsigned channel = 0; channel < channelCount; ++channel)
		{
			if((asic == 1 or asic == 2) and channel >= 10 and channel <= 20)
			{
				BOOST_CHECK_EQUAL(result.pedestalList()[asic].values[channel], 87);
			}
			else
			{
				BOOST_CHECK_EQUAL(result.pedestalList()[asic].values[channel], 0);
			}
		}
	}
}

// =============================================================================
// Thresholds
// -----------------------------------------------------------------------------

BOOST_AUTO_TEST_CASE(test_threshold_get_set_one)
{
	feminos.threshold_set(0, 0, 7952);
	ThrValue thr = feminos.threshold_get(0, 0);
	BOOST_CHECK_EQUAL(thr, 7952);
}

BOOST_AUTO_TEST_CASE(test_threshold_get_set_channel)
{
	feminos.threshold_set(AllAsics, 0, 34856);
	for(int i = 0; i < 4; ++i)
	{
		PedValue ped = feminos.threshold_get(i, 0);
		BOOST_CHECK_EQUAL(ped, 34856);
	}
}

BOOST_AUTO_TEST_CASE(test_threshold_list_set_asic)
{
	feminos.threshold_set(0, AllChannels, 65247);
	ListResult result;
	feminos.threshold_list(result, 0);
	checkPedThrList(result, 1, true, 65247);
}

BOOST_AUTO_TEST_CASE(test_threshold_list_set_all)
{
	feminos.threshold_set(AllAsics, AllChannels, 351);
	ListResult result;
	feminos.threshold_list(result, AllAsics);
	checkPedThrList(result, 4, true, 351);
}

// =============================================================================
// DAQ
// -----------------------------------------------------------------------------

BOOST_AUTO_TEST_CASE(test_daq_get_set_target)
{
	feminos.daq_setTarget(DataToPedestal);
	DaqTarget target = feminos.daq_getTarget();
	BOOST_CHECK_EQUAL(target, DataToPedestal);
}

BOOST_AUTO_TEST_CASE(test_daq_get_set_lossPolicy)
{
	feminos.daq_setLossPolicy(DaqLossRecredit);
	DaqLossPolicy policy = feminos.daq_getLossPolicy();
	BOOST_CHECK_EQUAL(policy, DaqLossRecredit);
}

BOOST_AUTO_TEST_CASE(test_daq_add_get_credits)
{
	DaqCredit credit;
	DaqCreditUnit unit;
	feminos.daq_clearCredits();
	feminos.daq_addCredits(5000, ByteCredits);
	feminos.daq_getCredits(credit, unit);
	BOOST_CHECK_EQUAL(credit, 5000);
	BOOST_CHECK_EQUAL(unit, ByteCredits);
//	feminos.daq_addCredits(2, FrameCredits);
}

BOOST_AUTO_TEST_CASE(test_daq_get_set_CreditWaitTime)
{
	BOOST_FAIL("Test not defined");
}

// =============================================================================
// Buffer
// -----------------------------------------------------------------------------

BOOST_AUTO_TEST_CASE(test_buffer_get_set_timeout)
{
	feminos.buffer_setTimeout(BufferTimeout_100ms);
	BufferTimeout timeout = feminos.buffer_getTimeout();
	BOOST_CHECK_EQUAL(timeout, BufferTimeout_100ms);
}

BOOST_AUTO_TEST_CASE(test_buffer_isTimeoutEnabled)
{
	feminos.buffer_enableTimeout(false);
	bool isEnabled = feminos.buffer_isTimeoutEnabled();
	BOOST_CHECK_EQUAL(isEnabled, false);
}

BOOST_AUTO_TEST_CASE(test_buffer_reset_getInfo)
{
	feminos.buffer_reset();
	BufferInfo info;
	feminos.buffer_getInfo(info);
	BOOST_CHECK_EQUAL(info.baseAddress, 0xC2000000);
	BOOST_CHECK_EQUAL(info.capacity, 8092);
	BOOST_CHECK_EQUAL(info.pndCount, 0);
	BOOST_CHECK_EQUAL(info.isRunActive, false);
	BOOST_CHECK_EQUAL(info.hasTimeout, false);
	BOOST_CHECK_EQUAL(info.timeout, BufferTimeout_1ms);
}

BOOST_AUTO_TEST_CASE(test_buffer_setActive)
{
	BufferInfo info;
	feminos.buffer_setActive(true);
	feminos.buffer_getInfo(info);
	BOOST_CHECK_EQUAL(info.isRunActive, true);
	feminos.buffer_setActive(false);
	feminos.buffer_getInfo(info);
	BOOST_CHECK_EQUAL(info.isRunActive, false);
}

// =============================================================================
// Dead-time histogram
// -----------------------------------------------------------------------------

BOOST_AUTO_TEST_CASE(test_hbusy_clear)
{
	feminos.hbusy_clear();
	DeadtimeHistoResult deadtimeInfo;
	feminos.hbusy_getHisto(deadtimeInfo);

	BOOST_CHECK_EQUAL(deadtimeInfo.deadtimeStats()->minBin, 0);
	BOOST_CHECK_EQUAL(deadtimeInfo.deadtimeStats()->maxBin, 1023);
	BOOST_CHECK_EQUAL(deadtimeInfo.deadtimeStats()->binWidth, 1);
	BOOST_CHECK_EQUAL(deadtimeInfo.deadtimeStats()->binCount, 1024);
	BOOST_CHECK_EQUAL(deadtimeInfo.deadtimeStats()->minNonNullBin, 1023);
	BOOST_CHECK_EQUAL(deadtimeInfo.deadtimeStats()->maxNonNullBin, 0);
	BOOST_CHECK_EQUAL(deadtimeInfo.deadtimeStats()->mean, 0);
	BOOST_CHECK_EQUAL(deadtimeInfo.deadtimeStats()->stdev, 0);
	BOOST_CHECK_EQUAL(deadtimeInfo.deadtimeStats()->entryCount, 0);

	BOOST_CHECK_EQUAL(deadtimeInfo.deadtimeHisto().count(), 0);
}

// =============================================================================
// Pedestal histogram
// -----------------------------------------------------------------------------

BOOST_AUTO_TEST_CASE(test_hped_clear)
{
	//@TODO
}

BOOST_AUTO_TEST_CASE(test_hped_getSummary)
{
	Reg0 config = feminos.register_read(0);

	feminos.hped_clear(AllAsics, AllChannels);
	PedestalSummaryResult result;
	feminos.hped_getSummary(result, 0, AllChannels);
	unsigned nextAsic = 0;
	unsigned nextChannel = 0;
	for(unsigned i = 0; i < result.pedestalSummary().count(); ++i)
	{
		BOOST_CHECK_EQUAL(result.pedestalSummary()[i].channel.chip, nextAsic);
		BOOST_CHECK_EQUAL(result.pedestalSummary()[i].channel.channel, nextChannel);
		BOOST_CHECK_EQUAL(result.pedestalSummary()[i].mean, 0);
		BOOST_CHECK_EQUAL(result.pedestalSummary()[i].stdev, 0);
		++nextChannel;
		if((config.mode_after == 1 and nextChannel == AfterChannelCount) or
		   (config.mode_after == 0 and nextChannel == AgetChannelCount))
		{
			nextChannel = 0;
			++nextAsic;
		}
	}
}

BOOST_AUTO_TEST_CASE(test_hped_getStats)
{
	feminos.hped_setOffset(AllAsics, AllChannels, 0);
	feminos.hped_clear(AllAsics, AllChannels);
	PedestalStatsResult result;
	feminos.hped_getStats(result, 0, 0);

	BOOST_CHECK_EQUAL(result.histoChannel()->chip, 0);
	BOOST_CHECK_EQUAL(result.histoChannel()->channel, 0);
	BOOST_CHECK_EQUAL(result.pedestalStats()->minBin, 0);
	BOOST_CHECK_EQUAL(result.pedestalStats()->maxBin, 511);
	BOOST_CHECK_EQUAL(result.pedestalStats()->binWidth, 1);
	BOOST_CHECK_EQUAL(result.pedestalStats()->binCount, 512);
	BOOST_CHECK_EQUAL(result.pedestalStats()->minNonNullBin, 511);
	BOOST_CHECK_EQUAL(result.pedestalStats()->maxNonNullBin, 0);
	BOOST_CHECK_EQUAL(result.pedestalStats()->mean, 0);
	BOOST_CHECK_EQUAL(result.pedestalStats()->stdev, 0);
	BOOST_CHECK_EQUAL(result.pedestalStats()->entryCount, 0);
}

BOOST_AUTO_TEST_CASE(test_hped_getHisto)
{
	// @TODO: generate some pedestal
	PedestalHistoResult result;
	feminos.hped_getHisto(result, 0, 0);

	BOOST_CHECK_EQUAL(result.histoChannel()->chip, 0);
	BOOST_CHECK_EQUAL(result.histoChannel()->channel, 0);

	BOOST_CHECK_EQUAL(result.pedestalHisto().count(), 0);
}

BOOST_AUTO_TEST_CASE(test_hped_setOffset)
{
	feminos.hped_setOffset(0, 0, 1000);
	PedestalStatsResult result;
	feminos.hped_getStats(result, 0, 0);
	BOOST_CHECK_EQUAL(result.pedestalStats()->minBin, 1000);
	BOOST_CHECK_EQUAL(result.pedestalStats()->maxBin, 1511);
	feminos.hped_setOffset(0, 0, 0);
}

BOOST_AUTO_TEST_CASE(test_hped_setThreshold)
{
//	feminos.hped_setThreshold(AllAsics, AllChannels, 250, 5.0);
}

BOOST_AUTO_TEST_CASE(test_hped_center)
{
//	feminos.hped_center(AllAsics, AllChannels, 250);
}

// =============================================================================
// S-Histogram
// -----------------------------------------------------------------------------

BOOST_AUTO_TEST_CASE(test_shisto_clear)
{
	feminos.shisto_clear(AllAsics, AllChannels);

	SHistoResult result;
	feminos.shisto_getHisto(result, 0, 0);

	for(unsigned i = 0; i < SHisto::BIN_COUNT; ++i)
	{
		BOOST_CHECK_EQUAL(result.sHisto()->bin[i], 0);
	}
}

BOOST_AUTO_TEST_CASE(test_shisto_setThreshold)
{
	feminos.shisto_setThreshold(10);
}

// =============================================================================
// Hit-count histogram
// -----------------------------------------------------------------------------

BOOST_AUTO_TEST_CASE(test_hhit_clear)
{
	feminos.hhit_clear(AllAsics);
	HitCountHistoResult result;
	feminos.hhit_getHisto(result, 0);

	BOOST_CHECK_EQUAL(result.hitCountHisto()->minBin, 0);
	BOOST_CHECK_EQUAL(result.hitCountHisto()->maxBin, 79);
	BOOST_CHECK_EQUAL(result.hitCountHisto()->binWidth, 1);
	BOOST_CHECK_EQUAL(result.hitCountHisto()->binCount, 80);
	BOOST_CHECK_EQUAL(result.hitCountHisto()->minNonNullBin, 79);
	BOOST_CHECK_EQUAL(result.hitCountHisto()->maxNonNullBin, 0);
	BOOST_CHECK_EQUAL(result.hitCountHisto()->mean, 0);
	BOOST_CHECK_EQUAL(result.hitCountHisto()->stdev, 0);
	BOOST_CHECK_EQUAL(result.hitCountHisto()->entryCount, 0);
	for(unsigned i = 0; i < HitCountHisto::BIN_COUNT; ++i)
	{
		BOOST_CHECK_EQUAL(result.hitCountHisto()->bin[i], 0);
	}
}

BOOST_AUTO_TEST_SUITE_END()
