/**
 * @file Application.cpp
 * @date 23 avr. 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "ApplicationLoader.h"
#include "Application.h"
#include <mdcs/cfg/AppConfigFromFile.h>
#include <mdcs/cfg/CommandLineIdentityProvider.h> // for reference injection
#include <mdcs/cfg/AbstractConfig.h>
#include <mdcs/cfg/Constants.h>
#include <mdcs/utl/ConsoleLoggingBackend.h>
#include <utl/Factory.ipp>
#include <utl/Singleton.ipp>
#include <utl/Logging.h>
#include ICE_HEADER(Properties.h)
#include ICE_HEADER(Initialize.h)

#ifdef __unix__
#include <dlfcn.h>
#endif

namespace mdcs
{
namespace acq
{

// =============================================================================
// mdcs::acq::ApplicationLoader class methods body
// -----------------------------------------------------------------------------

ApplicationLoader::ApplicationLoader(CompiledParams& cfg)
	: argc_(0), argv_(0), config(cfg), loaderParams_()
{
}

ApplicationLoader::~ApplicationLoader()
{
}

void ApplicationLoader::initFromArgv(int& argc, char* argv[])
{
	Ice::PropertiesPtr props = Ice::createProperties();
	Ice::StringSeq args = Ice::argsToStringSeq(argc, argv);
	args = props->parseCommandLineOptions(cfg::COMMAND_LINE_NAMESPACE, args);

	Ice::InitializationData id;
	id.properties = props;
	ic = Ice::initialize(args, id);

	argc_ = argc;
	argv_ = argv;

	initFromProperties(props);
}

void ApplicationLoader::initFromProperties(const Ice::PropertiesPtr& props)
{
	retrieveIdentity(props);
	processAppConfig(props);
}

void ApplicationLoader::installLogger()
{
	::utl::BackendLogger::setBackend(::utl::LoggingBackendPtr(new utl::ConsoleLoggingBackend));
}

void ApplicationLoader::retrieveIdentity(const Ice::PropertiesPtr& props)
{
	cfg::IdentityProvider::instance().setDefaultProvider(config.defaultIdentityProvider);
	cfg::IdentityProvider::instance().setDefaultType(config.defaultIdentityType);
	cfg::IdentityProvider::instance().init(props);
	LOG_INFO() << "Identity set to " << cfg::IdentityProvider::instance().id();
}

void ApplicationLoader::processAppConfig(const Ice::PropertiesPtr& props)
{
	cfg::AbstractConfig::selectConfigFormat(config.appConfigFormat);
	cfg::AppConfigProvider cfgProvider;
	cfgProvider.setDefaultProvider(cfg::AppConfigFromFile::PROVIDER_NAME);
	cfg::ConfigPtr appConfig = cfgProvider.retrieveConfig(props);
	cfg::Configurator::instance().configure(loaderParams_, *appConfig);
	cfg::AbstractConfig::selectConfigFormat(loaderParams_.configFormat);

	installLogger(); // TODO: when the logger factory will be available, put this after loadPlugins()
	loadPlugins();

	LOG_DEBUG() << "Creating application params: " << loaderParams_.applicationName;
	appParams_.reset(ApplicationFactory::instance()[loaderParams_.applicationName].create().release());
	LOG_DEBUG() << "Configuring application params";
	::utl::GlobalConfigurator::instance().configure(*appParams_, *appConfig);
}

void ApplicationLoader::loadPlugin(const std::string& pluginName)
{
#ifdef __unix__
	LOG_INFO() << "Loading plugin: " << pluginName;
	if(dlopen(pluginName.c_str(), RTLD_LAZY | RTLD_LOCAL) == 0)
	{
		LOG_ERROR() << "Error loading plugin:" << dlerror();
	}
#else
// TODO: implement plugins on other platforms
	LOG_ERROR() << "Loading plugins is not yet implemented on this platform!";
#endif
}

void ApplicationLoader::loadPlugins()
{
	for(Params::PluginList::const_iterator iPlugin = loaderParams_.plugins.begin(),
		ePlugin = loaderParams_.plugins.end(); iPlugin != ePlugin; ++iPlugin)
	{
		if(!iPlugin->empty())
		{
			loadPlugin(*iPlugin);
		}
	}
}

std::auto_ptr<Application> ApplicationLoader::createApplication()
{
	std::auto_ptr<Application> app;
	LOG_INFO() << "Creating application: " << loaderParams_.applicationName;
	app.reset(appParams_->createApplication(*this).release());
	return app;
}

void ApplicationLoader::main(int argc, char* argv[])
{
	try
	{
		initFromArgv(argc, argv);
		boost::scoped_ptr<Application> application(createApplication().release());
		application->mainLoop();

		LOG_INFO() << "Normal shutdown.";
	}
	catch(std::exception& e)
	{
		LOG_ERROR() << e.what();
	}
	catch(...)
	{
	}
	LOG_INFO() << "Closing application.";
	if(ic)
	{
		ic->destroy();
	}
}

// =============================================================================
// mdcs::acq::ApplicationLoader::CompiledParams class methods body
// -----------------------------------------------------------------------------

ApplicationLoader::CompiledParams::CompiledParams()
	: appConfigFormat(cfg::DEFAULT_CONFIG_TYPE)
{
}

ApplicationLoader::CompiledParams::~CompiledParams()
{
}

} /* namespace acq */
} /* namespace mdcs */
