/**
 * @file Application.cpp
 * @date 23 avr. 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "ServerApplication.h"
#include "ApplicationLoader.h"
#include "mon/IceStormMetricsCollector.h"
#include <mdcs/cfg/CommandLineIdentityProvider.h> // for reference injection
#include <mdcs/cfg/AbstractConfig.h>
#include <mdcs/utl/ConsoleLoggingBackend.h>
#include <dhsm/Engine.h>
#include <utl/Factory.ipp>
#include <utl/Singleton.ipp>
#include <utl/Logging.h>
#include ICE_HEADER(Properties.h)
#include ICE_HEADER(Initialize.h)

namespace mdcs
{
namespace acq
{

// =============================================================================
// mdcs::acq::ServerApplication class methods body
// -----------------------------------------------------------------------------

ServerApplication::ServerApplication(const Params& params,
	ApplicationLoader& loader,
	const DistributedEngineCreator& creator,
	const std::string& identity,
	const std::string& stateMachineName)
	: Application(params, loader),
	  adapter_(createAdapter(loader, params)),
	  server_(creator, identity, stateMachineName, adapter_),
	  collector_(new mon::IceStormMetricsCollector(loader.communicator()))
{
	configureLogger();
	createTopContext(server());
}

ServerApplication::~ServerApplication()
{
}

void ServerApplication::createTopContext(const StateMachineServer& server)
{
	server.engine()->setTopContext(dhsm::StateDataPtr(new SmContext(*this)));
}

void ServerApplication::configureLogger()
{
	utl::ConsoleLoggingBackend& loggingBackend = dynamic_cast<utl::ConsoleLoggingBackend&>(*::utl::BackendLogger::backend());

	std::ostringstream idStr;
	const cfg::NodeId& id = cfg::IdentityProvider::instance().id();
	idStr << id.type()[0]; // first character of identity
	if(!id.isUnique())
	{
		idStr << id.num();
	}
	loggingBackend.setIdentity(idStr.str());
}

Ice::ObjectAdapterPtr ServerApplication::createAdapter(const ApplicationLoader& loader, const Params& params)
{
	LOG_INFO() << "Creating primary adapter: " << params.primaryAdapter;
	return loader.communicator()->createObjectAdapter(params.primaryAdapter);

}

void ServerApplication::mainLoop()
{
	LOG_INFO() << "Activating primary adapter.";
	adapter_->activate();

	LOG_INFO() << "Waiting for communicator shutdown...";
	loader_.communicator()->waitForShutdown();
}

// =============================================================================
// mdcs::acq::ServerApplication::Params class methods body
// -----------------------------------------------------------------------------

ServerApplication::Params::~Params()
{
}

// =============================================================================
// mdcs::acq::BaseState class methods body
// -----------------------------------------------------------------------------

ApplicationState::ApplicationState(const StateData& base)
	: dhsm::StateData(base)
{
}

ApplicationState::~ApplicationState()
{
}

ServerApplication& ApplicationState::application()
{
	return engine()->topContext<ServerApplication::SmContext>().application();
}

// =============================================================================
// mdcs::acq::ConfiguredState class methods body
// -----------------------------------------------------------------------------

ConfiguredState::ConfiguredState(const StateData& base)
	: ApplicationState(base)
{
}

ConfiguredState::~ConfiguredState()
{
}

void ConfiguredState::configure(const std::string& configStr)
{
	cfg::ConfigPtr config = cfg::ConfigFactory::instance().create(configStr);
	cfg::Configurator::instance().configure(*this, *config);

	apply();
}

// =============================================================================
// mdcs::acq::StartedState class methods body
// -----------------------------------------------------------------------------

StartedState::StartedState(const StateData& base)
	: ApplicationState(base)
{
}

StartedState::~StartedState()
{
}

void StartedState::start(const std::string& runCfg)
{
	cfg::ConfigPtr config = cfg::ConfigFactory::instance().create(runCfg);
	cfg::Configurator::instance().configure(runInfo_, *config);

	doStart();
}

} /* namespace acq */
} /* namespace mdcs */
