/* =============================================================================
 * StateMachineFactory.cpp
 * -----------------------------------------------------------------------------
 * class mdcs::acq::StateMachineFactory
 * Created on: 27 mars 2013 at Irfu/Sedi/Lilas, CEA Saclay, F-91191, France
 * Author: fchateau
 * -----------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * -----------------------------------------------------------------------------
 * Contributors: fchateau
 * -----------------------------------------------------------------------------
 * This file is part of the Mordicus software project.
 * -----------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding
 * by the rules of distribution of free software. You can use, modify and/or
 * redistribute the software under the terms of the CeCILL license as circulated
 * by CEA, CNRS and INRIA at the following URL: "http://www.cecill.info".
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability. In this
 * respect, the user's attention is drawn to the risks associated with loading,
 * using, modifying and/or developing or reproducing the software by the user in
 * light of its specific status of free software, that may mean that it is
 * complicated to manipulate, and that also therefore means that it is reserved
 * for developers and experienced professionals having in-depth computer
 * knowledge. Users are therefore encouraged to load and test the software's
 * suitability as regards their requirements in conditions enabling the security
 * of their systems and/or data to be ensured and, more generally, to use and
 * operate it in the same conditions as regards security.
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * -----------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for
 * commercial purposes. For this you will need to negotiate a specific contract
 * with a legal representative of CEA.
 * =============================================================================
 */

#include "StateMachineFactory.h"
#include "StateMachineServer.h"
#include <mdcs/cfg/AbstractConfig.h>
#include <dhsm/Engine.h>
#include <dhsm/EngineServant.h>
#include <dhsm/StateMachine.h>
#include <utl/Factory.ipp>
#include <utl/Singleton.ipp>
#include <utl/Logging.h>

namespace mdcs
{
namespace acq
{
using namespace dhsm;

std::auto_ptr<dhsm::StateMachine> createBaseStateMachine()
{
	StateMachinePtr acqSm(new StateMachine("DAQ_SM"));

	State& standby    = acqSm->state("STANDBY");
	State& configured = acqSm->state("CONFIGURED");
	acqSm->initial(standby);

	StateMachine& configuredSm = configured.stateMachine("Configured");
	State& ready   = configuredSm.state("READY");
	State& started = configuredSm.state("STARTED");
	configuredSm.initial(ready);

	acqSm->transition("configure", standby, configured);
	acqSm->transition("reset", configured, standby);

	configuredSm.transition("start", ready, started);
	configuredSm.transition("stop", started, ready);

	return acqSm;
}

std::auto_ptr<dhsm::StateMachine> createCtrlStateMachine()
{
	StateMachinePtr acqSm(new StateMachine("DAQ_RC_SM"));

	State& disconnected = acqSm->state("DISCONNECTED");
	State& connected    = acqSm->state("CONNECTED");
	acqSm->initial(disconnected);

	StateMachine& connectedSm = connected.setStateMachine(createBaseStateMachine());
	connectedSm.label("Connected");

	acqSm->transition("connect", disconnected, connected);
	acqSm->transition("disconnect", connected, disconnected);

	return acqSm;
}

static dhsm::StateMachine* createBaseStateMachineWrapper()
{
	return createBaseStateMachine().release();
}

static dhsm::StateMachine* createCtrlStateMachineWrapper()
{
	return createCtrlStateMachine().release();
}

REGISTER_FACTORY(StateMachineFactory, "MDaqBaseStateMachine", createBaseStateMachineWrapper)
REGISTER_FACTORY(StateMachineFactory, "MDaqCtrlStateMachine", createCtrlStateMachineWrapper)


// =============================================================================
// mdcs::acq::DistributedEngineCreator class methods body
// -----------------------------------------------------------------------------

DistributedEngineCreator::~DistributedEngineCreator()
{
}


// =============================================================================
// mdcs::acq::EngineCreator class methods body
// -----------------------------------------------------------------------------

EnginePtr LeafEngineCreator::createSmEngine(StateMachineServer& /*server*/) const
{
	return EnginePtr(new Engine);
}

std::auto_ptr<EngineServant> LeafEngineCreator::createSmServant(StateMachineServer& /*server*/) const
{
	return std::auto_ptr<EngineServant>(new EngineServant);
}

} /* namespace acq */
} /* namespace mdcs */

template class utl::FactoryMap<std::string, dhsm::StateMachine*()>; // FactoryMap_StateMachine
template class utl::Singleton<mdcs::acq::FactoryMap_StateMachine>;  // StateMachineFactory
template class utl::FactoryRegistrar<mdcs::acq::StateMachineFactory>;
