/**
 * @file FileStorage.cpp
 * @date 16 avr. 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "FileStorage.h"
#include <mdcs/acq/mon/ManagedMetric.ipp>
#include <mdcs/acq/mon/MetricsCollector.h>
#include <mdcs/acq/ServerApplication.h>
#include <utl/Logging.h>
#include <iomanip>
#include <sstream>
#include <cmath>

#if defined(VXWORKS) || defined(__APPLE__) || defined(_WIN32)
double pow10(double arg)
{
	return pow(10, arg);
}
#endif

namespace mdcs
{
namespace acq
{
namespace back
{
namespace policy
{

// =============================================================================
// mdcs::acq::back::policy::FileStorage::Params class methods body
// -----------------------------------------------------------------------------

FileStorage::Params::Params()
	: storeData(true),
	  limitFileSize(false),
	  maxFileSize(0x7FFFFFFF),
	  seqDigitCount(DEFAULT_DIGIT_COUNT),
	  dataDir(),
	  filePrefix(),
	  fileSuffix()
{
}

// =============================================================================
// mdcs::acq::back::policy::FileStorage::Setup class methods body
// -----------------------------------------------------------------------------

FileStorage::Setup::Setup(ApplicationState& state, const Params& params)
	: utl::BaseSetupData<Policy>(state, params),
	  bytesWrittenMetric_(state.application().collector().createMetric("FileStorage.bytesWritten")),
	  eventsReceivedMetric_(state.application().collector().createMetric("FileStorage.eventsReceived"))
{
}

// =============================================================================
// mdcs::acq::back::policy::FileStorage::RunImpl class methods body
// -----------------------------------------------------------------------------

FileStorage::RunImpl::RunImpl(ApplicationState& state, Setup& setup, const RunInfo& runInfo)
	: utl::BaseRunImpl<Policy>(state, setup, runInfo),
	  bytesWritten_(0, *setup.bytesWrittenMetric_),
	  eventsReceived_(0, *setup.eventsReceivedMetric_),
	  hasFailed(false),
	  sequenceNumber(0),
	  currentFileSize(0)
{
}

FileStorage::RunImpl::~RunImpl()
{
	endFile();
	mon::Metric::SampleTimestamp ts = mon::Metric::currentTimestamp();
	bytesWritten_.reportSample(ts);
	eventsReceived_.reportSample(ts);
}

void FileStorage::RunImpl::process(const uint8_t* buffer, uint32_t nBytes)
{
	++eventsReceived_;
	if(params.storeData and !hasFailed)
	{
		if(!currentFile.is_open())
		{
			beginFile();
		}
		else if(params.limitFileSize and currentFileSize + nBytes > params.maxFileSize)
		{
			endFile();
			beginFile();
		}
		currentFile.write(reinterpret_cast<const char*>(buffer), nBytes);
		if(currentFile.fail() or currentFile.bad())
		{
			LOG_WARN() << "Failed writing to file: " << currentFilePath << ", disabling storage.";
			hasFailed = true;
		}
		else
		{
			currentFileSize += nBytes;
			bytesWritten_ += nBytes;
		}
	}
	nextStageProcess(buffer, nBytes);
}

/**
 * Returns the maximum sequence number, given the specified number of digits.
 */
unsigned FileStorage::RunImpl::computeSequenceMax(unsigned digitCount)
{
	return static_cast<unsigned>(pow10(digitCount)) - 1;
}

/**
 * Starts using a new file.
 */
void FileStorage::RunImpl::beginFile()
{
	unsigned seqMax = computeSequenceMax(params.seqDigitCount);
	if(sequenceNumber > seqMax)
	{
		LOG_ERROR() << "Cannot create more than " << seqMax+1 << " acquisition files, "
			"because configuration says the suffix should have only " << params.seqDigitCount << " digit(s).";
		hasFailed = true;
	}
	else
	{
		currentFilePath = getCurrentDataFilePath();
		currentFile.open(currentFilePath.c_str(), std::ios::trunc | std::ios::out | std::ios::binary);
		currentFileSize = 0;
		if(currentFile.fail() or currentFile.bad())
		{
			LOG_WARN() << "Failed opening file: " << currentFilePath << ", disabling storage.";
			hasFailed = true;
		}
	}
}

/**
 * Stops using the current file, and increment the sequence number.
 */
void FileStorage::RunImpl::endFile()
{
	if(currentFile.is_open())
	{
		currentFile.close();
		++sequenceNumber;
	}
}

/**
 * Returns the full-path to the current data file.
 */
std::string FileStorage::RunImpl::getCurrentDataFilePath()
{
	std::ostringstream fileName;
	fileName << params.dataDir << '/' << params.filePrefix << runInfo.fileStem();
	if(params.limitFileSize)
	{
		fileName << '.' << std::setw(params.seqDigitCount) << std::setfill('0') << sequenceNumber;
	}
	fileName << params.fileSuffix;
	return fileName.str();
}

} /* namespace policy */
} /* namespace back */
} /* namespace acq */
} /* namespace mdcs */
