/**
 * @file IceDataReceiver.cpp
 * @date 10 avr. 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "IceDataReceiver.h"
#include "../DataProcessorApplication.h"
#include <mdcs/acq/mon/ManagedMetric.ipp>
#include <mdcs/acq/mon/MetricsCollector.h>
#include <mdcs/acq/ServerApplication.h>
#include <mdcs/cfg/Constants.h>
#include <utl/Logging.h>
#include ICE_HEADER(Communicator.h)

namespace mdcs
{
namespace acq
{
namespace back
{
namespace policy
{

// =============================================================================
// mdcs::acq::back::policy::IceDataReceiver::Params class methods body
// -----------------------------------------------------------------------------

IceDataReceiver::Params::Params()
	: collocated(true)
{
}

// =============================================================================
// mdcs::acq::back::policy::IceDataReceiver::Setup class methods body
// -----------------------------------------------------------------------------

IceDataReceiver::Setup::Setup(ApplicationState& state, const Params& params)
	: utl::BaseSetupData<Policy>(state, params),
	  bytesReceivedMetric_(state.application().collector().createMetric("IceDataReceiver.bytesReceived"))
{
	ServerApplication& app = state.application();
	if(params.collocated)
	{
		LOG_INFO() << "Collocating flow on existing adapter";
		adapter_ = app.server().adapter();
	}
	else
	{
		LOG_INFO() << "Creating Flow object adapter";
		flowAdapter_ = app.server().adapter()->getCommunicator()->createObjectAdapter(cfg::PROCESSOR_FLOW_ADAPTER);

		LOG_INFO() << "Activating Object Adapter: " << flowAdapter_->getName();
		flowAdapter_->activate();

		adapter_ = flowAdapter_;
	}
	receiver_ = new Receiver;
	adapter_->add(receiver_, processorIdentity());
}

IceDataReceiver::Setup::~Setup()
{
	try
	{
		adapter_->remove(processorIdentity());

		if(flowAdapter_)
		{
			flowAdapter_->deactivate();
			flowAdapter_ = 0;
		}
		adapter_ = 0;
	}
	catch(const Ice::Exception& e)
	{
		LOG_ERROR() << e.what();
	}
}

void IceDataReceiver::Setup::activateReceiver(RunImpl* run)
{
	receiver_->setRunImpl(run);
}

Ice::Identity IceDataReceiver::Setup::processorIdentity()
{
	return adapter_->getCommunicator()->stringToIdentity(cfg::PROCESSOR_FLOW_SERVANT);
}

// =============================================================================
// mdcs::acq::back::policy::IceDataReceiver::RunPolicy class methods body
// -----------------------------------------------------------------------------

IceDataReceiver::RunImpl::RunImpl(ApplicationState& state, Setup& setup, const RunInfo& runInfo)
	: utl::BaseRunImpl<Policy>(state, setup, runInfo),
	  processor_(state.application<DataProcessorApplication>()),
	  bytesReceived_(0, setup.bytesReceivedMetric())
{
	setup.activateReceiver(this);
}

IceDataReceiver::RunImpl::~RunImpl()
{
}

void IceDataReceiver::RunImpl::process(const uint8_t* buffer, uint32_t nBytes)
{
	bytesReceived_ += nBytes;
	processor_.processFrame(buffer, nBytes);
}

// =============================================================================
// mdcs::acq::back::policy::IceDataReceiver::RunImpl::Receiver class methods body
// -----------------------------------------------------------------------------

IceDataReceiver::Setup::Receiver::Receiver()
	: impl_(0)
{
}

void IceDataReceiver::Setup::Receiver::processData(const std::pair<const Ice::Byte*, const Ice::Byte*>& data, const Ice::Current&)
{
	if(impl_ != 0)
	{
		impl_->process(data.first, data.second - data.first);
	}
}

} /* namespace policy */
} /* namespace back */
} /* namespace acq */
} /* namespace mdcs */
