/**
 * @file DataGeneratorApplication.cpp
 * @date 22 avr. 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "DataGeneratorApplication.h"
#include "AbstractDataProducer.h"
#include <mdcs/acq/StateMachineFactory.h>
#include <mdcs/acq/front/AbstractDataSender.h>
#include <mdcs/acq/front/AbstractInputBuffer.h>
#include <mdcs/cfg/IdentityProvider.h>
#include <mdcs/cfg/Constants.h>
#include <mdcs/cfg/CCfgUtils.ipp>
#include <mdcs/utl/PolicyTuple.hpp>
#include <mdcs/utl/Tuples.hpp>
#include <dhsm/Engine.h>
#include <utl/Factory.ipp>
#include <utl/Singleton.ipp>
#include <utl/Construct.hpp>

namespace mdcs
{
namespace acq
{
namespace front
{
using namespace dhsm;

// =============================================================================
// mdcs::acq::front::DataGeneratorApplication class methods body
// -----------------------------------------------------------------------------

DataGeneratorApplication::DataGeneratorApplication(const Params& params, ApplicationLoader& loader)
	: ServerApplication(params,
			loader,
			LeafEngineCreator(),
			cfg::GENERATOR_CTRL_SERVANT,
			cfg::GENERATOR_SM_NAME)
{
}

DataGeneratorApplication::~DataGeneratorApplication()
{
}

std::auto_ptr<dhsm::StateMachine> DataGeneratorApplication::createStateMachine()
{
	StateMachinePtr acqSm = createBaseStateMachine();
	State& configured = acqSm->findState("CONFIGURED");
	configured.setStateDataType<Configured>();
	configured.enterActions().add(&Configured::configure);

	State& started = configured.stateMachine().findState("STARTED");
	started.setStateDataType<Started>();
	started.enterActions().add(&Started::start);
	return acqSm;
}

static dhsm::StateMachine* stateMachineFactoryWrapper()
{
	return DataGeneratorApplication::createStateMachine().release();
}

REGISTER_FACTORY(StateMachineFactory, cfg::GENERATOR_SM_NAME, stateMachineFactoryWrapper)
REGISTER_FACTORY(ApplicationFactory, cfg::GENERATOR_APPNAME, ::utl::Construct<DataGeneratorApplication::Params*>())

// =============================================================================
// mdcs::acq::front::DataGeneratorApplication::Configured class methods body
// -----------------------------------------------------------------------------

DataGeneratorApplication::Configured::Configured(const StateData& base)
	: ConfiguredPolicies<Policies>(base)
{
}

DataGeneratorApplication::Configured::~Configured()
{
}

// =============================================================================
// mdcs::acq::front::DataGeneratorApplication::Started class methods body
// -----------------------------------------------------------------------------

DataGeneratorApplication::Started::Started(const StateData& base)
	: StartedPolicies<Policies>(base)
{
}

void DataGeneratorApplication::Started::doStart()
{
	StartedPolicies<Policies>::doStart();

	senderThread_ = new SenderThread(*dataSender(), *dataProducer());
	senderThreadCtrl_ = senderThread_->start();

	senderThread_->waitReady();
	dataProducer()->start();
}

DataGeneratorApplication::Started::~Started()
{
	try
	{
		LOG_INFO() << "stopping data producer...";
		dataProducer()->stop();
		LOG_INFO() << "stopping sender thread";
		senderThread_->askStop();
		LOG_INFO() << "waiting for sender thread to complete...";
		senderThreadCtrl_.join();
		LOG_INFO() << "leaving started state";
	}
	catch(std::exception& e)
	{
		LOG_ERROR() << e.what();
	}
}

// =============================================================================
// mdcs::acq::front::DataGeneratorApplication::Started class methods body
// -----------------------------------------------------------------------------

DataGeneratorApplication::SenderThread::SenderThread(AbstractDataSender& sender, AbstractDataProducer& producer)
	: sender_(sender), producer_(producer), buffer_(), readyMonitor_(),
	  isReady_(false)
{
}

void DataGeneratorApplication::SenderThread::run()
{
	try
	{
		buffer_.reset(producer_.createBuffer().release());
		signalReady();
		while(sender_.sendData(*buffer_));
	}
	catch(std::exception& e)
	{
		LOG_ERROR() << "Exception in sender thread: " << e.what();
	}
}

void DataGeneratorApplication::SenderThread::askStop()
{
	buffer_->interruptRead();
}

void DataGeneratorApplication::SenderThread::waitReady()
{
	IceUtil::Monitor<IceUtil::Mutex>::Lock lock(readyMonitor_);
	while(!isReady_)
	{
		readyMonitor_.wait();
	}
}

void DataGeneratorApplication::SenderThread::signalReady()
{
	IceUtil::Monitor<IceUtil::Mutex>::Lock lock(readyMonitor_);
	isReady_ = true;
	readyMonitor_.notify();
}

} /* namespace front */
} /* namespace acq */
} /* namespace mdcs */
