/**
 * @file IceStormMetricsCollector.cpp
 * @date 30 sept. 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "IceStormMetricsCollector.h"
#include <utl/Logging.h>

namespace mdcs
{
namespace acq
{
namespace mon
{
// =============================================================================
// mdcs::acq::mon::IceStormMetricsCollector class methods body
// -----------------------------------------------------------------------------

IceStormMetricsCollector::IceStormMetricsCollector(const Ice::CommunicatorPtr& ic)
	: topicManager_(IceStorm::TopicManagerPrx::checkedCast(ic->stringToProxy("IceStorm/TopicManager")))
{
	metaMetric_.reset(createMetric(Metric::META_METRIC_ID).release());
}

IceStormMetricsCollector::~IceStormMetricsCollector()
{
}

MetricPtr IceStormMetricsCollector::createMetric(const std::string& id)
{
	return MetricPtr(new IceStormMetric(*this, id));
}

// =============================================================================
// mdcs::acq::mon::IceStormMetricsCollector::IceStormMetric class methods body
// -----------------------------------------------------------------------------

IceStormMetricsCollector::IceStormMetric::IceStormMetric(IceStormMetricsCollector& collector, const std::string& topicName)
	: Metric(collector, topicName)
{
    while(!topic_)
	{
		try
		{
			topic_ = collector.topicManager()->retrieve(topicName);
			LOG_DEBUG() << "Topic " << topicName << " retrieved";
		}
		catch(const IceStorm::NoSuchTopic&)
		{
			try
			{
				topic_ = collector.topicManager()->create(topicName);
				LOG_DEBUG() << "Topic " << topicName << " created";
			}
			catch(const IceStorm::TopicExists&)
			{
				// Another client created the topic between the calls to retrieve and create.
			}
		}
	}
    Metric* metaMetric = collector.metaMetric();
    if(metaMetric != 0) // the meta metric cannot report its creation itself...
    {
    	metaMetric->reportSample(topicName, Metric::currentTimestamp());
    }
    publisher_ = MetricTopicPrx::uncheckedCast(topic_->getPublisher()->ice_oneway());
	LOG_DEBUG() << "Publisher for " << topicName << " created";
}

IceStormMetricsCollector::IceStormMetric::~IceStormMetric()
{
// Calling destroy() would not be safe, because several sources may emit on the same topic
// and reference counting may not be so easy to implement, in particular to support
// the crash of a collector.
//	topic_->destroy();
//	LOG_INFO() << "topic destroyed";
}

const Metric::SampleTimestamp& IceStormMetricsCollector::IceStormMetric::iceTimestampEpoch()
{
	static const SampleTimestamp epoch(boost::gregorian::date(2000, 1, 1));
	return epoch;
}

Ice::Long IceStormMetricsCollector::IceStormMetric::toTimestamp(const SampleTimestamp& time)
{
	return (time - iceTimestampEpoch()).total_microseconds();
}

void IceStormMetricsCollector::IceStormMetric::reportSample(const int64_t& sample, const SampleTimestamp& time)
{
	LOG_DEBUG() << "reporting long " << sample;
	publisher_->intScalar(sample, toTimestamp(time));
}

void IceStormMetricsCollector::IceStormMetric::reportSample(const double& sample, const SampleTimestamp& time)
{
	LOG_DEBUG() << "reporting dble " << sample;
	publisher_->fltScalar(sample, toTimestamp(time));
}

void IceStormMetricsCollector::IceStormMetric::reportSample(const std::string& sample, const SampleTimestamp& time)
{
	LOG_DEBUG() << "reporting str " << sample;
	publisher_->strScalar(sample, toTimestamp(time));
}

void IceStormMetricsCollector::IceStormMetric::reportSample(const std::vector<int64_t>& samples, const SampleTimestamp& time)
{
	LOG_DEBUG() << "reporting long vector";
	publisher_->intVector(samples, toTimestamp(time));
}

void IceStormMetricsCollector::IceStormMetric::reportSample(const std::vector<double>& samples, const SampleTimestamp& time)
{
	LOG_DEBUG() << "reporting dble vector";
	publisher_->fltVector(samples, toTimestamp(time));
}

void IceStormMetricsCollector::IceStormMetric::reportSample(const std::vector<std::string>& samples, const SampleTimestamp& time)
{
	LOG_DEBUG() << "reporting str vector";
	publisher_->strVector(samples, toTimestamp(time));
}
void IceStormMetricsCollector::IceStormMetric::reportSample(const ::utl::io::ConstBuffer& sample, const SampleTimestamp& time)
{
	LOG_DEBUG() << "reporting custom data";
	publisher_->customData(std::make_pair(sample.begin(), sample.end()), toTimestamp(time));
}

} /* namespace mon */
} /* namespace acq */
} /* namespace mdcs */
