/**
 * @file MetricsCollector.cpp
 * @date 30 sept. 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include <boost/asio/deadline_timer.hpp>
#include <boost/asio/placeholders.hpp>
#include "MetricsCollector.h"
#include "ManagedMetric.ipp"
#include <utl/Logging.h>

namespace mdcs
{
namespace acq
{
namespace mon
{

// =============================================================================
// mdcs::acq::mon::MetricsCollector::Thread class methods body
// -----------------------------------------------------------------------------

class MetricsCollector::Thread : public IceUtil::Thread
{
public:
	Thread(MetricsCollector& collector, boost::posix_time::time_duration period)
		: collector_(collector),period_(period),
		  timer_(io_, period_), stop_(false)
	{
	}
	void stop()
	{
		stop_ = true;
	}
protected:
	void run()
	{
		timer_.async_wait(boost::bind(&Thread::onTimer, this, boost::asio::placeholders::error));
		io_.run();
	}
	void onTimer(const boost::system::error_code& /*e*/)
	{
		collector_.collect();
		if(!stop_)
		{
			timer_.expires_at(timer_.expires_at() + period_);
			timer_.async_wait(boost::bind(&Thread::onTimer, this, boost::asio::placeholders::error));
		}
	}
	MetricsCollector& collector_;
	boost::posix_time::time_duration period_;
	boost::asio::io_service io_;
	boost::asio::deadline_timer timer_;
	bool stop_;
};

// =============================================================================
// mdcs::acq::mon::MetricsCollector class methods body
// -----------------------------------------------------------------------------

MetricsCollector::MetricsCollector()
	: collectorThread_(new Thread(*this, boost::posix_time::seconds(1))),
	  collectorThreadCtrl_(collectorThread_->start())
{
}

MetricsCollector::~MetricsCollector()
{
	collectorThread_->stop();
	collectorThreadCtrl_.join();
}

void MetricsCollector::registerManagedMetric(BaseManagedMetric& metric)
{
	IceUtil::Mutex::Lock lock(managedMetricsMutex_);
	managedMetrics_.push_back(metric);
}

void MetricsCollector::unregisterManagedMetric(BaseManagedMetric& metric)
{
	IceUtil::Mutex::Lock lock(managedMetricsMutex_);
	managedMetrics_.erase(MetricsList::s_iterator_to(metric));
}

void MetricsCollector::collect()
{
	IceUtil::Mutex::Lock lock(managedMetricsMutex_); // prevent any modification of the list
	Metric::SampleTimestamp defaultTimestamp = Metric::currentTimestamp();
	for(MetricsList::iterator metric = managedMetrics_.begin(), endMetric = managedMetrics_.end();
		metric != endMetric; ++metric)
	{
		metric->conditionalReportSample(defaultTimestamp);
	}
}

} /* namespace mon */
} /* namespace acq */
} /* namespace mdcs */
