/**
 * @file FileRunRepository.cpp
 * @date 25 juin 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "FileRunRepository.h"
#include <utl/Logging.h>
#include <utl/Construct.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/date_time/c_local_time_adjustor.hpp>
#include <fstream>
#include <sstream>
#include <iomanip>

namespace mdcs
{
namespace ctrl
{

using namespace acq;

// =============================================================================
// mdcs::ctrl::FileRunRepository::Params class methods body
// -----------------------------------------------------------------------------

FileRunRepository::Params::~Params()
{
}

RunRepositoryPtr FileRunRepository::Params::createPolicy(acq::ServerApplication& /*app*/) const
{
	return RunRepositoryPtr(new FileRunRepository(*this));
}

REGISTER_FACTORY(RunRepositoryFactory, "FileRunRepository", ::utl::Construct<FileRunRepository::Params*>())

// =============================================================================
// mdcs::ctrl::FileRunRepository::Params class methods body
// -----------------------------------------------------------------------------

FileRunRepository::FileRunRepository(const Params& p)
	: params(p)
{
}

FileRunRepository::~FileRunRepository()
{
}

RunInfoPtr FileRunRepository::createRunInfo()
{
	std::string indexFileName(params.directory);
	indexFileName += '/';
	indexFileName += params.siteName;
	indexFileName += "-run_id.txt";
	RunId runId = 0;
	{
		std::ifstream indexFile(indexFileName.c_str());
		if(indexFile.good())
		{
			indexFile >> runId;
		}
	}
	{
		std::ofstream indexFile(indexFileName.c_str());
		indexFile << runId + 1;
	}
	RunTimestamp timestamp = boost::posix_time::second_clock::universal_time();
	static std::locale loc(std::cout.getloc(), new boost::posix_time::time_facet("%Y%m%d-%H%M%S"));

	std::ostringstream fileStem;
	fileStem.imbue(loc);
	fileStem << params.siteName << '-'
		<< std::setw(params.seqDigitCount) << std::setfill('0') << runId
		<< '-' << boost::date_time::c_local_adjustor<RunTimestamp>::utc_to_local(timestamp);

	return RunInfoPtr(new RunInfo(params.siteName, runId, timestamp, fileStem.str()));
}

void FileRunRepository::storeRun(RunInfo& info, const RunConfigList& configs)
{
	LOG_INFO() << "Storing run " << info.runSite() << ' ' << info.runId();
	for(RunConfigList::const_iterator cfg = configs.begin(), endCfg = configs.end(); cfg != endCfg; ++cfg)
	{
		std::ostringstream configFileName;
		configFileName
			<< params.directory
			<< '/'
			<< params.cfgFilePrefix
			<< info.fileStem()
			<< '-'
			<< cfg->first
			<< ".xcfg";
		LOG_INFO() << "Creating file: " << configFileName.str();
		std::ofstream configFile(configFileName.str().c_str());
		if(configFile.fail())
		{
			throw std::runtime_error("failed creating config file");
		}
		configFile << cfg->second;
		//TODO: maybe we should use a config specific file storage method instead.
	}
}

std::string FileRunRepository::loadConfig(const std::string& configRef)
{
	//TODO: if configRef is an URL, we should load the file from filesystem
	return configRef; // considering configRef is not an URL but a full config, so we just return it.
}

} /* namespace ctrl */
} /* namespace mdcs */
