/* =============================================================================
 * RunControlApplication.cpp
 * -----------------------------------------------------------------------------
 * class mdcs::ctrl::RunControlApplication
 * Created on: 28 mars 2013 at Irfu/Sedi/Lilas, CEA Saclay, F-91191, France
 * Author: fchateau
 * -----------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * -----------------------------------------------------------------------------
 * Contributors: fchateau
 * -----------------------------------------------------------------------------
 * This file is part of the Mordicus software project.
 * -----------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding
 * by the rules of distribution of free software. You can use, modify and/or
 * redistribute the software under the terms of the CeCILL license as circulated
 * by CEA, CNRS and INRIA at the following URL: "http://www.cecill.info".
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability. In this
 * respect, the user's attention is drawn to the risks associated with loading,
 * using, modifying and/or developing or reproducing the software by the user in
 * light of its specific status of free software, that may mean that it is
 * complicated to manipulate, and that also therefore means that it is reserved
 * for developers and experienced professionals having in-depth computer
 * knowledge. Users are therefore encouraged to load and test the software's
 * suitability as regards their requirements in conditions enabling the security
 * of their systems and/or data to be ensured and, more generally, to use and
 * operate it in the same conditions as regards security.
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * -----------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for
 * commercial purposes. For this you will need to negotiate a specific contract
 * with a legal representative of CEA.
 * =============================================================================
 */

#include "RunControlApplication.h"
#include "AbstractRunRepository.h"
#include <mdcs/acq/mon/MetricsCollector.h>
#include <mdcs/acq/ServerApplication.h>
#include <mdcs/acq/ApplicationLoader.h>
#include <mdcs/acq/RunInfo.h>
#include <mdcs/cfg/AbstractConfig.h>
#include <mdcs/cfg/Constants.h>
#include <dhsm/EngineServant.h>
#include <dhsm/MasterEngine.h>
#include <dhsm/CompositeEngine.h>
#include <dhsm/StateMachine.h>
#include <utl/Factory.ipp>
#include <utl/Singleton.ipp>
#include <utl/Construct.hpp>

namespace mdcs
{
namespace ctrl
{
using namespace dhsm;

// =============================================================================
// Private classes declarations
// -----------------------------------------------------------------------------

class RunControlApplication::BranchEngineCreator : public acq::DistributedEngineCreator
{
public:
	std::auto_ptr<Engine> createSmEngine(acq::StateMachineServer& server) const;
	std::auto_ptr<EngineServant> createSmServant(acq::StateMachineServer& server) const;
};

class RunControlApplication::RootEngineCreator : public acq::DistributedEngineCreator
{
public:
	std::auto_ptr<Engine> createSmEngine(acq::StateMachineServer& server) const;
	std::auto_ptr<EngineServant> createSmServant(acq::StateMachineServer& server) const;
};

class RunControlApplication::RootEngineServant : public EngineServant
{
public:
	~RootEngineServant();

protected:
	int stepEvent(const std::string& event, const Ice::Current&);
	int stepStringMsg(const std::string& event, const std::string& param, const Ice::Current&);
	int stepIntMsg(const std::string& event, int param, const Ice::Current&);
};

// =============================================================================
// mdcs::ctrl::RunControlApplication class methods body
// -----------------------------------------------------------------------------

RunControlApplication::RunControlApplication(const Params& params, acq::ApplicationLoader& loader)
	: ServerApplication(params,
		loader,
		BranchEngineCreator(),
		cfg::PROCESSOR_CTRL_SERVANT,
		"MDaqCtrlStateMachine"),
	compositeServer_(
		RootEngineCreator(),
		cfg::RUNCONTROL_CTRL_SERVANT,
		cfg::RUNCONTROL_STATEMACHINE,
		server().adapter()),
	configQueue_(new RunConfigQueue),
	runInfoMetric_(collector().createMetric("RunControl.runInfo"))
{
	resetCompositeServer();
	createTopContext(compositeServer());

	Ice::CommunicatorPtr ic = loader_.communicator();

	LOG_DEBUG() << "Creating RunConfig adapter";
	configAdapter_ = ic->createObjectAdapter("RunConfig");
	LOG_DEBUG() << "Adding ConfigQueue servant";
	configAdapter_->add(configQueue_, ic->stringToIdentity("ConfigQueue"));
	LOG_DEBUG() << "Activating adapter";
	configAdapter_->activate();
	LOG_INFO() << "RunConfig adapter ready";
}

RunControlApplication::~RunControlApplication()
{
}

AbstractRunRepository& RunControlApplication::runRepository()
{
	return dynamic_cast<Connected&>(*compositeServer_.engine()->stateMachine()->currentStateData()).runRepository();
}

void RunControlApplication::resetCompositeServer()
{
	compositeServer_.engine<CompositeEngine>().clearEngines();
	compositeServer_.engine<CompositeEngine>().appendEngine(server().engine());
}

std::auto_ptr<dhsm::StateMachine> RunControlApplication::createStateMachine()
{
	StateMachinePtr acqSm = acq::createCtrlStateMachine();
	State& connected = acqSm->findState("CONNECTED");
	connected.setStateDataType<Connected>();
	connected.enterActions().add(&Connected::configure);
	connected.enterActions().add(&ConfigHolder::setConfig);

	State& configured = connected.stateMachine().findState("CONFIGURED");
	configured.setStateDataType<Configured>();
	configured.enterActions().add(&Configured::configure);
	configured.enterActions().add(&ConfigHolder::setConfig);

	State& started = configured.stateMachine().findState("STARTED");
	started.setStateDataType<Started>();
	started.enterActions().add(&Started::recordRun);
	return acqSm;
}

static dhsm::StateMachine* stateMachineFactoryWrapper()
{
	return RunControlApplication::createStateMachine().release();
}

REGISTER_FACTORY(acq::StateMachineFactory, cfg::RUNCONTROL_STATEMACHINE, stateMachineFactoryWrapper)
REGISTER_FACTORY(acq::ApplicationFactory, cfg::RUNCONTROL_APPNAME, utl::Construct<RunControlApplication::Params*>())

// =============================================================================
// mdcs::ctrl::RunControlApplication::SubEngineCreator class methods body
// -----------------------------------------------------------------------------

EnginePtr RunControlApplication::BranchEngineCreator::createSmEngine(acq::StateMachineServer& server) const
{
	return EnginePtr(new MasterEngine(server.adapter()->getCommunicator()));
}

std::auto_ptr<EngineServant> RunControlApplication::BranchEngineCreator::createSmServant(acq::StateMachineServer& /*server*/) const
{
	return std::auto_ptr<EngineServant>(new EngineServant);
}

// =============================================================================
// mdcs::ctrl::RunControlApplication::MainEngineCreator class methods body
// -----------------------------------------------------------------------------

EnginePtr RunControlApplication::RootEngineCreator::createSmEngine(acq::StateMachineServer& /*server*/) const
{
	EnginePtr ret;
	CompositeEngine* engine = new CompositeEngine;
	ret.reset(engine);
	engine->setEventSequence("connect",    true);
	engine->setEventSequence("configure",  true);
	engine->setEventSequence("start",      true);
	engine->setEventSequence("stop",       false);
	engine->setEventSequence("reset",      false);
	engine->setEventSequence("disconnect", false);
	return ret;
}
std::auto_ptr<EngineServant> RunControlApplication::RootEngineCreator::createSmServant(acq::StateMachineServer& /*server*/) const
{
	return std::auto_ptr<EngineServant>(new RootEngineServant);
}

// =============================================================================
// mdcs::ctrl::RunControlApplication::Connected class methods body
// -----------------------------------------------------------------------------

RunControlApplication::Connected::Connected(const StateData& base)
	: ConfiguredState(base)
{
}

RunControlApplication::Connected::~Connected()
{
	application().server().engine<MasterEngine>().disconnectAllServants();
	application<RunControlApplication>().resetCompositeServer();
}

AbstractRunRepository& RunControlApplication::Connected::runRepository()
{
	return *runRepository_;
}

acq::StateMachineServer& RunControlApplication::Connected::addStateMachine(const std::string& id, const std::string& nameInFactory)
{
	acq::StateMachineServer& server = *stateMachines_.insert(id,
		std::auto_ptr<acq::StateMachineServer>(new acq::StateMachineServer(
			BranchEngineCreator(),
			id,
			nameInFactory,
			application().server().adapter()))
		).first->second;
	application().createTopContext(server);
	server.engine<MasterEngine>().start();
	application<RunControlApplication>().compositeServer().engine<CompositeEngine>().appendEngine(server.engine());
	return server;
}

void RunControlApplication::Connected::connectServant(const acq::StateMachineServer& server, const std::string& proxy)
{
	server.engine<MasterEngine>().connectServant(proxy, true);
}

void RunControlApplication::Connected::apply()
{
	//TODO: move connection establishment from init method to here.
	runRepository_.reset(runRepositoryParams_->createPolicy(application()).release());
}

// =============================================================================
// mdcs::ctrl::RunControlApplication::Configured class methods body
// -----------------------------------------------------------------------------

RunControlApplication::Configured::Configured(const StateData& base)
	: ConfiguredState(base)
{
}

void RunControlApplication::Configured::apply()
{

}

// =============================================================================
// mdcs::ctrl::RunControlApplication::Started class methods body
// -----------------------------------------------------------------------------

RunControlApplication::Started::Started(const StateData& base)
	: ApplicationState(base)
{
}

RunControlApplication::Connected& RunControlApplication::Started::connectedCtx()
{
	return dynamic_cast<Connected&>(*context()->context());
}

RunControlApplication::Configured& RunControlApplication::Started::configuredCtx()
{
	return dynamic_cast<Configured&>(*context());
}

void RunControlApplication::Started::recordRun(const std::string& runCfg)
{
	acq::RunInfo runInfo;
	cfg::ConfigPtr config = cfg::ConfigFactory::instance().create(runCfg);
	cfg::Configurator::instance().configure(runInfo, *config);

	RunConfigList configs;
	configs["daq"] = configuredCtx().config();
	configs["connections"] = connectedCtx().config();

	connectedCtx().runRepository().storeRun(runInfo, configs);

	reportNewRun(runInfo);
}

void RunControlApplication::Started::reportNewRun(const acq::RunInfo& runInfo)
{
	std::vector<std::string> runMetrics;
	runMetrics.resize(3);
	runMetrics[0] = runInfo.runSite();
	runMetrics[1] = boost::lexical_cast<std::string>(runInfo.runId());
	runMetrics[2] = boost::lexical_cast<std::string>(runInfo.runTimestamp());
	application<RunControlApplication>().runInfoMetric().reportSample(runMetrics, acq::mon::Metric::currentTimestamp());
}

// =============================================================================
// mdcs::ctrl::RunControlApplication::ConfigHolder class methods body
// -----------------------------------------------------------------------------

void RunControlApplication::ConfigHolder::setConfig(const std::string& cfg)
{
	config_ = cfg;
}

const std::string& RunControlApplication::ConfigHolder::config() const
{
	return config_;
}

// =============================================================================
// mdcs::ctrl::RunControlApplication::EngineServant class methods body
// -----------------------------------------------------------------------------

RunControlApplication::RootEngineServant::~RootEngineServant()
{
}

int RunControlApplication::RootEngineServant::stepEvent(const std::string& event, const Ice::Current&)
{
	try
	{
		RunControlApplication& rcs = engine().topContext<ServerApplication::SmContext>().application<RunControlApplication>();
		std::string configStr;
		if(event == "start")
		{
			acq::RunInfoPtr run = rcs.runRepository().createRunInfo();
			cfg::ConfigPtr config = cfg::ConfigFactory::instance().create(std::string()); // create empty config
			cfg::Configurator::instance().configure(*config, *run); // init the config with a runInfo
			configStr = config->encode();

			LOG_DEBUG() << "Run config: " << configStr;
		}
		else
		{
			if(!rcs.configQueue().empty(event))
			{
				configStr = rcs.configQueue().popConfig(event);
			}
		}

		//TODO: loadConfig requires the RunRepository to be set: we have to set it through an "application" config.
//		std::string config = rcs.runRepository().loadConfig(configRef);

		if(configStr.empty())
		{
			return engine().step(Event(event.c_str()));
		}
		else
		{
			return engine().step(Message<std::string>(event.c_str(), configStr));
		}
	}
	catch(...)
	{
		handleExceptions();
		throw;
	}
}

int RunControlApplication::RootEngineServant::stepStringMsg(const std::string& /*event*/, const std::string& /*param*/, const Ice::Current&)
{
	throw ServantCriticalError("Parameters are not allowed on Run Control events");
}

int RunControlApplication::RootEngineServant::stepIntMsg(const std::string& /*event*/, int /*param*/, const Ice::Current&)
{
	throw ServantCriticalError("Parameters are not allowed on Run Control events");
}

} /* namespace ctrl */
} /* namespace mdcs */
