/**
 * @file IceStormMetricsMonitor.cpp
 * @date 3 oct. 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "IceStormMetricsMonitor.h"
#include "MetricTopic.h"
#include <mdcs/acq/mon/IceStormMetricsCollector.h>
#include <utl/Logging.h>
#include <boost/date_time/posix_time/posix_time_duration.hpp>
#include ICE_HEADER(ObjectAdapter.h)

namespace mdcs
{
namespace ctrl
{
namespace mon
{
using namespace acq::mon;

// =============================================================================
// mdcs::ctrl::mon::IceStormMetricsMonitor::MetricTopicImpl class methods body
// -----------------------------------------------------------------------------

class IceStormMetricsMonitor::MetricTopicImpl : public MetricTopic
{
public:
	MetricTopicImpl(IceStormMetric& metric)
		: metric_(metric)
	{
	}
	~MetricTopicImpl()
	{
	}
	static Metric::SampleTimestamp fromTimestamp(Ice::Long timestamp)
	{
		return IceStormMetricsCollector::IceStormMetric::iceTimestampEpoch() + boost::posix_time::microseconds(timestamp);
	}
	void intScalar(Ice::Long sample, Ice::Long timestamp, const Ice::Current&)
	{
		LOG_DEBUG() << "Long: " << sample << " ts=" << timestamp;
		metric_.callback().reportSample(sample, fromTimestamp(timestamp));
	}
	void fltScalar(Ice::Double sample, Ice::Long timestamp, const Ice::Current&)
	{
		LOG_DEBUG() << "Dble: " << sample << " ts=" << timestamp;
		metric_.callback().reportSample(sample, fromTimestamp(timestamp));
	}
	void strScalar(const std::string& sample, Ice::Long timestamp, const Ice::Current&)
	{
		LOG_DEBUG() << "Str: " << sample << " ts=" << timestamp;
		metric_.callback().reportSample(sample, fromTimestamp(timestamp));
	}
	void intVector(const IntVec& sample, Ice::Long timestamp, const Ice::Current&)
	{
		metric_.callback().reportSample(sample, fromTimestamp(timestamp));
	}
	void fltVector(const FltVec& sample, Ice::Long timestamp, const Ice::Current&)
	{
		metric_.callback().reportSample(sample, fromTimestamp(timestamp));
	}
	void strVector(const StrVec& sample, Ice::Long timestamp, const Ice::Current&)
	{
		for(size_t i = 0; i < sample.size(); ++i)
		{
			LOG_DEBUG() << "Str[" << i << "]: " << sample[i] << " ts=" << timestamp;
		}
		metric_.callback().reportSample(sample, fromTimestamp(timestamp));
	}
	void customData(const std::pair<const Ice::Byte*, const Ice::Byte*>& sample, Ice::Long timestamp, const Ice::Current&)
	{
		metric_.callback().reportSample(utl::io::ConstBuffer(sample.first, sample.second - sample.first), fromTimestamp(timestamp));
	}
private:
	IceStormMetric& metric_;
};

// =============================================================================
// mdcs::ctrl::mon::IceStormMetricsMonitor class methods body
// -----------------------------------------------------------------------------

IceStormMetricsMonitor::IceStormMetricsMonitor(const Ice::CommunicatorPtr& ic)
{
	LOG_INFO() << "Connecting to TopicManager...";
	topicManager_ = IceStorm::TopicManagerPrx::checkedCast(ic->stringToProxy("IceStorm/TopicManager"));

	LOG_INFO() << "Creating monitoring adapter...";
	adapter_ = ic->createObjectAdapterWithEndpoints("MetricsMonitor", "tcp");

	LOG_INFO() << "Activating monitoring adapter...";
	adapter_->activate();
}

IceStormMetricsMonitor::~IceStormMetricsMonitor()
{
	LOG_INFO() << "Destroying the monitoring adapter...";
	adapter_->destroy();
}

MonitoredMetricPtr IceStormMetricsMonitor::subscribeMetric(const std::string& id, acq::mon::AbstractMetric& callback)
{
	return MonitoredMetricPtr(new IceStormMetric(*this, id, callback));
}

// =============================================================================
// mdcs::ctrl::mon::IceStormMetricsMonitor::IceStormMetric class methods body
// -----------------------------------------------------------------------------

IceStormMetricsMonitor::IceStormMetric::IceStormMetric(
		IceStormMetricsMonitor& monitor,
		const std::string& topicName,
		acq::mon::AbstractMetric& callback
)
	: MonitoredMetric(monitor, topicName, callback)
{
	subscribe();
}

IceStormMetricsMonitor::IceStormMetric::~IceStormMetric()
{
	if(subscriber_)
	{
		LOG_INFO() << "Unsubscribing from topic: " << topic_->getName();
		topic_->unsubscribe(subscriber_);

		LOG_INFO() << "Removing the metric topic servant from adapter";
		monitor().adapter()->remove(subscriber_->ice_getIdentity());
	}
}

void IceStormMetricsMonitor::IceStormMetric::subscribe()
{
	try
	{
		LOG_INFO() << "Retrieving topic " << id_;
		topic_ = monitor().topicManager()->retrieve(id_);

		LOG_INFO() << "Creating a metric topic servant for " << id_;
		MetricTopicPtr topicMonitor = new MetricTopicImpl(*this);
		subscriber_ = monitor().adapter()->addWithUUID(topicMonitor)->ice_oneway();

		LOG_INFO() << "Subscribing to topic " << id_;
		IceStorm::QoS qos;
		topic_->subscribeAndGetPublisher(qos, subscriber_);

		metaCallback_.reset(); // Subscription successful, destroying the metaCallback if existing.
	}
	catch(const IceStorm::NoSuchTopic&)
	{
		if(id_ != Metric::META_METRIC_ID)
		{
			LOG_INFO() << "Topic not found! monitoring the meta-metric to wait for its creation...";
			metaCallback_.reset(new MetaCallback(*this));
		}
		else
		{
			// TODO: throw a proper exception
			throw std::runtime_error("Cannot find the meta-metric topic!");
		}
	}
}

} /* namespace mon */
} /* namespace ctrl */
} /* namespace mdcs */
