/**
 * @file ConfigSetMaster.cpp
 * @date 4 sept. 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "ConfigSetMaster.h"
#include <mdcs/ctrl/IRunConfigQueue.h>
#include <CCfg/Model/ConfigSetModel.h>
#include <utl/Logging.h>
#include <fstream>
#include <stdexcept>

namespace mdcs
{
namespace gui
{
using namespace CCfg::Model;

// =============================================================================
// mdcs::gui::ConfigSetMaster class methods body
// -----------------------------------------------------------------------------

ConfigSetMaster::ConfigSetMaster(Ice::CommunicatorPtr ic, ConfigSetModel* model, QObject* parent)
	: QObject(parent), model_(model),
	  configQueue_(ctrl::IRunConfigQueuePrx::checkedCast(ic->stringToProxy("ConfigQueue")))
{
	connect(model, SIGNAL(repositoryChanged()), this, SLOT(onRepositoryChanged()));
	connect(model, SIGNAL(configModified(const QString&)),
			this, SLOT(onConfigSlotUpdated(const QString&)));
	connect(model, SIGNAL(configBindingChanged(const QString&, const QString&, const QString&)),
			this, SLOT(onConfigSlotUpdated(const QString&)));
}

ConfigSetMaster::~ConfigSetMaster()
{
}

void ConfigSetMaster::onRepositoryChanged()
{
	configQueue_->clearAll();
}

void ConfigSetMaster::onConfigSlotUpdated(const QString& configSlotName)
{
	ConfigSetModel::ConfigSlot* configSlot = model_->searchByName(configSlotName);
	if(configSlot != 0)
	{
		updateQueue(configSlot->group());
	}
	else
	{
		LOG_WARN() << "cannot find config slot: " << qPrintable(configSlotName);
	}
}

void ConfigSetMaster::updateQueue(const std::string& queueName)
{
	LOG_INFO() << "updating config queue: " << queueName;
	configQueue_->pushConfig(queueName, generateConfig(queueName));
}

std::string ConfigSetMaster::generateConfig(const std::string& queueName)
{
	try
	{
		const ConfigSetModel::ConfigSlotList& configSlots = model_->configSlots();
		for(size_t i = 0; i < configSlots.size(); ++i)
		{
			// TODO: handle multiple configurations for a single queue (=> merging)
			if(configSlots[i].group() == queueName)
			{
				std::string url = configSlots[i].config().url().toStdString();
				if(model_->isDatabaseRepository())
				{
					return url;
				}
				else
				{
					return loadConfig(url);
				}
			}
		}
		LOG_ERROR() << "cannot find config queue named: " << queueName;
	}
	catch(const std::exception& e) //TODO: maybe catch a subclass
	{
		LOG_ERROR() << e.what();
	}
	return std::string();
}

std::string ConfigSetMaster::loadConfig(const std::string& url)
{
    std::ifstream configFile(url.c_str());
	if(configFile.fail())
	{
		throw std::runtime_error("unable to open specified configuration");
	}
	return std::string((std::istreambuf_iterator<char>(configFile)), std::istreambuf_iterator<char>());
}

} /* namespace gui */
} /* namespace mdcs */
