/**
 * @file RunInfoWidget.cpp
 * @date 3 oct. 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "RunInfoWidget.h"
#include <mdcs/gui/mon/RunInfoWidget.ui.h>
#include <mdcs/ctrl/mon/MonitoredMetric.h>
#include <mdcs/ctrl/mon/MetricsMonitor.h>
#include <boost/date_time/posix_time/posix_time_io.hpp>
#include <boost/lexical_cast.hpp>

namespace mdcs
{
namespace gui
{
namespace mon
{

// =============================================================================
// mdcs::gui::mon::RunInfoWidget class methods body
// -----------------------------------------------------------------------------

RunInfoWidget::RunInfoWidget(QWidget* parent)
	: Utils::Ui::DesignedWidget<Ui_RunInfoWidget, QWidget>(parent),
	  runInfoSignal_(this, SLOT(updateRunInfo(const QStringList&, const SampleTimestamp&))),
	  eventCountSignal_(this, SLOT(updateEventCount(qint64, const SampleTimestamp&))),
	  fileSizeSignal_(this, SLOT(updateFileSize(qint64)))
{
}

RunInfoWidget::~RunInfoWidget()
{
}

void RunInfoWidget::setMonitor(ctrl::mon::MetricsMonitor& monitor)
{
	runInfoMetric_    = monitor.subscribeMetric("RunControl.runInfo", runInfoSignal_);

	// TODO: will not work with multiple data storage storing different events (would require aggregation).
	eventCountMetric_ = monitor.subscribeMetric("FileStorage.eventsReceived", eventCountSignal_);
	fileSizeMetric_   = monitor.subscribeMetric("FileStorage.bytesWritten", fileSizeSignal_);
}

void RunInfoWidget::updateRunInfo(const QStringList& runInfo, const SampleTimestamp& time)
{
	if(runInfo.size() >= 3)
	{
		ui_->lineEditExperiment->setText(runInfo[0]);
		ui_->lineEditNumber->setText(runInfo[1]);
		runStart = boost::lexical_cast<SampleTimestamp>(runInfo[2].toStdString());
		static const SampleTimestamp epoch(boost::gregorian::date(1970, 1, 1));
		QDateTime dt = QDateTime::fromTime_t((runStart - epoch).total_seconds()).toLocalTime();
		ui_->dateEdit->setDateTime(dt);
		ui_->timeEdit->setDateTime(dt);

		updateEventCount(0, time);
		updateFileSize(0);
	}
}

void RunInfoWidget::updateEventCount(qint64 count, const SampleTimestamp& time)
{
	ui_->lineEditEventCount->setText(QString::number(count));
	updateDuration(time);
}

void RunInfoWidget::updateFileSize(qint64 size)
{
	QString sizeStr = QString::number(size / (1024.0 * 1024.0), 'f', 2);
	sizeStr += " MiB";
	ui_->lineEditFileSize->setText(sizeStr);
}

void RunInfoWidget::updateDuration(const SampleTimestamp& time)
{
	if(not runStart.is_special())
	{
		int16_t elapsed = (time - runStart).total_seconds();
		int16_t days = elapsed / (3600 * 24);
		int16_t secsInDay = elapsed % (3600 * 24);
		QLatin1Char fillChar('0');
		ui_->lineEditDuration->setText(QString("%1d %2:%3:%4")
				.arg(days)
				.arg(secsInDay / 3600, 2, 10, fillChar)
				.arg((secsInDay / 60) % 60, 2, 10, fillChar)
				.arg(secsInDay % 60, 2, 10, fillChar));
	}
}

} /* namespace mon */
} /* namespace gui */
} /* namespace mdcs */
