/**
 * @file MasterStateMachineModel.cpp
 * @date 5 fev 2014
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "MasterStateMachineModel.h"
#include "StateMachineWidget.h"
#include <mdcs/cfg/Constants.h>
#include <dhsm/MasterEngine.h>
#include <utl/Logging.h>

namespace mdcs
{
namespace gui
{
namespace sm
{
using namespace dhsm;

// =============================================================================
// mdcs::gui::sm::MasterStateMachineModel class methods body
// -----------------------------------------------------------------------------

MasterStateMachineModel::MasterStateMachineModel(std::auto_ptr<StateMachine> sm, const Ice::CommunicatorPtr& ic)
	: engine_(new MasterEngine),
	  smThread_(new Thread(*this))
{
	MasterEngine& engine = static_cast<MasterEngine&>(*engine_);
	engine.setStateMachine(sm);
	LOG_INFO() << "engine init";
	engine.init(ic);
	LOG_INFO() << "connect run control";
	engine.connectServant(cfg::RUNCONTROL_CTRL_SERVANT);
	LOG_INFO() << "activate statemachine";
	engine.start();
}

MasterStateMachineModel::~MasterStateMachineModel()
{
}

void MasterStateMachineModel::doGetReady()
{
	smThread_->start();
}

void MasterStateMachineModel::onStepEvent(const QString& event, const QString& param)
{
	emit stepEvent(event, param);
}

void MasterStateMachineModel::onReinit()
{
	emit reinit();
}

// =============================================================================
// mdcs::gui::sm::MasterStateMachineModel::AsyncModel class methods body
// -----------------------------------------------------------------------------

MasterStateMachineModel::AsyncModel::AsyncModel(StateMachineModel& syncModel)
	: engine_(syncModel.engine())
{
	connect(this,       SIGNAL(ready()),
	        &syncModel, SIGNAL(ready()));
	connect(this,       SIGNAL(beginTransition(const QString&)),
			&syncModel, SIGNAL(beginTransition(const QString&)));
	connect(this,       SIGNAL(endTransition(mdcs::gui::sm::StateMachineModel::TransitionStatus, const QString&)),
	        &syncModel, SIGNAL(endTransition(mdcs::gui::sm::StateMachineModel::TransitionStatus, const QString&)));
	connect(this,       SIGNAL(beginReinit()),
	        &syncModel, SIGNAL(beginReinit()));
	connect(this,       SIGNAL(endReinit()),
	        &syncModel, SIGNAL(endReinit()));

	connect(&syncModel, SIGNAL(stepEvent(const QString&, const QString&)),
	        this,       SLOT(onStepEvent(const QString&, const QString&)));
	connect(&syncModel, SIGNAL(reinit()),
	        this,       SLOT(onReinit()));
}

MasterStateMachineModel::AsyncModel::~AsyncModel()
{
}

void MasterStateMachineModel::AsyncModel::doGetReady()
{
	emit ready();
}

void MasterStateMachineModel::AsyncModel::onStepEvent(const QString& event, const QString& param)
{
	emit beginTransition(event);
	try
	{
		Engine::ProcessingResult result =
			param.isEmpty() ?
				engine_.step(Event(event.toAscii())) :
				engine_.step(Message<std::string>(event.toAscii(), param.toStdString()));
		emit endTransition(static_cast<TransitionStatus>(result), QString());
	}
	catch(const std::exception& e)
	{
		emit endTransition(TransitionException, QString::fromStdString(e.what()));
	}
}

void MasterStateMachineModel::AsyncModel::onReinit()
{
	emit beginReinit();
	try
	{
		engine_.reinitiate(engine_.currentStateMachine());
	}
	catch(const std::exception& e)
	{
		LOG_ERROR() << "reinit failed: " << e.what();
	}
	emit endReinit();
}

// =============================================================================
// mdcs::gui::sm::MasterStateMachineModel::Thread class methods body
// -----------------------------------------------------------------------------

MasterStateMachineModel::Thread::Thread(StateMachineModel& syncModel)
	: syncModel_(syncModel)
{
}

MasterStateMachineModel::Thread::~Thread()
{
	stop();
}

void MasterStateMachineModel::Thread::run()
{
	MasterStateMachineModel::AsyncModel asyncModel(syncModel_);
	asyncModel.doGetReady();
	exec();
}

void MasterStateMachineModel::Thread::stop()
{
	this->exit(0);
	if(!this->wait(3000))
	{
		LOG_ERROR() << "StateMachine thread is not responding";
		this->terminate();
		LOG_WARN() << "StateMachine thread killed";
	}
}


} /* namespace sm */
} /* namespace gui */
} /* namespace mdcs */
