/**
 * @file StateMachineWidget.cpp
 * @date 22 mai 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "StateMachineWidget.h"
#include "StateMachineModel.h"
#include <mdcs/gui/sm/StateMachine.ui.h>
#include <dhsm/MasterEngine.h>
#include <dhsm/State.h>
#include <QMessageBox>
#include <QCloseEvent>
#include <QTimerEvent>

namespace mdcs
{
namespace gui
{
namespace sm
{
using namespace dhsm;

// =============================================================================
// Private classes declarations
// -----------------------------------------------------------------------------

/**
 * This class flashes allows to flash a button.
 */
class StateMachineWidget::ButtonFlasher : public QObject
{
public:
	ButtonFlasher(QPushButton* button);
	~ButtonFlasher();

	static const int FLASH_BUTTON_PERIOD = 500; // In milliseconds

private:
	void flash();
	void timerEvent(QTimerEvent* event);

	QPushButton* flashingButton;
	bool flashState;
	QPalette defaultPalette;
	QPalette flashPalette;
	int timer;
};

// =============================================================================
// mdcs::gui::sm::StateMachineWidget::ButtonFlasher class methods body
// -----------------------------------------------------------------------------

StateMachineWidget::ButtonFlasher::ButtonFlasher(QPushButton* button)
	: flashingButton(button),
	  flashState(false),
	  defaultPalette(flashingButton->palette()),
	  flashPalette(defaultPalette)
{
	flashPalette.setColor(QPalette::Button,  Qt::red);
	flashingButton->setAutoFillBackground(true);
	timer = startTimer(FLASH_BUTTON_PERIOD);
}

StateMachineWidget::ButtonFlasher::~ButtonFlasher()
{
	killTimer(timer);
	flashingButton->setPalette(defaultPalette);
	flashingButton->setAutoFillBackground(false);

	// restore normal look
	flashState = true;
	flash();
}

void StateMachineWidget::ButtonFlasher::flash()
{
	flashState = !flashState;
	flashingButton->setPalette(flashState ? flashPalette : defaultPalette);
	flashingButton->update();
}

void StateMachineWidget::ButtonFlasher::timerEvent(QTimerEvent* event)
{
	if(event->timerId() == timer)
	{
		flash();
	}
}

// =============================================================================
// mdcs::gui::sm::StateMachineWidget class methods body
// -----------------------------------------------------------------------------

StateMachineWidget::StateMachineWidget(QWidget* parent)
	: Utils::Ui::DesignedWidget<Ui_StateMachine, QWidget>(parent), model_(0)
{
	buttons.clear();
	ui_->btConnect->setCheckedText("disconnect");
	ui_->btConnect->setUncheckedText("connect");
	ui_->btConfigure->setCheckedText("reset");
	ui_->btConfigure->setUncheckedText("configure");
	ui_->btStart->setCheckedText("stop");
	ui_->btStart->setUncheckedText("start");
	buttons.push_back(ui_->btConnect);
	buttons.push_back(ui_->btConfigure);
	buttons.push_back(ui_->btStart);

	for(size_t i = 0; i < buttons.size(); ++i)
	{
		connect(buttons[i], SIGNAL(clicked(bool)), this, SLOT(onButtonClicked(bool)));
	}
}

StateMachineWidget::~StateMachineWidget()
{
}

void StateMachineWidget::setModel(StateMachineModel* model)
{
	model_ = model;

	connect(model, SIGNAL(ready()),
			this,  SLOT(onReady()));
	connect(model, SIGNAL(beginTransition(const QString&)),
			this,  SLOT(onBeginTransition(const QString&)));
	connect(model, SIGNAL(endTransition(mdcs::gui::sm::StateMachineModel::TransitionStatus, const QString&)),
	        this,  SLOT(onEndTransition(mdcs::gui::sm::StateMachineModel::TransitionStatus, const QString&)));
	connect(model, SIGNAL(beginReinit()),
	        this,  SLOT(onBeginReinit()));
	connect(model, SIGNAL(endReinit()),
	        this,  SLOT(onEndReinit()));

	connect(this,  SIGNAL(eventBroadcasted(const QString&, const QString&)),
			model, SLOT(onStepEvent(const QString&, const QString&)));

	model_->doGetReady();
}

QPushButton* StateMachineWidget::findButtonForEvent(const std::string& event) const
{
	QString eventStr = QString::fromStdString(event);
	for(size_t i = 0; i < buttons.size(); ++i)
	{
		DualTextButton& button = static_cast<DualTextButton&>(*buttons[i]);
		if(button.checkedText() == eventStr)
		{
			return buttons[i];
		}
		else if(button.uncheckedText() == eventStr)
		{
			return buttons[i];
		}
	}
	return 0;
}

void StateMachineWidget::setEventParam(const std::string& event, const std::string& param)
{
	params[QString::fromStdString(event)] = QString::fromStdString(param);
}

void StateMachineWidget::onReady()
{
	update();
}

void StateMachineWidget::onBeginTransition(const QString& event)
{
	this->setEnabled(false);
	QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));

	QPushButton* evtButton = findButtonForEvent(event.toStdString());
	if(evtButton != 0)
	{
		flashMonitor.reset(new ButtonFlasher(evtButton));
	}
	else
	{
		LOG_DEBUG() << "cannot find button for event!";
	}
}

void StateMachineWidget::onEndTransition(StateMachineModel::TransitionStatus result, const QString& excepMessage)
{
	flashMonitor.reset();
	QApplication::restoreOverrideCursor();
	update();
	this->setEnabled(true);
	if(result == StateMachineModel::TransitionException)
	{
		QMessageBox::warning(this, "Transition error", excepMessage);
		model_->onReinit();
	}
}

void StateMachineWidget::onBeginReinit()
{
	this->setEnabled(false);
	QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
}

void StateMachineWidget::onEndReinit()
{
	QApplication::restoreOverrideCursor();
	update();
	this->setEnabled(true);
}

void StateMachineWidget::onButtonClicked(bool checked)
{
	DualTextButton* sender = static_cast<DualTextButton*>(this->sender());
	QString eventName = checked ? sender->uncheckedText() : sender->checkedText();
	emit eventBroadcasted(eventName, params[eventName]);
}

void StateMachineWidget::update()
{
	if(model_ != 0)
	{
		State* currentState = &model_->engine().currentState();
		ui_->labelState->setText(QString::fromStdString(currentState->label()));
		for(size_t i = 0; i < buttons.size(); ++i)
		{
			DualTextButton& button = static_cast<DualTextButton&>(*buttons[i]);
			State* state = currentState;
			bool shouldBeChecked = false;
			bool isSensibleToEvent = false;
			while(state != 0 and !isSensibleToEvent)
			{
				const State::TransitionMap& transitionMap = state->transitionMap();
				state = state->parentStateMachine()->parentState();

				Event event(button.checkedText().toAscii());
				shouldBeChecked = (transitionMap.find(event) != transitionMap.end());

				if(!shouldBeChecked)
				{
					event.label(button.uncheckedText().toStdString());
					isSensibleToEvent = (transitionMap.find(event) != transitionMap.end());
				}
				else
				{
					isSensibleToEvent = true;
				}
			}
			button.setChecked(shouldBeChecked);
			button.setEnabled(isSensibleToEvent);
		}
	}
}

void StateMachineWidget::closeEvent(QCloseEvent* event)
{
	if(model_ != 0)
	{
		// TODO: maybe there is a cleaner way ?
		if(model_->engine().currentState(0).label() != "DISCONNECTED")
		{
			QMessageBox::StandardButton r =
				QMessageBox::warning(this, tr("Closing"),
					tr("The run control is still connected, want do you want to do ?"),
					QMessageBox::Ignore | QMessageBox::Cancel,
					QMessageBox::Cancel);
			if(r == QMessageBox::Cancel)
			{
				event->ignore();
			}
			else
			{
				QWidget::closeEvent(event);
			}
		}
	}
}

} /* namespace sm */
} /* namespace gui */
} /* namespace mdcs */
