/**
 * @file AbstractRegister.h
 * @created Oct 26, 2012
 * @author anvar
 * ---------------------------------------------------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * ---------------------------------------------------------------------------------------------------------------------
 * Contributors: Shebli Anvar (shebli.anvar@cea.fr)
 * ---------------------------------------------------------------------------------------------------------------------
 * This file is part of the Mordicus software project.
 * ---------------------------------------------------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 * ---------------------------------------------------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 * =====================================================================================================================
 */

#ifndef mdcs_hw_AbstractRegister_h_INCLUDED
#define mdcs_hw_AbstractRegister_h_INCLUDED

#include <typeinfo>
#include <boost/static_assert.hpp>
#include <string>

namespace mdcs
{
namespace hw
{
/**
 * This class is the base class for all register classes. This is to provide for register polymorphism in factories.
 */
class AbstractRegister
{
public:
	AbstractRegister();
	virtual ~AbstractRegister();
	virtual const std::string policyName() const = 0; ///< Return the name of the register access policy used by this register type.
};

/**
 * Template class used for specializing specific init methods. This struct is used instead of a method to allow for
 * partial template specializations. This general definition is invalid, only specialized templates are expected to be
 * valid, hence the use of boost::static_assert in the general-case init method.
 * @tparam O The class of the object to be initialized
 * @tparam C The type of the configuration object used as input to initialize the object.
 */
template<class O, class C>
struct Init
{
	typedef O ObjectType;
	typedef C ConfigType;
	static void init(ObjectType&, const ConfigType&)
	{
		throw std::string("Template struct Init<O,C> must be specialized for O = '") +
				typeid(ObjectType).name() + "' and C = '" + typeid(ConfigType).name() + "'";
		//BOOST_STATIC_ASSERT_MSG(false, "template struct Init must be specialized");
	}
};

/**
 * This template method uses the Init template struct to call the Init::init method. This is ahelper method to allow
 * automatic template parameter type binding detection when calling Init::init.
 * @tparam O The class of the object to be initialized
 * @tparam C The type of the configuration object used as input to initialize the object.
 */
template<class O, class C>
inline void call_init(O& o, const C& c)
{
	Init<O,C>::init(o,c);
}

} // namespace hw
} // namespace mdcs

#endif /* mdcs_hw_AbstractRegister_h_INCLUDED */
