/**
 * @file BaseDevice.cpp
 * @created Oct 26, 2012
 * @author anvar
 * ---------------------------------------------------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * ---------------------------------------------------------------------------------------------------------------------
 * Contributors: Shebli Anvar (shebli.anvar@cea.fr)
 * ---------------------------------------------------------------------------------------------------------------------
 * This file is part of the Mordicus software project.
 * ---------------------------------------------------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 * ---------------------------------------------------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 * =====================================================================================================================
 */

#include "BaseDevice.h"
#include "AbstractRegister.h"

namespace mdcs
{
namespace hw
{

/**
 * @fn BaseDevice::factoryMap()
 * Class method returning the unique instance of the application's map of device factory methods.
 */

/**
 * @fn BaseDevice::deviceMap()
 * Class method returning the unique instance of the application's device map.
 */

/**
 * Destroys BaseDevice by deallocating all its register instances.
 */
BaseDevice::~BaseDevice()
{
	clearRegs();
}

/**
 * Adds specified register to the register set of this device. Ownership of register object is transferred to device.
 * This method is protected: not supposed to be accessed by other than specialized classes.
 * @param reg The register to be added. Has to be allocated on the heap.
 * @return This device, so that register management can be cascaded.
 * @throw NullRegisterException is specified register pointer is null
 */
BaseDevice& BaseDevice::addReg(AbstractRegister* reg)
{
	try
	{
		regList.push_back(reg); // If reg is null, a boost exception is thrown.
		return *this;
	}
	catch (const boost::bad_ptr_container_operation& e)
	{
		throw NullRegisterException(handle);
	}
}

/**
 * Removes all the registers of this device.
 */
void BaseDevice::clearRegs()
{
	regList.clear(); // Deallocation of registers done by boost ptr_list
}

/**
 * Formats a human-readable message for this exception.
 * @param stream The stream to which the formatted output is written.
 */
void NullRegisterException::writeMessage(std::ostream& stream) const
{
	stream << "Access to Null Register in: " << deviceHandle.toString();
}

} // namespace hw
} // namespace mdcs
