/**
 * @file BitField.h
 * @created Aug 28, 2013
 * @author anvar
 * ---------------------------------------------------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * ---------------------------------------------------------------------------------------------------------------------
 * Contributors: Shebli Anvar (shebli.anvar@cea.fr)
 * ---------------------------------------------------------------------------------------------------------------------
 * This file is part of the Mordicus software project.
 * ---------------------------------------------------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 * ---------------------------------------------------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 * =====================================================================================================================
 */
#ifndef mdcs_hw_BitField_h_INCLUDED
#define mdcs_hw_BitField_h_INCLUDED

#include "mdcs/hw/AbstractRegister.h"
#include "utl/Exception.h"
#include "utl/BitFieldHelperVector.hpp"
#include <map>
#include <set>

namespace mdcs
{
namespace hw
{

/**
 * This structure specifies a register bit field. By convention, the least significant bit is numbered 0. For instance,
 * the bits of a 32-bit register are numbered 0 through 31 from the least significant to the most significant.
 */
struct BitField
{
	template < class R >
	class   Operator; // Defined further
	template < typename K >
	class Pool; // Defined further.

	BitField(unsigned pos = 0u, unsigned wid = 1u) : pos(pos), wid(wid) {}
	unsigned pos; ///< Position of the first bit of the bit field.
	unsigned wid; ///< Number of bits (width) of the bit field.
};

/**
 * This template class allows for chained operations on the bit fields of a register, using a cached register data type
 * and actually reading/writing the register before/after the first/last operation has been completed. Avoids repeating
 * for every bit field a possibly costly read/write access to the register itself.
 * @tparam R The class of the register on which this bit field operator operates.
 */
template < class R >
class BitField::Operator
{
public:
	typedef R Register;
	typedef typename Register::DataType DataType;
	typedef typename Register::BitFieldKey BitFieldKey;

public:
	Operator(Register& reg) : reg(reg), dirty(false) { reg.peek(data); /* read reg in cache */ }
	Operator(const Operator& bitFieldOp)
		: reg(bitFieldOp.reg), dirty(bitFieldOp.dirty), data(bitFieldOp.data)
	{
		const_cast<Operator&>(bitFieldOp).dirty = false; // Transfer responsibility for register write-back
	}
	~Operator() { if (dirty) reg.poke(data); /* Write back register with latest "cache" value */ }

	template < typename ValueType >
	Operator& peek(unsigned pos, unsigned wid, ValueType& value)
	{
		LOG_DEBUG() << "Reading value from bit field(" << std::dec << pos << ',' << wid << ')';
		::utl::BitFieldHelper<DataType>::getField(data, pos, wid, value);
		return *this;
	}

	template < typename ValueType >
	Operator& poke(unsigned pos, unsigned wid, const ValueType& value)
	{
		LOG_DEBUG() << "Writing value " << std::hex << value << " to bit field(" << std::dec << pos << ',' << wid << ')';
		::utl::BitFieldHelper<DataType>::setField(data, pos, wid, value);
		dirty = true; // "cache" has been modified => needs eventual register write-back.
		return *this;
	}

	template < typename ValueType >
	Operator& poke(const BitField   &    bf, const ValueType& value) { return poke(bf.pos, bf.wid, value); }

	template < typename ValueType >
	Operator& peek(const BitField   &    bf,       ValueType& value) { return peek(bf.pos, bf.wid, value); }

	template < typename ValueType >
	Operator& poke(const BitFieldKey& bfKey, const ValueType& value) { return poke(reg[bfKey], value); }

	template < typename ValueType >
	Operator& peek(const BitFieldKey& bfKey,       ValueType& value) { return peek(reg[bfKey], value); }

private:
	/// This is meant to forbid the use of the new operator for the BitField::Operator class. The logic of this class
	/// uses the semantics of automatic C++ variables.
	void* operator new (std::size_t size) { return 0; } // Forbids dynamic allocation

protected:
	Register& reg; ///< The register on which this operator operates.
	bool dirty; ///< Set to true if reg has been modified by any operation, false otherwise.
	DataType data; ///< Variable used as data cache for reg.
};

template < typename K >
class BitField::Pool
{
public:
	typedef K KeyType;
	typedef std::map<KeyType, BitField> Map;
	struct  BitFieldNotFoundException; ///< Thrown if access to non-existent bit field is attempted
	typedef std::set<KeyType> KeySet;

	Pool(AbstractRegister* reg = 0) : reg(reg) {}
	Pool& mapBitField(const KeyType& key, const BitField& bitField);
	Pool& clearBitFields();
	size_t countBitFields();
	const BitField& bitField(const KeyType& key) const;
	KeySet keySet();

private:
	AbstractRegister* reg;
	Map reg2bitField;
};

template < typename K >
inline BitField::Pool<K>& BitField::Pool<K>::mapBitField(const KeyType& key, const BitField& bitField)
{
	reg2bitField[key] = bitField;
	return *this;
}

template < typename K >
inline BitField::Pool<K>& BitField::Pool<K>::clearBitFields()
{
	reg2bitField.clear();
	return *this;
}

template < typename K >
inline size_t BitField::Pool<K>::countBitFields()
{
	return reg2bitField.size();
}

template < typename K >
const BitField& BitField::Pool<K>::bitField(const KeyType& key) const
{
	typename Map::const_iterator it = reg2bitField.find(key);
	if (it == reg2bitField.end())
		throw BitFieldNotFoundException(key);

	return it->second;
}

template < typename K >
typename BitField::Pool<K>::KeySet BitField::Pool<K>::keySet()
{
	KeySet _keySet;
	for (typename Map::const_iterator it = reg2bitField.begin(); it != reg2bitField.end(); ++it)
	{
		_keySet.insert(it->first);
	}
	return _keySet;
}

/**
 * Nested exception class BitField::Pool<K>::BitFieldNotFoundException. Thrown when a bit field key is not found in the
 * bit field map.
 */
template < typename K >
struct BitField::Pool<K>::BitFieldNotFoundException : public ::utl::Exception
{
	BitFieldNotFoundException(const KeyType& bfKey) : bfKey(bfKey) {}
	~BitFieldNotFoundException() throw() {}
	void writeMessage(std::ostream& stream) const { stream << "Cannot find bit field \"" << bfKey << "\""; }
	const KeyType bfKey;
};

} // namespace hw
} // namespace mdcs

#endif /* mdcs_hw_BitField_h_INCLUDED */
