/**
 * @file DeviceT.hpp
 * @created Oct 26, 2012
 * @author anvar
 * ---------------------------------------------------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * ---------------------------------------------------------------------------------------------------------------------
 * Contributors: Shebli Anvar (shebli.anvar@cea.fr)
 * ---------------------------------------------------------------------------------------------------------------------
 * This file is part of the Mordicus software project.
 * ---------------------------------------------------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 * ---------------------------------------------------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 * =====================================================================================================================
 */

#ifndef mdcs_hw_DeviceT_hpp_INCLUDED
#define mdcs_hw_DeviceT_hpp_INCLUDED

#include "BaseDevice.h"
#include "mdcs/hw/RegisterT.hpp"
#include "mdcs/hw/DeviceHandle.h"
#include <map>
#include <set>

namespace mdcs
{
namespace hw
{

/**
 * Generic register manager template. This class contains all the functionalities expected from a device as a collection
 * of registers.
 * @tparam KR Type of key used to map the registers of this device (typically string or enum).
 * @tparam CD Type of core device class wrapped by this device.
 */
template<typename KR, class R>
class DeviceT : public BaseDevice
{
	template<class DType, class DCType> friend struct Init;

public:
	typedef KR RegKey; ///< Register key type
	typedef R Register; ///< Register type
	typedef typename Register::Policy Policy; ///< Register access policy

	typedef std::map<RegKey, Register*> RegMap; ///< Register map type
	typedef std::map<Register*, RegKey> InvRegMap; ///< Inverted Register map type
	typedef std::set<RegKey> RegKeySet; ///< Set of register keys within this device.

	struct RegisterNotFoundException; ///< Exception thrown when register search within device is unsuccessful.
	struct AlreadyUsedRegKeyException; ///< Exception thrown when a second mapping to a register key is attempted.

public:
	DeviceT(const std::string& instanceId) : BaseDevice(handle(instanceId)) {}
	virtual ~DeviceT() { clearRegs(); }
	static DeviceHandle handle(const std::string& instanceId) { return DeviceHandle(Policy::name(), instanceId); }


	template <class RegConfig>
	DeviceT& newReg(const RegKey& regKey, const RegConfig& cfg);

	DeviceT& addReg(const RegKey& regKey, std::auto_ptr<Register> reg) { return addReg(regKey, reg.release()); }
	virtual void clearRegs(); ///< @override
	virtual Register* srchReg(const RegKey& regKey);
	  const Register* srchReg(const RegKey& regKey) const { Register* reg = srchReg(regKey); return reg; }
	virtual Register& findReg(const RegKey& regKey);
	  const Register& findReg(const RegKey& regKey) const { Register& reg = srchReg(regKey); return reg; }

	      Register& operator[] (const RegKey& regKey)       { return findReg(regKey); }
	const Register& operator[] (const RegKey& regKey) const { return findReg(regKey); }

	size_t regCount () const { return regMap.size(); }
	RegKeySet regKeySet() const;

protected:
	virtual DeviceT& addReg(const RegKey& regKey, Register* reg);

private:
	RegMap regMap;
};

template<typename KR, class R>
struct DeviceT<KR,R>::RegisterNotFoundException : public ::utl::Exception
{
	RegisterNotFoundException(const RegKey& regKey) : regKey(regKey) {}
	~RegisterNotFoundException() throw() {}
	void writeMessage(std::ostream& stream) const { stream << "Cannot find register: " << regKey; }
	const RegKey regKey;
};

template<typename KR, class R>
struct DeviceT<KR,R>::AlreadyUsedRegKeyException : public ::utl::Exception
{
	AlreadyUsedRegKeyException(const RegKey& regKey) : regKey(regKey) {}
	~AlreadyUsedRegKeyException() throw() {}
	void writeMessage(std::ostream& stream) const { stream << "Register Key already in use: " << regKey; }
	const RegKey regKey;
};

} // namespace hw
} // namespace mdcs

#endif /* mdcs_hw_DeviceT_hpp_INCLUDED */
