/**
 * @file: DeviceT.ipp
 * @created: Feb 7, 2013
 * @author: anvar
 * ---------------------------------------------------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * ---------------------------------------------------------------------------------------------------------------------
 * Contributors: Shebli Anvar (shebli.anvar@cea.fr)
 * ---------------------------------------------------------------------------------------------------------------------
 * This file is part of the Mordicus software project.
 * ---------------------------------------------------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 * ---------------------------------------------------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licensing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 * =====================================================================================================================
 */

#ifndef mdcs_hw_DeviceT_ipp_INCLUDED
#define mdcs_hw_DeviceT_ipp_INCLUDED

#include "DeviceT.hpp"
#include <utl/Logging.h>
#include <memory>

namespace mdcs
{
namespace hw
{

/**
 * @fn DeviceT<KR,R>::DeviceT(const std::string& instanceId)
 * Device constructor.
 * @param instanceId Unique string ID attached to this device instance.
 */

/**
 * @fn DeviceT<KR,R>::~DeviceT()
 * Device destructor: releases all registers and other resources.
 */

/**
 * @fn DeviceHandle DeviceT<KR,R>::handle(const std::string& instanceId)
 * This class method returns a DeviceHandle object initialized with specified instance ID and the policy name associated
 * with this device (through the policy (P) template parameter).
 * @param instanceId Unique string associated with a device instance.
 */

/**
 * Inserts a new register object within this device, taking ownership of the register object.
 * @fn DeviceT<KR,R>& DeviceT<KR,R>::addReg (const RegKey& regKey, std::auto_ptr<Register> reg)
 * @param regKey The unique key to which the register is mapped within this device.
 * @param reg An autopointer containing the new register to be inserted into this device.
 * @return A reference this device to allow chained operations.
 * @throw AlreadyUsedRegKeyException if specified key is already in use (i.e. already mapped to another register).
 * @throw NullRegisterException is inserted register is null.
 */

/**
 * Inserts a new register object within this device, taking ownership of the register object.
 * @param regKey The unique key to which the register is mapped within this device.
 * @param reg Pointer to the new register to be inserted into this device.
 * @return A reference this device to allow chained operations.
 * @throw AlreadyUsedRegKeyException if specified key is already in use (i.e. already mapped to another register).
 * @throw NullRegisterException is inserted register is null.
 */
template<typename KR, class R>
DeviceT<KR,R>& DeviceT<KR,R>::addReg (const RegKey& regKey, Register* reg)
{
	if (regMap.find(regKey) != regMap.end())
	{
		// Since we are expected to seize ownership of reg, we need to delete it if we can't insert it in our map.
		if (reg != 0) delete reg;
		throw AlreadyUsedRegKeyException(regKey);
	}

	BaseDevice::addReg(reg);
	regMap[regKey] = reg;
	return *this;
}
/**
 * This template method creates and adds a new register associated to specified key and configured by config object.
 * @tparam RegConfig The type of the config object used to initialize a register. For this function to work, a
 * specialized mdcs::hw::Init structure with proper register type, register config type and init method must have been
 * implemented.
 * @param regKey The key used to map the new device register.
 * @param cfg The configuration object from which the new register is initialized.
 * @return A reference to this device to allow chained operations.
 */
template<typename KR, class R>
template <class RegConfig>
DeviceT<KR,R>& DeviceT<KR,R>::newReg(const RegKey& regKey, const RegConfig& cfg)
{
	LOG_DEBUG() << "Creating register on key: " << regKey;
	Register* reg = new Register;
	call_init(*reg, cfg);
	addReg(regKey, reg);
	return *this;
}

/**
 * Destroys all the registers of this device.
 */
template<typename KR, class R>
void DeviceT<KR,R>::clearRegs()
{
	regMap.clear();
	BaseDevice::clearRegs();
}

/**
 * Searches for a register mapped to specified key.
 * @param regKey The key to which the register is mapped within this device.
 * @return Pointer to mapped register or 0 if not found.
 */
template<typename KR, class R>
typename DeviceT<KR,R>::Register* DeviceT<KR,R>::srchReg(const RegKey& regKey)
{
	typename RegMap::iterator it = regMap.find(regKey);
	return (it != regMap.end()) ? it->second : 0;
}

/**
 * Looks for a register mapped to specified key.
 * @param regKey The key to which the register is mapped within this device.
 * @return Pointer to mapped register.
 * @throw RegisterNotFoundException if no register is mapped to specified key.
 */
template<typename KR, class R>
typename DeviceT<KR,R>::Register& DeviceT<KR,R>::findReg(const RegKey& regKey)
{
	typename RegMap::iterator it = regMap.find(regKey);
	if (it != regMap.end())
	{
		return *it->second;
	}
	throw RegisterNotFoundException(regKey);
}

/**
 * @fn size_t DeviceT<KR,R>::regCount() const
 * Counts the total number of registers defined within this device.
 * @return Number of registers.
 */

/**
 * Gets the set of all register keys of this device.
 * @return A RegKeySet object containing all the keys.
 */
template<typename KR, class R>
typename DeviceT<KR,R>::RegKeySet DeviceT<KR,R>::regKeySet() const
{
	RegKeySet regKeySet_;
	for (typename RegMap::const_iterator it = regMap.begin(); it != regMap.end(); ++it)
	{
		regKeySet_.insert(it->first);
	}
	return regKeySet_;
}

} // namespace hw
} // namespace mdcs

#endif // mdcs_hw_DeviceT_ipp_INCLUDED
