/**
 * @file RegisterT.hpp
 * @created Oct 26, 2012
 * @author anvar
 * ---------------------------------------------------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * ---------------------------------------------------------------------------------------------------------------------
 * Contributors: Shebli Anvar (shebli.anvar@cea.fr)
 * ---------------------------------------------------------------------------------------------------------------------
 * This file is part of the Mordicus software project.
 * ---------------------------------------------------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 * ---------------------------------------------------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 * =====================================================================================================================
 */

#ifndef mdcs_hw_RegisterT_hpp_INCLUDED
#define mdcs_hw_RegisterT_hpp_INCLUDED

#include "mdcs/hw/AbstractRegister.h"
#include "mdcs/hw/BitField.h"
#include "mdcs/utl/TypeInfo.hpp"

namespace mdcs
{
namespace hw
{

/**
 * This is a helper template class to allow for "partial" template specialization of the template peek and poke
 * operations of the RegisterT template class.
 * @tparam P A register access policy class.
 * @tparam V The type of value that is read from or written to a register.
 */
template < class P, typename V >
struct PeekPokeOp
{
	typedef P Policy; ///< Register access policy
	typedef V ValueType; ///< Type of value used in peek/poke operation
	typedef typename Policy::DataType DataType; ///< Type of data that is read from / written to this type of register
	typedef typename Policy::AddrType AddrType; ///< Register address type allowing register instance referencing

	static void poke(const AddrType& addr, const ValueType& value);
	static void peek(const AddrType& addr,       ValueType& value);
};

template < class P, typename V >
void PeekPokeOp<P,V>::poke(const AddrType& addr, const ValueType& value)
{
	LOG_DEBUG() << "Creating transient " << utl::TypeInfo<DataType>::name()
	            << " initialized with value: " << std::hex << value << std::dec;
	DataType data(value);
	LOG_DEBUG() << "Poking transient " << utl::TypeInfo<DataType>::name() << " @ " << ((void*)addr);
	Policy::poke(addr, data);
}

template < class P, typename V >
void PeekPokeOp<P,V>::peek(const AddrType& addr, ValueType& value)
{
	DataType data;
	Policy::peek(addr, data);
	value = (ValueType) data; // cast operation must exist
}

template < class P >
struct PeekPokeOp<P, typename P::DataType>
{
	typedef P Policy; ///< Register access policy
	typedef typename Policy::DataType DataType; ///< Type of data that is read from / written to this type of register
	typedef typename Policy::AddrType AddrType; ///< Register address type allowing register instance referencing
	typedef DataType ValueType;

	static void poke(const AddrType& addr, const ValueType& value) { Policy::poke(addr, value); }
	static void peek(const AddrType& addr,       ValueType& value) { Policy::peek(addr, value); }
};

/**
 * This template class represents a register. It is not meant to be specialized, the template parameters allowing its
 * the needed variability. The register also contains a bitfield map (which can be empty) associating a key of type
 * given by 1st template parameter to a BitField definition. The bitfield map is an instance of the BitField::Pool
 * template class.
 * @tparam KB The type of key associated to a bitfield of this register.
 * @tparam P The register access policy, defining:
 *   1) the data type associated to the register value: typename DataType (what type of value you read / write)
 *   2) the address type associated to the register: typename AddrType (allows full referencing of a register instance)
 *   3) how the register is read out: static void peek(const AddrType& addr, DataType& value)
 *   4) how the register is written to: static void poke(const AddrType& addr, const DataType& value)
 *   5) a unique string identifying this policy (returned by a class operation):  static std::string name()
 *
 *   NOTE: If more than one value types are to be read out from and written to the registers, then the peek / poke class
 *   operations should be defined as template methods where the template parameter represents the value type. Then
 *   specializations for each supported value type should be provided (e.g. int, unsigned, short, char...)
 */
template < typename KB, class P >
class RegisterT : public AbstractRegister
{
	template<class RType, class RCType> friend struct Init;

public:
	typedef P Policy; ///< Register access policy
	typedef KB BitFieldKey; ///< Key type to which register bit fields are mapped
	typedef typename Policy::DataType DataType; ///< Type of data that is read from / written to this type of register
	typedef typename Policy::AddrType AddrType; ///< Register address type allowing register instance referencing

public:
	RegisterT(const AddrType& addr) : _address(addr) {} ///< Constructor with specified register address
	RegisterT() {} /// Default constructor: register address might me invalid; needs to be initialized using call_init
	virtual ~RegisterT() {}
	const std::string policyName() const { return  Policy::name(); } ///< Implementation of abstract operation policyName()
	const AddrType& address() const { return _address; }
	BitField::Pool<BitFieldKey>& bitFieldPool() { return bfPool; }
	const BitField::Pool<BitFieldKey>& bitFieldPool() const { return bfPool; }

	///@{ Register read / write operations operations
	template <typename ValueType>
	RegisterT& poke(const ValueType& value) { PeekPokeOp<Policy,ValueType>::poke(address(), value); return *this; }
	template <typename ValueType>
	RegisterT& peek(ValueType& value) { PeekPokeOp<Policy,ValueType>::peek(address(), value); return *this; }
	template <typename ValueType>
	const RegisterT& peek(ValueType& value) const { PeekPokeOp<Policy,ValueType>::peek(address(), value); return *this; }

	DataType value() const { DataType value;  peek(value); return value; }
	template <typename ValueType>
	RegisterT& operator = (const ValueType& value) { poke(value); return *this; }
	operator DataType() { return value(); } ///< Cast to DataType operator
	///@}

	///@{ Bit field manipulation operations
	BitField::Operator<RegisterT> bitFieldOperator() const { BitField::Operator<RegisterT> bfop(*this); return bfop; }

	template < typename ValueType >
	BitField::Operator<RegisterT> poke(unsigned pos, unsigned wid, const ValueType& value)
	{
		return BitField::Operator<RegisterT>(*this).poke(pos, wid, value);
	}

	template < typename ValueType >
	BitField::Operator<RegisterT> peek(unsigned pos, unsigned wid, ValueType& value)
	{
		return BitField::Operator<RegisterT>(*this).peek(pos, wid, value);
	}
	template < typename ValueType >
	BitField::Operator<RegisterT> poke(const BitField& bf, const ValueType& value) { return poke(bf.pos, bf.wid, value); }
	template < typename ValueType >
	BitField::Operator<RegisterT> peek(const BitField& bf,       ValueType& value) { return peek(bf.pos, bf.wid, value); }

	const BitField& operator [] (const BitFieldKey& bfKey) const { return bfPool.bitField(bfKey); }

	template < typename ValueType >
	BitField::Operator<RegisterT> poke(const BitFieldKey& bfKey, const ValueType& value) { return poke((*this)[bfKey], value); }
	template < typename ValueType >
	BitField::Operator<RegisterT> peek(const BitFieldKey& bfKey,       ValueType& value) { return peek((*this)[bfKey], value); }
	///@}

private:
	AddrType _address; ///< Absolute reference of the register: whatever is needed by the peek/poke operations.
	BitField::Pool<BitFieldKey> bfPool;

};

} // namespace hw
} // namespace mdcs

#endif /* mdcs_hw_RegisterT_hpp_INCLUDED */
