/**
 * @file MemBus/Policy.hpp
 * @created Oct 30, 2012
 * @author anvar
 * ---------------------------------------------------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * ---------------------------------------------------------------------------------------------------------------------
 * Contributors: Shebli Anvar (shebli.anvar@cea.fr)
 * ---------------------------------------------------------------------------------------------------------------------
 * This file is part of the Mordicus software project.
 * ---------------------------------------------------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 * ---------------------------------------------------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 * =====================================================================================================================
 */

#ifndef mdcs_hw_policy_MemBus_Policy_hpp_INCLUDED
#define mdcs_hw_policy_MemBus_Policy_hpp_INCLUDED

#include "mdcs/hw/RegisterT.hpp"
#include "mdcs/utl/TypeInfo.hpp"
#include "utl/RtBitSet.hpp"
#include "utl/BitSet.hpp"
#include <boost/type_traits/conditional.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/cstdint.hpp>
#include <string>
#include <vector>
#include <cstring>

namespace mdcs
{
namespace hw
{
namespace policy
{
namespace MemBus
{

/**
 * Base policy class for MemBus register access policy
 */
class BasePolicy
{
public:
	typedef void* AddrType; ///< Address type allowing to unambiguously reference this type of registers.
	typedef ptrdiff_t Offset; ///< Offset of a register relative to a base address.
	/// this structure references a register within a device by holding the device base address and the register offset.
	struct InDeviceRef
	{
		InDeviceRef(const AddrType& baseAddress, const Offset& offset) : baseAddress(baseAddress), offset(offset) {}
		const AddrType baseAddress;
		const Offset offset;
	};

public:
	/// Returns a unique string naming the full policy (including data type)
	/// @param dataTypeName String uniquely referring to the policy's data type
	static std::string name(const std::string& dataTypeName) { return "MemBus<" + dataTypeName + ">"; }
};

/**
 * General memory bus access policy template.
 * @tparam W The complete data representation of the memory bus register. Typically it would be either a scalar type or
 * a bit set type.
 */
template<typename W>
class Policy : public BasePolicy
{
public:
	typedef W DataType; ///< The type of data that is read out from and written to this type of registers.
	typedef BasePolicy::AddrType AddrType; ///< Address type allowing to unambiguously reference this type of registers.
	typedef BasePolicy::Offset Offset; ///< Offset of a register relative to a base address.

public:
	/// Returns a unique string naming the full policy (including data type)
	static std::string name() { return BasePolicy::name(::mdcs::utl::TypeInfo<DataType>::name()); }

	/// Method for writing data into register.
	static void poke(const AddrType& addr, const DataType& value);

	/// Method for reading from register.
	static void peek(const AddrType& addr, DataType& value);
};

/**
 * Specialized memory bus template for RtBitSet (run time bitset) objects
 * @tparam SGNSS Signedness of the considered register.
 */
template < ::utl::Signedness SGNSS >
class Policy< ::utl::RtBitSet<SGNSS> > : public BasePolicy
{
public:
	typedef ::utl::RtBitSet<SGNSS> DataType;
	typedef BasePolicy::AddrType AddrType; ///< Address type allowing to unambiguously reference this type of registers.
	typedef BasePolicy::Offset Offset; ///< Offset of a register relative to a base address.

public:
	/// Returns a unique string naming the full policy (including data type)
	static std::string name() { return BasePolicy::name(::mdcs::utl::TypeInfo<DataType>::name()); }

	/// Method for writing data into register.
	static void poke(const AddrType& addr, const DataType& sourceBitSet);

	/// Method for reading from register.
	static void peek(const AddrType& addr, DataType& targetBitSet);
};

/**
 * Specialized memory bus template for BitSet (compile time bitset) objects.
 * @tparam NBITS The number of bits of this bitset.
 * @tparam SGNSS Signedness of the considered register.
 */
template < size_t NBITS,  ::utl::Signedness SGNSS >
class Policy< ::utl::BitSet<NBITS,SGNSS> > : public BasePolicy
{
public:
	typedef ::utl::BitSet<NBITS,SGNSS> DataType;
	typedef BasePolicy::AddrType AddrType; ///< Address type allowing to unambiguously reference this type of registers.
	typedef BasePolicy::Offset Offset; ///< Offset of a register relative to a base address.

public:
	/// Returns a unique string naming the full policy (including data type)
	static std::string name() { return BasePolicy::name(::mdcs::utl::TypeInfo<DataType>::name()); }

	/// Method for writing data into register.
	static void poke(const AddrType& addr, const DataType& sourceBitSet);

	/// Method for reading from register.
	static void peek(const AddrType& addr, DataType& targetBitSet);
};

} // namespace MemBus
} // namespace policy

/**
 * Specialized template class used to initialize a memory-mapped register directly using base address and offset values.
 */
template<typename KB, typename W>
struct Init< RegisterT< KB, policy::MemBus::Policy<W> >, typename policy::MemBus::BasePolicy::InDeviceRef >
{
	typedef RegisterT< KB, policy::MemBus::Policy<W> > RegisterType;
	typedef typename policy::MemBus::BasePolicy::InDeviceRef ConfigType;
	static void init(RegisterType& reg, const ConfigType& cfg)
	{
		reg._address = reinterpret_cast<char*>(cfg.baseAddress) + cfg.offset;
	}
};

} // namespace hw
} // namespace mdcs

#endif /* mdcs_hw_policy_MemBus_Policy_hpp_INCLUDED */
