/**
 * @file MemBus/Policy.ipp
 * @created Sep 16, 2013
 * @author anvar
 * ---------------------------------------------------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * ---------------------------------------------------------------------------------------------------------------------
 * Contributors: Shebli Anvar (shebli.anvar@cea.fr)
 * ---------------------------------------------------------------------------------------------------------------------
 * This file is part of the Mordicus software project.
 * ---------------------------------------------------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 * ---------------------------------------------------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 * =====================================================================================================================
 */

#ifndef mdcs_hw_policy_MemBus_Policy_ipp_INCLUDED
#define mdcs_hw_policy_MemBus_Policy_ipp_INCLUDED

#include "mdcs/hw/policy/MemBus/Policy.hpp"
#include <utl/Logging.h>

namespace mdcs
{
namespace hw
{
namespace policy
{
namespace MemBus
{

/**
 * Writes data value into referenced register. This implements the generic case where DataType (W) is a standard scalar.
 * @tparam W The complete data representation of the memory bus register.
 * @param addr The memory address of the register (lowest byte, LSB in little-endian, MSB in big-endian).
 * @param value The value written into the register.
 */
template<typename W>
inline void Policy<W>::poke(const AddrType& addr, const DataType& value)
{
	LOG_DEBUG() << name() << "::poke " << std::hex << value << " @ address: " << addr << std::dec;
	static_cast<DataType*>(addr)[0] = value;
}

/**
 * Reads data value from referenced register. This implements the generic case where DataType (W) is a standard scalar.
 * @tparam W The complete data representation of the memory bus register.
 * @param addr The memory address of the register (lowest byte).
 * @param value The value read out from the register.
 */
template<typename W>
inline void Policy<W>::peek(const AddrType& addr, DataType& value)
{
	LOG_DEBUG() << name() << "::peek @ address: " << addr;
	value = static_cast<DataType*>(addr)[0];
	LOG_DEBUG() << name() << "::peeked: " << std::hex << value << std::dec;
}

/**
 * Poke method for RunTime BitSet Policy data type and bitset value types.
 * @param addr Register address
 * @param sourceBitSet This parameter is a bit set (an RtBitSet actually) from which the register is written to.
 * @note: The number of actual bits written into the register are given by sourceBitSet.bitCount()
 */
template < ::utl::Signedness SGNSS >
inline void Policy< ::utl::RtBitSet<SGNSS> >::poke(const AddrType& addr, const DataType& sourceBitSet)
{
	LOG_DEBUG() << name() << "::poke " << " @ address: " << addr << std::dec;
	char* addr_ = static_cast<char*>(addr);
	sourceBitSet.getValue(addr_, sourceBitSet.minByteCount());
}

/**
 * Peek method for RunTime BitSet Policy data type and bitset value types.
 * @param addr Register address
 * @param targetBitSet This parameter is a bit set (an RtBitSet actually) into which the register is read out.
 * @note: The number of actual bits written into the register are given by targetBitSet.bitCount(). As a consequence,
 *        users need to set this bitCount value before calling this method.
 */
template < ::utl::Signedness SGNSS >
inline void Policy< ::utl::RtBitSet<SGNSS> >::peek(const AddrType& addr, DataType& targetBitSet)
{
	LOG_DEBUG() << name() << "::peek @ address: " << addr;
	const char* addr_ = static_cast<const char*>(addr);
	targetBitSet.setValue(addr_, targetBitSet.minByteCount());
}

/**
 * Poke method for compile time BitSet Policy data type and bitset value types.
 * @param addr Register address
 * @param sourceBitSet This parameter is a bit set (an BitSet actually) from which the register is written to.
 * @note: The number of actual bits written into the register are given by sourceBitSet.bitCount()
 */
template < size_t NBITS,  ::utl::Signedness SGNSS >
inline void Policy< ::utl::BitSet<NBITS,SGNSS> >::poke(const AddrType& addr, const DataType& sourceBitSet)
{
	LOG_DEBUG() << name() << "::poke " << " @ address: " << addr << std::dec;
	char* addr_ = static_cast<char*>(addr);
	sourceBitSet.getValue(addr_, sourceBitSet.minByteCount());
}

/**
 * Peek method for compile time BitSet Policy data type and bitset value types.
 * @param addr Register address
 * @param targetBitSet This parameter is a bit set (an BitSet actually) into which the register is read out.
 * @note: The number of actual bits written into the register are given by targetBitSet.bitCount(). As a consequence,
 *        users need to set this bitCount value before calling this method.
 */
template < size_t NBITS,  ::utl::Signedness SGNSS >
inline void Policy< ::utl::BitSet<NBITS,SGNSS> >::peek(const AddrType& addr, DataType& targetBitSet)
{
	LOG_DEBUG() << name() << "::peek @ address: " << addr;
	const char* addr_ = static_cast<const char*>(addr);
	targetBitSet.setValue(addr_, targetBitSet.minByteCount());
}

} // namespace MemBus
} // namespace policy
} // namespace hw
} // namespace mdcs

#endif /* mdcs_hw_policy_MemBus_Policy_ipp_INCLUDED */
