/**
 * @file CircularBuffer.cpp
 * @date 15 mai 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "CircularBuffer.h"
#include <cassert>

namespace mdcs
{
namespace utl
{

// =============================================================================
// mdcs::utl::CircularBuffer class methods body
// -----------------------------------------------------------------------------

CircularBuffer::CircularBuffer()
	: begin_(0), end_(0), read_(0), write_(0)
{
}

CircularBuffer::CircularBuffer(uint8_t* base, size_t length)
	: begin_(base), end_(base + length), read_(base), write_(base)
{
	assert(length > 1);
}

void CircularBuffer::init(uint8_t* base, size_t length)
{
	assert(length > 1);
	begin_ = base;
	end_  = base + length;
	read_ = base;
	write_ = base;
}

void CircularBuffer::moveWritePtr(size_t nBytes)
{
	if(nBytes > sizeFree())
	{
		throw Exception(NotEnoughSpace);
	}
	write_ += nBytes;
	if(write_ >= end_)
	{
		write_ += begin_ - end_;
	}
}

void CircularBuffer::moveReadPtr(size_t nBytes)
{
	if(nBytes > sizeOccupied())
	{
		throw Exception(NotEnoughData);
	}
	read_ += nBytes;
	if(read_ >= end_)
	{
		read_ -= end_ - begin_;
	}
}

bool CircularBuffer::empty() const
{
	return read_ == write_;
}

bool CircularBuffer::full() const
{
	return (write_ + 1 == read_) or (read_ == begin_ and write_ + 1 == end_);
}

size_t CircularBuffer::size() const
{
	return end_ - begin_;
}

size_t CircularBuffer::sizeFree() const
{
	if(write_ < read_)
	{
		return read_ - write_ - 1;
	}
	else
	{
		return (end_ - begin_) + (read_ - write_) - 1;
	}
}

size_t CircularBuffer::sizeNextContiguousFree() const
{
	if(write_ < read_)
	{
		return read_ - write_ - 1;
	}
	else
	{
		size_t size = end_ - write_;
		if(read_ == begin_)
		{
			--size; // if the read pointer is at the beginning, we need to leave the last slot open.
		}
		return size;
	}
}

size_t CircularBuffer::sizeOccupied() const
{
	if(write_ < read_)
	{
		return (end_ - begin_) - (read_ - write_);
	}
	else
	{
		return write_ - read_;
	}
}

size_t CircularBuffer::sizeNextContiguousOccupied() const
{
	if(write_ < read_)
	{
		return end_ - read_;
	}
	else
	{
		return write_ - read_;
	}
}

// =============================================================================
// mdcs::utl::CircularBuffer::Exception class methods body
// -----------------------------------------------------------------------------

CircularBuffer::Exception::~Exception() throw()
{
}

void CircularBuffer::Exception::writeMessage(std::ostream& stream) const
{
	static const char* const message[InvalidError] = {"not enough space!", "not enough data!"};
	stream << "circular buffer: " << message[error];
}

} /* namespace utl */
} /* namespace mdcs */
