/**
 * @file ContiguousCircularBuffer.cpp
 * @date 15 mai 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include "ContiguousCircularBuffer.h"
#include <utl/Logging.h>
#include <limits>
#include <cassert>

#undef max

namespace mdcs
{
namespace utl
{

// =============================================================================
// mdcs::utl::ContiguousCircularBuffer class methods body
// -----------------------------------------------------------------------------

ContiguousCircularBuffer::ContiguousCircularBuffer()
	: begin_(0), end_(0), limit_(0), read_(0), write_(0)
{
}

ContiguousCircularBuffer::ContiguousCircularBuffer(uint8_t* base, size_t length)
	: begin_(base), end_(base + length), limit_(end_), read_(base), write_(base)
{
	assert(length > 1);
}

void ContiguousCircularBuffer::init(uint8_t* base, size_t length)
{
	assert(length > 1);
	begin_ = base;
	end_  = base + length;
	limit_ = end_;
	read_ = base;
	write_ = base;
}

ContiguousCircularBuffer::WriteSession ContiguousCircularBuffer::write(size_t nBytes)
{
	WriteSession session(*this, nBytes);
	if(nBytes > sizeFree())
	{
		throw Exception(NotEnoughSpace, nBytes, sizeFree());
	}
	if(write_ < read_)
	{
	}
	else
	{
		if(nBytes > sizeNextContiguousFree())
		{
			// trying to find a contiguous block large enough at the beginning of the buffer.
			if((nBytes + 1) > (read_ - begin_)) // nBytes + 1: because we need to leave one slot open
			{
				throw Exception(NotEnoughSpace, nBytes + 1, read_ - begin_);
			}
			else
			{
				limit_ = write_;
				write_ = begin_;
			}
		}
	}
	return session;
}

void ContiguousCircularBuffer::moveReadPtr(size_t nBytes)
{
	if(nBytes > sizeOccupied())
	{
		throw Exception(NotEnoughData, nBytes, sizeOccupied());
	}
	read_ += nBytes;
	if(read_ >= limit_)
	{
		read_ -= limit_ - begin_;
		limit_ = end_;
	}
	if(empty())
	{
		// if all data was consumed, we reset the pointers to the beginning
		// of the buffer: it will allow achieving bigger contiguous block
		// more easily.
		read_ = write_ = begin_;
	}
}

bool ContiguousCircularBuffer::empty() const
{
	return read_ == write_;
}

bool ContiguousCircularBuffer::full() const
{
	return (write_ + 1 == read_) or (read_ == begin_ and write_ + 1 == end_);
}

size_t ContiguousCircularBuffer::size() const
{
	return end_ - begin_;
}

size_t ContiguousCircularBuffer::sizeFree() const
{
	if(write_ < read_)
	{
		return read_ - write_ - 1;
	}
	else
	{
		return (end_ - begin_) + (read_ - write_) - 1;
	}
}

size_t ContiguousCircularBuffer::sizeNextContiguousFree() const
{
	if(write_ < read_)
	{
		return read_ - write_ - 1;
	}
	else
	{
		size_t size = end_ - write_;
		if(read_ == begin_)
		{
			--size; // if the read pointer is at the beginning, we need to leave the last slot open.
		}
		return size;
	}
}

size_t ContiguousCircularBuffer::sizeMaxContiguousFree() const
{
	if(write_ < read_)
	{
		return read_ - write_ - 1;
	}
	else
	{
		if(read_ == begin_)
		{
			return end_ - write_ - 1;
		}
		else
		{
			return std::max(end_ - write_, read_ - begin_ - 1);
		}
	}
}

size_t ContiguousCircularBuffer::sizeOccupied() const
{
	if(write_ < read_)
	{
		return (limit_ - begin_) - (read_ - write_);
	}
	else
	{
		return write_ - read_;
	}
}

size_t ContiguousCircularBuffer::sizeNextContiguousOccupied() const
{
	if(write_ < read_)
	{
		return limit_ - read_;
	}
	else
	{
		return write_ - read_;
	}
}


// =============================================================================
// mdcs::utl::ContiguousCircularBuffer::WriteSession class methods body
// -----------------------------------------------------------------------------

ContiguousCircularBuffer::WriteSession::WriteSession(const WriteSession& session)
	: buffer_(session.buffer_), maxBytes_(session.maxBytes_)
{
	const_cast<WriteSession&>(session).maxBytes_ = std::numeric_limits<size_t>::max();
}
ContiguousCircularBuffer::WriteSession::~WriteSession()
{
	if(maxBytes_ != std::numeric_limits<size_t>::max())
	{
		if(maxBytes_ != 0)
		{
			if(!std::uncaught_exception())
			{
				LOG_ERROR() << "Uncommitted write session in buffer";
			}
			rollback();
		}
	}
}

uint8_t* ContiguousCircularBuffer::WriteSession::pointer() const
{
	return buffer_.write_;
}

void ContiguousCircularBuffer::WriteSession::commit()
{
	doCommit(maxBytes_);
}

void ContiguousCircularBuffer::WriteSession::commit(size_t nBytes)
{
	if(nBytes > maxBytes_)
	{
		throw Exception(NotEnoughReserve, nBytes, maxBytes_);
	}
	if(nBytes == 0)
	{
		rollback();
	}
	else
	{
		doCommit(nBytes);
	}
}

void ContiguousCircularBuffer::WriteSession::rollback()
{
	// Only rollback when the write pointer has been moved because
	// there was not enough free space at the end of the buffer.
	// In this case, the write pointer is at the beginning of the buffer
	// and the limit pointer is not at the end of the buffer.
	if(buffer_.write_ == buffer_.begin_ and buffer_.limit_ != buffer_.end_)
	{
		buffer_.write_ = buffer_.limit_;
		buffer_.limit_ = buffer_.end_;
	}
	maxBytes_ = 0;
}

void ContiguousCircularBuffer::WriteSession::doCommit(size_t nBytes)
{
	buffer_.write_ += nBytes;
	assert(buffer_.write_ <= buffer_.end_); // by design, overflowing the buffer end is impossible
	if(buffer_.write_ == buffer_.end_) // but, it is possible to exactly reach the end.
	{
		buffer_.write_ = buffer_.begin_;
	}
	maxBytes_ = 0;
}

ContiguousCircularBuffer::WriteSession::WriteSession(ContiguousCircularBuffer& buffer, size_t maxBytes)
	: buffer_(buffer), maxBytes_(maxBytes)
{
}

ContiguousCircularBuffer::WriteSession& ContiguousCircularBuffer::WriteSession::operator=(const WriteSession& /*session*/)
{
	assert(false);
	return *this;
}

// =============================================================================
// mdcs::utl::ContiguousCircularBuffer::Exception class methods body
// -----------------------------------------------------------------------------

ContiguousCircularBuffer::Exception::~Exception() throw()
{
}

void ContiguousCircularBuffer::Exception::writeMessage(std::ostream& stream) const
{
	static const char* const message[InvalidError] = {"not enough space!", "not enough data!", "more data was written than reserved!"};
	stream << "contiguous circular buffer: " << message[error] << " requested: " << sizeRequested << " actual: " << sizeActual;
}

} /* namespace utl */
} /* namespace mdcs */
