/**
 * @file MetaPolicy.hpp
 * @date 18 juil. 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#ifndef mdcs_utl_MetaPolicy_hpp_INCLUDED
#define mdcs_utl_MetaPolicy_hpp_INCLUDED

#include <utl/Factory.hpp>
#include <utl/Singleton.hpp>
#include <utl/Configurator.h>
#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/type_traits/is_base_of.hpp>
#include <boost/type_traits/is_abstract.hpp>
#include <boost/static_assert.hpp>
#include <boost/utility.hpp>
#include <memory>

namespace mdcs
{
namespace utl
{

template<class MetaPolicy> class AbstractParamData;

template<class Policy>
struct PolicyArray
{
};

template<class I, class C, class R>
struct MetaPolicyType
{
	typedef I Interface;
	typedef C Context;
	typedef R RunInfo;

	typedef ::utl::FactoryMap<std::string, AbstractParamData<MetaPolicyType<I, C, R> >*()> FactoryMap;
	typedef ::utl::Singleton<FactoryMap> Factory;
	typedef ::utl::FactoryRegistrar<Factory> Registrar;

	typedef PolicyArray<MetaPolicyType<I, C, R> > Array; ///< typedef for use in mpl::vectors of types
};

template<class MetaPolicy>
class AbstractRunImpl : public MetaPolicy::Interface, private boost::noncopyable
{
public:
	BOOST_STATIC_ASSERT(boost::is_abstract<typename MetaPolicy::Interface>::value);

	virtual ~AbstractRunImpl() {}
};

template<class MetaPolicy>
class AbstractSetupData : private boost::noncopyable
{
public:
	typedef typename MetaPolicy::Context Context;
	typedef typename MetaPolicy::RunInfo RunInfo;

	typedef AbstractRunImpl<MetaPolicy> RunImpl;
	typedef std::auto_ptr<RunImpl> RunImplPtr;

	virtual ~AbstractSetupData() {}
	virtual RunImplPtr createRun(Context& context, const RunInfo& runInfo) = 0;
};

template<class MetaPolicy>
class AbstractParamData : private boost::noncopyable
{
public:
	typedef typename MetaPolicy::Context Context;

	typedef AbstractSetupData<MetaPolicy> SetupData;
	typedef std::auto_ptr<SetupData> SetupDataPtr;

	virtual ~AbstractParamData() {}
	virtual SetupDataPtr createSetup(Context& context) const = 0;
};

template<class MetaPolicy> class PolicySetup;
template<class MetaPolicy> class PolicyParams;

template<class MetaPolicy>
class PolicyRunImpl
{
public:
	typedef boost::scoped_ptr<AbstractRunImpl<MetaPolicy> > Ptr;
	typedef typename MetaPolicy::Interface Interface;
	typedef typename MetaPolicy::Context   Context;
	typedef typename MetaPolicy::RunInfo   RunInfo;

	void create(PolicySetup<MetaPolicy>& setup, Context& context, const RunInfo& runInfo)
	{
		policy_.reset(setup.setup->createRun(context, runInfo).release());
	}
	Interface* operator->() { return policy_.get(); }
	Interface& operator*()  { return *policy_; }
	const Interface* operator->() const { return policy_.get(); }
	const Interface& operator*()  const { return *policy_; }

	const Ptr& policy() const { return policy_; }
	Ptr& policy() { return policy_; }

private:
	Ptr policy_;
};

template<class MetaPolicy>
class PolicySetup
{
public:
	typedef boost::scoped_ptr<AbstractSetupData<MetaPolicy> > Ptr;
	typedef typename MetaPolicy::Context Context;

	friend class PolicyRunImpl<MetaPolicy>;

	void create(const PolicyParams<MetaPolicy>& params, Context& context)
	{
		setup.reset(params.params->createSetup(context).release());
	}

private:
	Ptr setup;
};

template<class MetaPolicy>
class PolicyParams
{
public:
	typedef boost::scoped_ptr<AbstractParamData<MetaPolicy> > Ptr;

	friend class PolicySetup<MetaPolicy>;

	void create(const std::string& className)
	{
		typedef typename MetaPolicy::Factory Factory;
		params.reset(Factory::instance()[className].create().release());
	}
	template<class C> void init(const C& config)
	{
		::utl::GlobalConfigurator::instance().configure(*params, config);
	}

private:
	Ptr params;
};



template<class MetaPolicy> class PolicySetupArray;
template<class MetaPolicy> class PolicyParamsArray;

template<class MetaPolicy>
class PolicyRunImplArray : private boost::noncopyable
{
public:
	typedef boost::ptr_vector<AbstractRunImpl<MetaPolicy> > Container;
	typedef typename MetaPolicy::Interface Interface;
	typedef typename MetaPolicy::Context   Context;
	typedef typename MetaPolicy::RunInfo   RunInfo;

	void create(PolicySetupArray<MetaPolicy>& setupArray, Context& context, const RunInfo& runInfo)
	{
		for(size_t i = 0; i < setupArray.setup.size(); ++i)
		{
			policy_.push_back(setupArray.setup[i].createRun(context, runInfo).release());
		}
	}

	size_t size() const { return policy_.size(); }
	Interface& operator[](size_t i) { return policy_[i]; }
	const Interface& operator[](size_t i) const { return policy_[i]; }

	const Container& policy() const { return policy_; }
	Container& policy() { return policy_; }

private:
	Container policy_;
};

template<class MetaPolicy>
class PolicySetupArray : private boost::noncopyable
{
public:
	typedef boost::ptr_vector<AbstractSetupData<MetaPolicy> > Container;
	typedef typename MetaPolicy::Context Context;

	friend class PolicyRunImplArray<MetaPolicy>;

	void create(const PolicyParamsArray<MetaPolicy>& paramsArray, Context& context)
	{
		for(size_t i = 0; i < paramsArray.params.size(); ++i)
		{
			setup.push_back(paramsArray.params[i].createSetup(context).release());
		}
	}
	size_t size() const { return setup.size(); }

private:
	Container setup;
};

template<class MetaPolicy>
class PolicyParamsArray : private boost::noncopyable
{
public:
	typedef boost::ptr_vector<AbstractParamData<MetaPolicy> > Container;

	friend class PolicySetupArray<MetaPolicy>;

	void create(const std::string& className)
	{
		typedef typename MetaPolicy::Factory Factory;
		params.push_back(Factory::instance()[className].create().release());
	}
	template<class C> void init(const C& config)
	{
		::utl::GlobalConfigurator::instance().configure(params.back(), config);
	}
	size_t size() const { return params.size(); }

private:
	Container params;
};


} /* namespace utl */
} /* namespace mdcs */

#define _FACTORY_MAP(MetaPolicy)       utl::FactoryMap<std::string, ::mdcs::utl::AbstractParamData<MetaPolicy>*()>
#define _FACTORY(MetaPolicy)           utl::Singleton<_FACTORY_MAP(MetaPolicy) >


#define INSTANTIATE_META_POLICY(MetaPolicy) \
	template class _FACTORY_MAP(MetaPolicy); \
	template class _FACTORY(MetaPolicy);

#endif /* mdcs_utl_MetaPolicy_hpp_INCLUDED */
