/**
 * @file Policy.hpp
 * @date 18 juil. 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#ifndef mdcs_utl_Policy_hpp_INCLUDED
#define mdcs_utl_Policy_hpp_INCLUDED

#include <utl/Factory.hpp>
#include <utl/Singleton.hpp>
#include <boost/type_traits/is_base_of.hpp>
#include <boost/static_assert.hpp>
#include <memory>

namespace mdcs
{
namespace utl
{

template<class M, class P, class S, class R>
struct PolicyType
{
	typedef M Meta;
	typedef P ParamData;
	typedef S SetupData;
	typedef R RunImpl;
	typedef typename Meta::Interface Interface;
	typedef typename Meta::Context   Context;
	typedef typename Meta::RunInfo   RunInfo;

	static ParamData* creator() { return new ParamData; }
};

template<class Policy>
class BaseRunImpl : public AbstractRunImpl<typename Policy::Meta>
{
public:
	typedef typename Policy::ParamData ParamData;
	typedef typename Policy::SetupData SetupData;
	typedef typename Policy::Context   Context;
	typedef typename Policy::RunInfo   RunInfo;

	BaseRunImpl(Context& /*ctx*/, SetupData& s, const RunInfo& i) : params(s.params), setup(s), runInfo(i) {}
	virtual ~BaseRunImpl() {}

protected:
	const ParamData& params;
	SetupData& setup;
	const RunInfo& runInfo;
};

template<class Policy>
class BaseSetupData : public AbstractSetupData<typename Policy::Meta>
{
public:
	typedef AbstractSetupData<typename Policy::Meta> Base;
	typedef typename Base::RunImplPtr  RunImplPtr;
	typedef typename Policy::ParamData ParamData;
	typedef typename Policy::SetupData SetupData;
	typedef typename Policy::RunImpl   RunImpl;
	typedef typename Policy::Context   Context;
	typedef typename Policy::RunInfo   RunInfo;

	friend class BaseRunImpl<Policy>;

	BaseSetupData(Context& /*ctx*/, const ParamData& p) : params(p)
	{
		BOOST_STATIC_ASSERT((boost::is_base_of<BaseRunImpl<Policy>, RunImpl>::value));
	}
	virtual ~BaseSetupData() {}

	RunImplPtr createRun(Context& context, const RunInfo& runInfo)
	{
		return RunImplPtr(new RunImpl(context, static_cast<SetupData&>(*this), runInfo));
	}

protected:
	const ParamData& params;
};

template<class Policy>
class BaseParamData : public AbstractParamData<typename Policy::Meta>
{
public:
	typedef AbstractParamData<typename Policy::Meta> Base;
	typedef typename Base::SetupDataPtr SetupDataPtr;
	typedef typename Policy::ParamData  ParamData;
	typedef typename Policy::SetupData  SetupData;
	typedef typename Policy::Context    Context;

	BaseParamData()
	{
		BOOST_STATIC_ASSERT((boost::is_base_of<BaseSetupData<Policy>, SetupData>::value));
	}
	virtual ~BaseParamData() {}

	SetupDataPtr createSetup(Context& context) const
	{
		return SetupDataPtr(new SetupData(context, static_cast<const ParamData&>(*this)));
	}

	template<class C> static void init(ParamData& r, const C& config);
};

} /* namespace utl */
} /* namespace mdcs */

#define FORCE_REFERENCE_POLICY(PolicyClass) \
	static ::utl::Once<PolicyClass> BOOST_JOIN(PolicyClass, __LINE__)

#define BEGIN_POLICY_EXT(MetaPolicyClass, PolicyClass, policyName, ParamsClass, SetupClass, RunImplClass) \
	struct PolicyClass \
	{ \
		PolicyClass() { PolicyClass::Policy::Meta::Factory::instance().add(policyName, PolicyClass::Policy::creator); } \
		class ParamsClass; \
		class SetupClass; \
		class RunImplClass; \
		typedef ::mdcs::utl::PolicyType<MetaPolicyClass, ParamsClass, SetupClass, RunImplClass> Policy;

#define BEGIN_POLICY_NAMED(MetaPolicyClass, PolicyClass, policyName) \
	BEGIN_POLICY_EXT(MetaPolicyClass, PolicyClass, policyName, Params, Setup, RunImpl)

#define BEGIN_POLICY(MetaPolicyClass, PolicyClass) \
	BEGIN_POLICY_EXT(MetaPolicyClass, PolicyClass, #PolicyClass, Params, Setup, RunImpl)

#define END_POLICY(PolicyClass) \
	}; \
	BOOST_STATIC_ASSERT((boost::is_base_of< ::mdcs::utl::BaseParamData<PolicyClass::Policy>, PolicyClass::Policy::ParamData>::value)); \
	FORCE_REFERENCE_POLICY(PolicyClass);

#endif /* mdcs_utl_Policy_hpp_INCLUDED */
