/**
 * @file PolicyTuple.hpp
 * @date 22 juil. 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#ifndef mdcs_utl_PolicyTuple_hpp_INCLUDED
#define mdcs_utl_PolicyTuple_hpp_INCLUDED

#include "MetaPolicy.hpp"
#include <boost/tuple/tuple.hpp>
#include <boost/mpl/vector.hpp>
#include <boost/mpl/transform.hpp>
#include <boost/mpl/or.hpp>

namespace mdcs
{
namespace utl
{

template<class PolicyVector, template<class> class Holder>
struct PolicyTuple
{
	typedef typename boost::mpl::reverse_fold<
			PolicyVector,
			boost::tuples::null_type,
			boost::tuples::cons<Holder<boost::mpl::_2>, boost::mpl::_1>
	>::type type;
};

template<class MetaPolicy>
struct ParamsHolder
{
	typedef PolicyParams<MetaPolicy> type;
};
template<class MetaPolicy>
struct ParamsHolder<PolicyArray<MetaPolicy> >
{
	typedef PolicyParamsArray<MetaPolicy> type;
};
template<class MetaPolicy>
struct SetupHolder
{
	typedef PolicySetup<MetaPolicy> type;
};
template<class MetaPolicy>
struct SetupHolder<PolicyArray<MetaPolicy> >
{
	typedef PolicySetupArray<MetaPolicy> type;
};
template<class MetaPolicy>
struct RunImplHolder
{
	typedef PolicyRunImpl<MetaPolicy> type;
};
template<class MetaPolicy>
struct RunImplHolder<PolicyArray<MetaPolicy> >
{
	typedef PolicyRunImplArray<MetaPolicy> type;
};

#ifdef VXWORKS
template<class B, class D>
struct is_base_of : boost::mpl::or_<boost::is_base_of<B, D>, boost::is_same<B, D> > {};
#else
template<class B, class D>
struct is_base_of : boost::is_base_of<B, D> {};
#endif

struct PolicyTupleFunc
{
	template<class Config>
	class ParamsInit
	{
	public:
		ParamsInit(Config& config) : config(config) {}

		template<class MetaPolicy>
		void operator()(PolicyParams<MetaPolicy>& params)
		{
			std::string policyName;
			doInit(params, policyName, getPolicyConfig(params, config, MetaPolicy::Interface::NAME, policyName));
		}
		template<class MetaPolicy>
		void operator()(PolicyParamsArray<MetaPolicy>& params)
		{
			iterate(params, policyIterate(config, MetaPolicy::Interface::NAME));
		}

	private:
		template<class PolicyParams, class PolicyConfig>
		void doInit(PolicyParams& params, const std::string& policyName, const PolicyConfig& cfg)
		{
			params.create(policyName);
			params.init(cfg);
		}
		template<class MetaPolicy, class Iterator>
		void iterate(PolicyParamsArray<MetaPolicy>& params, Iterator iterator)
		{
			while(policyHasNext(iterator))
			{
				std::string policyName;
				doInit(params, policyName, policyNext(iterator, policyName));
			}
		}

		Config& config;
	};
	template<class Context>
	class SetupCreate
	{
	public:
		SetupCreate(Context& context) : context(context) {}

		template<class MetaPolicy>
		void operator()(PolicySetup<MetaPolicy>& setup, const PolicyParams<MetaPolicy>& params)
		{
			BOOST_STATIC_ASSERT((is_base_of<typename PolicySetup<MetaPolicy>::Context, Context>::value));
			setup.create(params, context);
		}
		template<class MetaPolicy>
		void operator()(PolicySetupArray<MetaPolicy>& setup, const PolicyParamsArray<MetaPolicy>& params)
		{
			BOOST_STATIC_ASSERT((is_base_of<typename PolicySetupArray<MetaPolicy>::Context, Context>::value));
			setup.create(params, context);
		}
	private:
		Context& context;
	};
	template<class Context, class RunInfo>
	class RunCreate
	{
	public:
		RunCreate(Context& context, const RunInfo& runInfo) : context(context), runInfo(runInfo) {}

		template<class MetaPolicy>
		void operator()(PolicyRunImpl<MetaPolicy>& run, PolicySetup<MetaPolicy>& setup)
		{
			BOOST_STATIC_ASSERT((is_base_of<typename PolicyRunImpl<MetaPolicy>::Context, Context>::value));
			BOOST_STATIC_ASSERT((is_base_of<typename PolicyRunImpl<MetaPolicy>::RunInfo, RunInfo>::value));
			run.create(setup, context, runInfo);
		}
		template<class MetaPolicy>
		void operator()(PolicyRunImplArray<MetaPolicy>& run, PolicySetupArray<MetaPolicy>& setup)
		{
			BOOST_STATIC_ASSERT((is_base_of<typename PolicyRunImplArray<MetaPolicy>::Context, Context>::value));
			BOOST_STATIC_ASSERT((is_base_of<typename PolicyRunImplArray<MetaPolicy>::RunInfo, RunInfo>::value));
			run.create(setup, context, runInfo);
		}
	private:
		Context& context;
		const RunInfo& runInfo;
	};

	template<class Config>
	static ParamsInit<Config> init(Config& config)
	{
		return ParamsInit<Config>(config);
	}

	template<class Context>
	static SetupCreate<Context> create(Context& context)
	{
		return SetupCreate<Context>(context);
	}

	template<class Context, class RunInfo>
	static RunCreate<Context, RunInfo> create(Context& context, const RunInfo& runInfo)
	{
		return RunCreate<Context, RunInfo>(context, runInfo);
	}
};

} /* namespace utl */
} /* namespace mdcs */

#endif /* mdcs_utl_PolicyTuple_hpp_INCLUDED */
