/**
 * @file CircularBufferTest.cpp
 * @date 15 mai 2013
 * @author fchateau
 * @note
 * SVN tag: $Id$
 *
 * Contributors: Frederic Chateau
 *
 * This file is part of the Mordicus software project.
 *
 * @copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 *
 * @par FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 *
 * @par COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 *
 */

#include <mdcs/utl/CircularBuffer.h>
#include <boost/test/auto_unit_test.hpp>

using namespace mdcs::utl;

struct CircularBufferFixture
{
	CircularBufferFixture()
		: buffer(array, sizeof(array))
	{
	}
	void write(size_t nBytes)
	{
		buffer.moveWritePtr(nBytes);
	}
	void read(size_t nBytes)
	{
		buffer.moveReadPtr(nBytes);
	}

	uint8_t array[4];
	CircularBuffer buffer;
};

bool checkExceptionNotEnoughSpace(const CircularBuffer::Exception& e)
{
	return e.error == CircularBuffer::NotEnoughSpace;
}

bool checkExceptionNotEnoughData(const CircularBuffer::Exception& e)
{
	return e.error == CircularBuffer::NotEnoughData;
}

BOOST_FIXTURE_TEST_SUITE(mdcs_utl_CircularBufferTest, CircularBufferFixture)

BOOST_AUTO_TEST_CASE(base_empty_test)
{
	BOOST_CHECK(buffer.empty());
	BOOST_CHECK(!buffer.full());
	BOOST_CHECK_EQUAL(buffer.size(), 4);
	BOOST_CHECK_EQUAL(buffer.sizeFree(), 3);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousFree(), 3);
	BOOST_CHECK_EQUAL(buffer.sizeOccupied(), 0);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousOccupied(), 0);
}

BOOST_AUTO_TEST_CASE(base_not_full_test)
{
	write(2);
	BOOST_CHECK(!buffer.empty());
	BOOST_CHECK(!buffer.full());
	BOOST_CHECK_EQUAL(buffer.size(), 4);
	BOOST_CHECK_EQUAL(buffer.sizeFree(), 1);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousFree(), 1);
	BOOST_CHECK_EQUAL(buffer.sizeOccupied(), 2);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousOccupied(), 2);
}

BOOST_AUTO_TEST_CASE(base_full_test)
{
	write(3);
	BOOST_CHECK(!buffer.empty());
	BOOST_CHECK(buffer.full());
	BOOST_CHECK_EQUAL(buffer.size(), 4);
	BOOST_CHECK_EQUAL(buffer.sizeFree(), 0);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousFree(), 0);
	BOOST_CHECK_EQUAL(buffer.sizeOccupied(), 3);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousOccupied(), 3);
}

BOOST_AUTO_TEST_CASE(base_overflow_test)
{
	BOOST_CHECK_EXCEPTION(write(4), CircularBuffer::Exception, checkExceptionNotEnoughSpace);
}

BOOST_AUTO_TEST_CASE(base_underflow_test)
{
	BOOST_CHECK_EXCEPTION(read(1), CircularBuffer::Exception, checkExceptionNotEnoughData);
}

BOOST_AUTO_TEST_CASE(shifted_not_full_test)
{
	write(2);
	read(1);
	BOOST_CHECK(!buffer.empty());
	BOOST_CHECK(!buffer.full());
	BOOST_CHECK_EQUAL(buffer.size(), 4);
	BOOST_CHECK_EQUAL(buffer.sizeFree(), 2);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousFree(), 2);
	BOOST_CHECK_EQUAL(buffer.sizeOccupied(), 1);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousOccupied(), 1);
}

BOOST_AUTO_TEST_CASE(shifted_empty_test)
{
	write(1);
	read(1);
	BOOST_CHECK(buffer.empty());
	BOOST_CHECK(!buffer.full());
	BOOST_CHECK_EQUAL(buffer.size(), 4);
	BOOST_CHECK_EQUAL(buffer.sizeFree(), 3);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousFree(), 3);
	BOOST_CHECK_EQUAL(buffer.sizeOccupied(), 0);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousOccupied(), 0);
}

BOOST_AUTO_TEST_CASE(end_not_full_test)
{
	write(3);
	read(2);
	write(1);
	BOOST_CHECK(!buffer.empty());
	BOOST_CHECK(!buffer.full());
	BOOST_CHECK_EQUAL(buffer.size(), 4);
	BOOST_CHECK_EQUAL(buffer.sizeFree(), 1);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousFree(), 1);
	BOOST_CHECK_EQUAL(buffer.sizeOccupied(), 2);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousOccupied(), 2);
}

BOOST_AUTO_TEST_CASE(end_empty_test)
{
	write(3);
	read(3);
	BOOST_CHECK(buffer.empty());
	BOOST_CHECK(!buffer.full());
	BOOST_CHECK_EQUAL(buffer.size(), 4);
	BOOST_CHECK_EQUAL(buffer.sizeFree(), 3);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousFree(), 1);
	BOOST_CHECK_EQUAL(buffer.sizeOccupied(), 0);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousOccupied(), 0);
}

BOOST_AUTO_TEST_CASE(splitted_not_full_test)
{
	write(3);
	read(3);
	write(2);
	BOOST_CHECK(!buffer.empty());
	BOOST_CHECK(!buffer.full());
	BOOST_CHECK_EQUAL(buffer.size(), 4);
	BOOST_CHECK_EQUAL(buffer.sizeFree(), 1);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousFree(), 1);
	BOOST_CHECK_EQUAL(buffer.sizeOccupied(), 2);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousOccupied(), 1);
}

BOOST_AUTO_TEST_CASE(splitted_full_test)
{
	write(3);
	read(3);
	write(3);
	BOOST_CHECK(!buffer.empty());
	BOOST_CHECK(buffer.full());
	BOOST_CHECK_EQUAL(buffer.size(), 4);
	BOOST_CHECK_EQUAL(buffer.sizeFree(), 0);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousFree(), 0);
	BOOST_CHECK_EQUAL(buffer.sizeOccupied(), 3);
	BOOST_CHECK_EQUAL(buffer.sizeNextContiguousOccupied(), 1);
}

BOOST_AUTO_TEST_SUITE_END()
