// =============================================================================
// BitFieldHelper.hpp
// -----------------------------------------------------------------------------
// class utl::BitFieldHelper
// -----------------------------------------------------------------------------
// Developed at the CEA (Saclay, France), in the Irfu/Sedi/Lilas laboratory, by:
// Frederic Chateau (frederic.chateau@cea.fr)
// -----------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation.
// =============================================================================

#ifndef utl_BitFieldHelper_h_INCLUDED
#define utl_BitFieldHelper_h_INCLUDED

#include "utl/Logging.h"
#include <boost/type_traits/make_signed.hpp>
#include <boost/type_traits/make_unsigned.hpp>
#include <boost/type_traits/is_signed.hpp>
#include <boost/type_traits/remove_cv.hpp>
#include <boost/mpl/if.hpp>
#include <climits>

namespace utl
{
/**
 * This class provides template methods to manipulate bit-fields easily and efficiently.
 */
template<class WordType>
class BitFieldHelper
{
public:
	typedef typename boost::remove_cv<WordType>::type    WordBasicType;
	typedef typename boost::make_signed<WordBasicType>::type   WordSignedType;
	typedef typename boost::make_unsigned<WordBasicType>::type WordUnsignedType;

	/**
	 * Compile-time mapping that returns the signed or unsigned version of WordType
	 * depending on the sign of the specified template parameter.
	 */
	template<typename ValueType>
	struct WordTypeMatchingSignOf
		: boost::mpl::if_c<
			boost::is_signed<ValueType>::value,
			WordSignedType,
			WordUnsignedType>
	{
	};

	static const WordBasicType ZERO = 0;
	static const WordBasicType ONE = 1;
	static const WordBasicType ALL = ~0;
	static const unsigned SIZEOF = sizeof(WordBasicType);
	static const unsigned BITCOUNT = SIZEOF * CHAR_BIT;

	///@name Compile-time bit-field manipulation.
	///@{
	template<unsigned pos>
	static bool getBit(WordBasicType var);

	template<unsigned pos>
	static void clearBit(WordType& var);

	template<unsigned pos>
	static void setBit(WordType& var);

	template<unsigned pos>
	static void setBit(WordType& var, bool askSet);

	template<unsigned pos, unsigned width>
	static WordBasicType getField(WordBasicType var);

	template<unsigned pos, unsigned width, typename FieldType>
	static FieldType getField(WordBasicType var);

	template<unsigned pos, unsigned width, typename FieldType>
	static void getField(WordBasicType var, FieldType& result);

	template<unsigned pos, unsigned width>
	static void setField(WordType& var, WordBasicType value);

	template<unsigned pos, unsigned width>
	static WordBasicType getReadMask();

	template<unsigned pos, unsigned width>
	static WordBasicType getClearMask();
	///@}

	///@name Run-time bit-field manipulation.
	///@{
	static bool getBit(WordBasicType var, unsigned pos);

	static void setBit(WordType& var, unsigned pos, bool askSet);

	static WordBasicType getField(WordBasicType var, unsigned pos, unsigned width);

	template<typename FieldType>
	static FieldType getField(WordBasicType var, unsigned pos, unsigned width);

	template<typename FieldType>
	static void getField(WordBasicType var, unsigned pos, unsigned width, FieldType& result);

	static void setField(WordType& var, unsigned pos, unsigned width, WordBasicType value);

	static WordBasicType getReadMask(unsigned pos, unsigned width);

	static WordBasicType getClearMask(unsigned pos, unsigned width);
	///@}

private:
	BitFieldHelper() {}
	~BitFieldHelper() {}
};

// =============================================================================
// utl::BitFieldHelper inline methods body
// -----------------------------------------------------------------------------

/**
 * Returns the value of a bit of an integer variable (compile-time version).
 */
template<class WordType>
template<unsigned pos>
inline bool BitFieldHelper<WordType>::getBit(WordBasicType var)
{
	return (var & (ONE << pos)) != 0;
}

/**
 * Returns the value of a bit of an integer variable.
 */
template<class WordType>
inline bool BitFieldHelper<WordType>::getBit(WordBasicType var, unsigned pos)
{
	return (var & (ONE << pos)) != 0;
}

/**
 * Clears the bit at the specified position.
 */
template<class WordType>
template<unsigned pos>
inline void BitFieldHelper<WordType>::clearBit(WordType& var)
{
	var &= ~(ONE << pos);
}

/**
 * Sets the bit at the specified position.
 */
template<class WordType>
template<unsigned pos>
inline void BitFieldHelper<WordType>::setBit(WordType& var)
{
	var |= ONE << pos;
}

/**
 * Sets or clears a bit of a variable (compile-time version).
 */
template<class WordType>
template<unsigned pos>
inline void BitFieldHelper<WordType>::setBit(WordType& var, bool askSet)
{
	if(askSet) setBit<pos>(var);
	else       clearBit<pos>(var);
}

/**
 * Sets or clears a bit of a variable.
 */
template<class WordType>
inline void BitFieldHelper<WordType>::setBit(WordType& var, unsigned pos, bool askSet)
{
	if(askSet) var |=   ONE << pos;
	else       var &= ~(ONE << pos);
}

/**
 * Returns the value of a bit-field, with the same type as the input (compile-time version).
 */
template<class WordType>
template<unsigned pos, unsigned width>
inline typename BitFieldHelper<WordType>::WordBasicType
BitFieldHelper<WordType>::getField(WordBasicType var)
{
	WordBasicType result;
	getField<pos, width>(var, result);
	return result;
}
/**
 * Returns the value of a bit-field, with the same type as the input.
 */
template<class WordType>
inline typename BitFieldHelper<WordType>::WordBasicType
BitFieldHelper<WordType>::getField(WordBasicType var, unsigned pos, unsigned width)
{
	WordBasicType result;
	getField(var, pos, width, result);
	return result;
}

/**
 * Returns the value of a bit-field, allowing the field to have a different type (compile-time version).
 */
template<class WordType>
template<unsigned pos, unsigned width, typename FieldType>
inline FieldType BitFieldHelper<WordType>::getField(WordBasicType var)
{
	FieldType result;
	getField<pos, width>(var, result);
	return result;
}

/**
 * Returns the value of a bit-field, allowing the field to have a different type.
 */
template<class WordType>
template<typename FieldType>
inline FieldType BitFieldHelper<WordType>::getField(WordBasicType var, unsigned pos, unsigned width)
{
	FieldType result;
	getField(var, pos, width, result);
	return result;
}

/**
 * Retrieves the value of a bit-field (compile-time version).
 * The field does not need to have the same type (or same signed-ness) as the input.
 */
template<class WordType>
template<unsigned pos, unsigned width, typename FieldType>
inline void BitFieldHelper<WordType>::getField(WordBasicType var, FieldType& result)
{
	typedef typename WordTypeMatchingSignOf<FieldType>::type SignCorrectWordType;

	result = static_cast<SignCorrectWordType>((var << (BITCOUNT - width - pos))) >> (BITCOUNT - width);
}

/**
 * Retrieves the value of a bit-field.
 * The field does not need to have the same type (or same signed-ness) as the input.
 */
template<class WordType>
template<typename FieldType>
inline void BitFieldHelper<WordType>::getField(WordBasicType var, unsigned pos, unsigned width, FieldType& result)
{
	typedef typename WordTypeMatchingSignOf<FieldType>::type SignCorrectWordType;

	result = static_cast<SignCorrectWordType>((var << (BITCOUNT - width - pos))) >> (BITCOUNT - width);
}

/**
 * Sets the value of a bit-field of a variable (compile-time version).
 */
template<class WordType>
template<unsigned pos, unsigned width>
inline void BitFieldHelper<WordType>::setField(WordType& var, WordBasicType value)
{
	var = (var & getClearMask<pos, width>()) | ((value << pos) & getReadMask<pos, width>());
}

/**
 * Sets the value of a bit-field of a variable.
 */
template<class WordType>
inline void BitFieldHelper<WordType>::setField(WordType& var, unsigned pos, unsigned width, WordBasicType value)
{
	var = (var & getClearMask(pos, width)) | ((value << pos) & getReadMask(pos, width));
}

/**
 * Returns the mask used to isolate the bit-field range, in order to read it (compile-time version).
 * Protection against the machine-dependent case (and seen at run-time)
 * for which width == size of the word type
 */
template<class WordType>
template<unsigned pos, unsigned width>
inline typename BitFieldHelper<WordType>::WordBasicType
BitFieldHelper<WordType>::getReadMask()
{
	return (width >= BITCOUNT) ? ALL : ((~(ALL << width)) << pos);
}

/**
 * Returns the mask used to isolate the bit-field range, in order to read it.
 * Protection against the machine-dependent case (and seen at run-time)
 * for which width == size of the word type
 */
template<class WordType>
inline typename BitFieldHelper<WordType>::WordBasicType
BitFieldHelper<WordType>::getReadMask(unsigned pos, unsigned width)
{
	return (width >= BITCOUNT) ? ALL : ((~(ALL << width)) << pos);
}

/**
 * Returns the mask used to exclude the bit-field range, in order to clear it (compile-time version).
 */
template<class WordType>
template<unsigned pos, unsigned width>
inline typename BitFieldHelper<WordType>::WordBasicType
BitFieldHelper<WordType>::getClearMask()
{
	return ~getReadMask<pos, width>();
}

/**
 * Returns the mask used to exclude the bit-field range, in order to clear it.
 */
template<class WordType>
inline typename BitFieldHelper<WordType>::WordBasicType
BitFieldHelper<WordType>::getClearMask(unsigned pos, unsigned width)
{
	return ~getReadMask(pos, width);
}

}

#endif // utl_BitFieldHelper_h_INCLUDED
