// =============================================================================
// BitFieldRegister.hpp
// -----------------------------------------------------------------------------
// class utl::BitFieldRegister
// -----------------------------------------------------------------------------
// Developed at the CEA (Saclay, France), in the Irfu/Sedi/Lilas laboratory, by:
// Frederic Chateau (frederic.chateau@cea.fr)
// -----------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation.
// =============================================================================

#ifndef utl_BitFieldRegister_INCLUDED
#define utl_BitFieldRegister_INCLUDED

#include "utl/BitFieldHelper.hpp"
#include <boost/type_traits/remove_volatile.hpp>

namespace utl
{
/**
 * This should be used as the base class of datatypes that implements a
 * bit field semantics.
 * It is designed to be used in conjunction with DECLARE_BIT and DECLARE_FIELD
 * macros. These macro must be called in the subclass body. They generate
 * accessors for each bit and fields.
 */
template<typename T>
class BitFieldRegister
{
public:
	typedef T WordType;
	typedef typename boost::remove_volatile<WordType>::type WordBasicType;
	typedef utl::BitFieldHelper<WordType> RegField;

	BitFieldRegister(WordType value = 0) : regValue(value) {}
	BitFieldRegister(const BitFieldRegister<T>& reg) : regValue(reg.value()) {}

	BitFieldRegister<T>& operator=(const BitFieldRegister<T>& reg) { regValue = reg.value(); return *this; }

	void setValue(WordType val) { regValue = val; }
	WordType value() const { return regValue; }

protected:
	WordType regValue;
};

#define DECLARE_BIT(name, pos) \
	bool name() const { return RegField::getBit<pos>(regValue); } \
	void set_##name(bool val) { RegField::setBit<pos>(regValue, val); }

#define DECLARE_FIELD(name, pos, width) \
	WordBasicType name() const { return RegField::getField<pos, width>(regValue); } \
	void set_##name(WordType val) { RegField::setField<pos, width>(regValue, val); }

}

#endif // utl_BitFieldRegister_INCLUDED
