/* =====================================================================================================================
 * Logging.h
 * ---------------------------------------------------------------------------------------------------------------------
 * class utl::Logger and macros for logging helpers.
 * Created on: Jan 11, 2012 at Irfu/Sedi/Lilas, CEA Saclay, F-91191, France
 * ---------------------------------------------------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * ---------------------------------------------------------------------------------------------------------------------
 * Contributors: Shebli Anvar (shebli.anvar@cea.fr)
 *               Frederic Chateau (frederic.chateau@cea.fr)
 * ---------------------------------------------------------------------------------------------------------------------
 * This software is part of the the Mordicus Real-Time Framework general utilities.
 * ---------------------------------------------------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 * ---------------------------------------------------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 * =====================================================================================================================
 */

#ifndef utl_Logging_h_INCLUDED
#define utl_Logging_h_INCLUDED

#include <string>
#include <iostream>
#include <memory>

namespace utl
{
enum SeverityLevel
{
	debug = 0,
	info,
	warn,
	error,
	fatal
};

class AbstractLoggingBackend;

typedef std::auto_ptr<AbstractLoggingBackend> LoggingBackendPtr;

/**
 * This class allows to declare log expressions that will be processed by
 * a parameterable global backend.
 * It is supposed to be instanciated at the beginning of each expression and
 * destroyed at the end.
 */
class BackendLogger
{
public:
	BackendLogger(SeverityLevel level, const char* source, unsigned line);
	~BackendLogger();

	template<class T>
	BackendLogger& operator<<(const T& val);

	static AbstractLoggingBackend* backend();
	static void setBackend(LoggingBackendPtr b);

private:
	static void destroyBackend();

	static AbstractLoggingBackend* backend_;
	std::ostream& stream;
};

/**
 * This class implements a very simple log expression that just outputs the
 * expression on stdout and appends it an endline.
 */
class StdoutLogger
{
public:
	~StdoutLogger() { std::cout << std::endl; }
	template<class T>
	StdoutLogger& operator<<(const T& val)
	{
		std::cout << val;
		return *this;
	}
};

/**
 * This class defines the interface for logging backends.
 * Classes deriving from it will have to implement how the log lines shall be
 * displayed/stored.
 */
class AbstractLoggingBackend
{
public:
	virtual ~AbstractLoggingBackend() = 0;
	virtual std::ostream& beginLog(SeverityLevel level, const char* source, unsigned line) = 0;
	virtual void endLog() = 0;
};

/**
 * This is a logging backend based on stdout. It ensures logs are written on stdout.
 * It is used as the default backend implementation for BackendLogger.
 */
class StdoutLoggingBackend : public AbstractLoggingBackend
{
public:
	std::ostream& beginLog(SeverityLevel level, const char* source, unsigned line);
	void endLog();
};

std::string xtract_base(const std::string& filePath);

// =============================================================================
// Inline methods of class BackendLogger
// -----------------------------------------------------------------------------
#ifndef __CINT__ // Let the ROOT dictionary generator rootcint  skip this
inline BackendLogger::BackendLogger(SeverityLevel level, const char* source, unsigned line)
	: stream(backend()->beginLog(level, source, line))
{
}

inline BackendLogger::~BackendLogger()
{
	backend_->endLog();
}

template<class T>
inline BackendLogger& BackendLogger::operator<<(const T& val)
{
	stream << val;
	return *this;
}
#endif // __CINT__
}

// =============================================================================
// Macros
// -----------------------------------------------------------------------------

#ifndef UTL_LOG_LEVEL
# ifdef UTL_LOGGING_ENABLED
#  define UTL_LOG_LEVEL 0
# else
#  define UTL_LOG_LEVEL 10
# endif
#else
# ifndef UTL_LOGGING_ENABLED
#  define UTL_LOGGING_ENABLED
# endif
#endif

#define LOG_DEBUG() if(UTL_LOG_LEVEL > 0) ; else ::utl::BackendLogger(::utl::debug, __FILE__, __LINE__)
#define LOG_INFO()  if(UTL_LOG_LEVEL > 1) ; else ::utl::BackendLogger(::utl::info,  __FILE__, __LINE__)
#define LOG_WARN()  if(UTL_LOG_LEVEL > 2) ; else ::utl::BackendLogger(::utl::warn,  __FILE__, __LINE__)
#define LOG_ERROR() if(UTL_LOG_LEVEL > 3) ; else ::utl::BackendLogger(::utl::error, __FILE__, __LINE__)
#define LOG_FATAL() if(UTL_LOG_LEVEL > 4) ; else ::utl::BackendLogger(::utl::fatal, __FILE__, __LINE__)

#define MSG() ::utl::BackendLogger(::utl::info,  __FILE__, __LINE__)
#define MSG_IF(cond)  if(!(cond)) ; else ::utl::BackendLogger(::utl::info,  __FILE__, __LINE__)

#ifdef UTL_LOGGING_ENABLED
# define LOG(expr) ::utl::StdoutLogger() << expr;
#endif

#if UTL_LOG_LEVEL < 1
# define MLOG_DEBUG(expr) ::utl::BackendLogger(::utl::debug, __FILE__, __LINE__) << expr;
#endif
#if UTL_LOG_LEVEL < 2
# define MLOG_INFO(expr)  ::utl::BackendLogger(::utl::info,  __FILE__, __LINE__) << expr;
#endif
#if UTL_LOG_LEVEL < 3
# define MLOG_WARN(expr)  ::utl::BackendLogger(::utl::warn,  __FILE__, __LINE__) << expr;
#endif
#if UTL_LOG_LEVEL < 4
# define MLOG_ERROR(expr) ::utl::BackendLogger(::utl::error, __FILE__, __LINE__) << expr;
#endif
#if UTL_LOG_LEVEL < 5
# define MLOG_FATAL(expr) ::utl::BackendLogger(::utl::fatal, __FILE__, __LINE__) << expr;
#endif

#ifndef LOG
# define LOG(expr)
#endif
#ifndef MLOG_DEBUG
# define MLOG_DEBUG(expr)
#endif
#ifndef MLOG_INFO
# define MLOG_INFO(expr)
#endif
#ifndef MLOG_WARN
# define MLOG_WARN(expr)
#endif
#ifndef MLOG_ERROR
# define MLOG_ERROR(expr)
#endif
#ifndef MLOG_FATAL
# define MLOG_FATAL(expr)
#endif

#define PROMPT_MSG(expr) LOG("--- " << ::utl::xtract_base(__FILE__) << ": " << expr)
#define PROMPT_ERR(expr) LOG("*** " << ::utl::xtract_base(__FILE__) << ": " << expr)

#endif // utl_Logging_h_INCLUDED
