/**
 * @file RtBitSet.h
 * @created Nov 8, 2012
 * @author anvar
 * ---------------------------------------------------------------------------------------------------------------------
 * © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
 * ---------------------------------------------------------------------------------------------------------------------
 * Contributors: Shebli Anvar (shebli.anvar@cea.fr)
 * ---------------------------------------------------------------------------------------------------------------------
 * This software is part of the Utilities software project.
 * ---------------------------------------------------------------------------------------------------------------------
 * FREE SOFTWARE LICENCING
 * This software is governed by the CeCILL license under French law and abiding  * by the rules of distribution of free
 * software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
 * CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
 * and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
 * and the software's author, the holder of the economic rights, and the successive licensors have only limited
 * liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
 * and/or developing or reproducing the software by the user in light of its specific status of free software, that may
 * mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
 * experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
 * to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
 * you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
 * ---------------------------------------------------------------------------------------------------------------------
 * COMMERCIAL SOFTWARE LICENCING
 * You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
 * negotiate a specific contract with a legal representative of CEA.
 * =====================================================================================================================
 */

#ifndef utl_RtBitSet_h_INCLUDED
#define utl_RtBitSet_h_INCLUDED

#include "utl/Logging.h"
#include "utl/WordIterator.hpp"
#include "utl/BitFieldHelper.hpp"
#include "utl/BinIO.hpp"
#include <boost/type_traits/make_signed.hpp>
#include <boost/type_traits/make_unsigned.hpp>
#include <boost/type_traits/is_signed.hpp>
#include <boost/type_traits/conditional.hpp>
#include <boost/math/common_factor_rt.hpp>
#include <vector>
#include <algorithm>
#include <iostream>
#include <iomanip>
#include <climits>

namespace utl
{

struct AtomOperator
{
	struct copy
	{
		template < typename AtomType >
		AtomType& operator() (AtomType& targetAtom, AtomType sourceAtom) { targetAtom  = sourceAtom; return targetAtom; }
	};

	struct bitAnd
	{
		template < typename AtomType >
		AtomType& operator() (AtomType& targetAtom, AtomType sourceAtom) { targetAtom &= sourceAtom; return targetAtom; }
	};

	struct bitOr
	{
		template < typename AtomType >
		AtomType& operator() (AtomType& targetAtom, AtomType sourceAtom) { targetAtom |= sourceAtom; return targetAtom; }
	};

	struct bitXor
	{
		template < typename AtomType >
		AtomType& operator() (AtomType& targetAtom, AtomType sourceAtom) { targetAtom ^= sourceAtom; return targetAtom; }
	};

	struct bitNot
	{
		template < typename AtomType >
		AtomType& operator() (AtomType& targetAtom, AtomType sourceAtom) { targetAtom = ~sourceAtom; return targetAtom; }
	};
};


/**
 * This enum type allows for an explicit distinction between signed and unsigned data.
 */
enum Signedness { SIGNED = -1, UNSIGNED = 1 };

/**
 * Forward declarations for friend statements and helpers
 */
template < class W > class BitFieldHelper;
template < typename W > struct make_RtBitSet;

/**
 * Template class representing a "word" made of an arbitrary number of bits set at run time (run time bitset)
 * @tparam S Indicates whether the bitset should be considered as a signed or unsigned word by taking on one of the 2
 *           enum values SIGNED or UNSIGNED.
 */
template < Signedness S = SIGNED >
class RtBitSet
{
public:
	/**
	 * Depending on its signed or unsigned character of this bitset will be made of a vector of signed or unsigned
	 * words. Atom refers to the type of this elementary word.
	 */
	typedef typename boost::conditional< S == SIGNED, signed char, unsigned char >::type Byte;
	typedef typename boost::conditional< S == SIGNED, int, unsigned >::type Atom;
	typedef typename boost::make_unsigned <Atom>::type unsigned_Atom; ///< Unsigned version of atom word.
	typedef typename boost::make_signed   <Atom>::type   signed_Atom; ///< Signed version of atom word.
	typedef WordIterator<Atom> iter_Atom;

	friend class BitFieldHelper< std::vector<Atom> >;

	/// The signed or unisgned character of this bitset is also stored in this class variable.
	static const Signedness signedness = S;

public:
	///@name Constructors, initializer & normalizer
	///@{
	RtBitSet(size_t _bitCount = 0);

	template < typename W2 >
	RtBitSet(size_t _bitCount, const W2& initWord);

	template < typename W2 >
	RtBitSet(const W2& initWord);

	template < typename W2 >
	RtBitSet(size_t _bitCount, const std::vector<W2>& initVect);

	template < typename W2 >
	RtBitSet(const std::vector<W2>& initVect);

	template < Signedness S2 >
	RtBitSet(size_t _bitCount, const RtBitSet<S2>& initRtBitSet);

	template < Signedness S2 >
	RtBitSet(const RtBitSet<S2>& initRtBitSet);

	void setBitCount(size_t _bitCount);
	void normalize();
	///@}

	///@name Value getters and setters
	///@{
	template < typename W2 >
	void setValue(const W2& w) { rasterOp<AtomOperator::copy, W2>(w); }
	template < typename W2 >
	void setValue(const std::vector<W2>& sourceVect) { rasterOp<AtomOperator::copy, W2>(sourceVect); }
	void setValue(const unsigned_Atom& w) { rasterOp<AtomOperator::copy>(w); }
	void setValue(const signed_Atom& w) { rasterOp<AtomOperator::copy>(w); }
	template < Signedness S_Source >
	void setValue(const RtBitSet<S_Source>& bitSet) { rasterOp<AtomOperator::copy, S_Source>(bitSet); }
	template < typename AtomType >
	void setValue(const AtomType* sourceArray, size_t arraySize) { rasterOp<AtomOperator::copy, AtomType>(sourceArray, arraySize); }

	template < typename W2 >
	void getValue(W2& w) const { rasterOp<AtomOperator::copy, W2>(w); }
	template < typename W2 >
	void getValue(std::vector<W2>& targetVect) const { rasterOp<AtomOperator::copy, W2>(targetVect); }
	void getValue(unsigned_Atom& w) const { rasterOp<AtomOperator::copy>(w); }
	void getValue(signed_Atom& w) const { rasterOp<AtomOperator::copy>(w); }
	template < Signedness S_Source >
	void getValue(RtBitSet<S_Source>& bitSet) const { bitSet.setValue(*this); }
	template < typename AtomType >
	void getValue(AtomType* targetArray, size_t arraySize) const { rasterOp<AtomOperator::copy, AtomType>(targetArray, arraySize); }
	///@}

	///@name Bitwise operations
	///@{
	template < typename W2 >
	void bitAnd(const W2& w) { rasterOp<AtomOperator::bitAnd, W2>(w); }
	void bitAnd(const unsigned_Atom& w) { rasterOp<AtomOperator::bitAnd>(w); }
	void bitAnd(const signed_Atom& w) { rasterOp<AtomOperator::bitAnd>(w); }
	template < Signedness S_Source >
	void bitAnd(const RtBitSet<S_Source>& bitSet) { rasterOp<AtomOperator::bitAnd, S_Source>(bitSet); }

	template < typename W2 >
	void bitOr(const W2& w) { rasterOp<AtomOperator::bitOr, W2>(w); }
	void bitOr(const unsigned_Atom& w) { rasterOp<AtomOperator::bitOr>(w); }
	void bitOr(const signed_Atom& w) { rasterOp<AtomOperator::bitOr>(w); }
	template < Signedness S_Source >
	void bitOr(const RtBitSet<S_Source>& bitSet) { rasterOp<AtomOperator::bitOr, S_Source>(bitSet); }

	template < typename W2 >
	void bitXor(const W2& w) { rasterOp<AtomOperator::bitXor, W2>(w); }
	void bitXor(const unsigned_Atom& w) { rasterOp<AtomOperator::bitXor>(w); }
	void bitXor(const signed_Atom& w) { rasterOp<AtomOperator::bitXor>(w); }
	template < Signedness S_Source >
	void bitXor(const RtBitSet<S_Source>& bitSet) { rasterOp<AtomOperator::bitXor, S_Source>(bitSet); }

	void bitNot();
	void shiftLeft  (size_t nBits);
	void shiftRight (size_t nBits);

	RtBitSet& operator <<= (size_t nBits) { shiftLeft (nBits); return *this; }
	RtBitSet& operator >>= (size_t nBits) { shiftRight(nBits); return *this; }
	RtBitSet  operator   ~ ()  { bitNot(); return *this; }
	RtBitSet  operator  << (size_t nBits) const { RtBitSet bitSetTmp(*this); bitSetTmp <<= nBits; return bitSetTmp; }
	RtBitSet  operator  >> (size_t nBits) const { RtBitSet bitSetTmp(*this); bitSetTmp >>= nBits; return bitSetTmp; }
	template < typename W2 >
	RtBitSet& operator  &= (const W2& w) { bitAnd(w); return *this; }
	template < typename W2 >
	RtBitSet& operator  |= (const W2& w) { bitOr (w); return *this; }
	template < typename W2 >
	RtBitSet& operator  ^= (const W2& w) { bitXor(w); return *this; }
	template < typename W2 >
	RtBitSet  operator   & (const W2& w) const { RtBitSet bitSetTmp(*this); bitSetTmp &= w; return bitSetTmp; }
	template < typename W2 >
	RtBitSet  operator   | (const W2& w) const { RtBitSet bitSetTmp(*this); bitSetTmp |= w; return bitSetTmp; }
	template < typename W2 >
	RtBitSet  operator   ^ (const W2& w) const { RtBitSet bitSetTmp(*this); bitSetTmp ^= w; return bitSetTmp; }
	///@}

	///@name Comparison operations
	///@{
	template < typename W2 >
	int compare(const W2& w) const;
	template < Signedness S2 >
	int compare(const RtBitSet<S2>& bitSet) const;
	template < typename W2 >
	bool isEqualTo(const W2& w) const { return compare(w) == 0; }
	template < typename W2 >
	bool isGreaterThan(const W2& w) const { return compare(w) > 0; }
	template < typename W2 >
	RtBitSet& operator = (const W2& w) { setValue(w); return *this; }
	template < typename W2 >
	bool operator == (const W2& w) const { return isEqualTo(w); }
	template < typename W2 >
	bool operator != (const W2& w) const { return not isEqualTo(w); }
	template < typename W2 >
	bool operator > (const W2& w) const { return isGreaterThan(w); }
	template < typename W2 >
	bool operator >= (const W2& w) const { return isGreaterThan(w) or isEqualTo(w); }
	template < typename W2 >
	bool operator <= (const W2& w) const { return not isGreaterThan(w); }
	template < typename W2 >
	bool operator < (const W2& w) const { return not isGreaterThan(w) and not isEqualTo(w); }
	///@}

	///@name Elementary word (atom) access & iteration operations
	///@{
	iter_Atom itAtomBegin() const { return iter_Atom::getBegin(buffer); }
	iter_Atom itAtomEnd  () const { return iter_Atom::getEnd  (buffer); }
	iter_Atom itAtomLast () const { return itAtomEnd() - 1; }

	      Atom& atomAt(std::ptrdiff_t iAtom)       { return itAtomBegin()[iAtom]; }
    const Atom& atomAt(std::ptrdiff_t iAtom) const { return itAtomBegin()[iAtom]; }

	      Atom& atomFirst()       { return atomAt(0); }
	const Atom& atomFirst() const { return atomAt(0); }

	      Atom& atomLast ()       { return atomAt(atomCount()-1); }
	const Atom& atomLast () const { return atomAt(atomCount()-1); }
	///@}

	///@name Miscellaneous operations
	///@{
	template < typename W2 >
	static unsigned wordByteCount(const W2& val = 0) { return sizeof(val); }

	template < typename W2 >
	static unsigned wordBitCount(const W2& val = 0) { return wordByteCount(val) * CHAR_BIT; }

	static unsigned atomByteCount() { return sizeof(Atom); }
	static unsigned atomBitCount() { return atomByteCount() * CHAR_BIT; }

	size_t bitCount () const { return bitCount_; }
	size_t byteCount() const { return atomCount() * atomByteCount(); }
	size_t atomCount() const { return buffer.size(); }
	size_t minByteCount() const { return bitCount()%CHAR_BIT == 0 ? bitCount()/CHAR_BIT : 1 + bitCount()/CHAR_BIT; }

	template < typename W2 >
	static bool isPositive(const W2& w) { return ((w >> 1) & (1 << (wordBitCount(w)-1))) == 0; }

	template < typename W2 >
	static bool isNegative(const W2& w) { return not isPositive(w); }

	bool isPositive() { return bitCount() == 0 or isPositive(atomLast()); }
	bool isNegative() { return not isPositive(); }

	static bool isBigEndian() { return iter_Atom::isBigEndian(); }

	template < typename W2 >
	W2* bufferPtr() { if (bitCount() > 0) return reinterpret_cast<W2*>(&buffer[0]); return 0; }

	template < typename W2 >
    const W2* bufferPtr() const { if (bitCount() > 0) return reinterpret_cast<const W2*>(&buffer[0]); return 0; }

	void print(std::ostream& os) const;

	template <typename W2>
	W2 toScalar() const { W2 w; getValue(w); return w; }
	///@}

	///@name Bit field access operations
	///@{
	bool getBit(unsigned pos) const;
	void setBit(unsigned pos, bool bitVal);

	template < typename value_t >
	void getField(unsigned pos, unsigned wid, value_t& value) const;

	template < typename value_t >
	value_t getField(unsigned pos, unsigned wid) const;

	template < typename value_t >
	void setField(unsigned pos, unsigned wid, const value_t& value);
	///@}

protected:
	///@name Raster (array) operations
	///@{
	template < class AtomOp, typename AtomType >
	static void rasterOp(AtomType* pTarget, size_t maxAtoms, const AtomType* pSource, size_t nAtoms);

	template < class AtomOp, typename AtomType >
	void rasterOp(const AtomType* sourceArray, size_t arraySize);

	template < class AtomOp, typename AtomType >
	void rasterOp(AtomType* targetArray, size_t arraySize) const;

	template < class AtomOp, typename AtomType >
	void rasterOp(const AtomType& sourceAtom) { rasterOp<AtomOp, AtomType>(&sourceAtom, 1u); }

	template < class AtomOp, typename AtomType >
	void rasterOp(AtomType& targetAtom) const { rasterOp<AtomOp, AtomType>(&targetAtom, 1u); }

	template < class AtomOp, typename SourceWord >
	void rasterOp(const std::vector<SourceWord>& sourceVect);

	template < class AtomOp, typename TargetWord >
	void rasterOp(std::vector<TargetWord>& targetVect) const;

	template < class AtomOp >
	void rasterOp(const unsigned_Atom& sourceAtom);

	template < class AtomOp >
	void rasterOp(unsigned_Atom& targetAtom) const;

	template < class AtomOp >
	void rasterOp(const signed_Atom& sourceAtom);

	template < class AtomOp >
	void rasterOp(signed_Atom& targetAtom) const;

	template < class AtomOp, Signedness S_Source >
	void rasterOp(const RtBitSet<S_Source>& sourceRtBitSet);
	///@}

private:
	size_t bitCount_;
	std::vector<Atom> buffer;
};

/**
 * Constructs a bitset given its total number of bits.
 * @param _bitCount The total number of bit of this RtBitSet instance.
 */
template < Signedness S >
RtBitSet<S>::RtBitSet(size_t _bitCount)
{
	setBitCount(_bitCount);
	setValue(0);
}

/**
 * Template constructor. The bitset is given its total number of bits and a word of the template parameter type to
 * initialize its value.
 * @tparam W2 The type of the initialization word.
 * @param _bitCount The total number of bit of this RtBitSet instance.
 * @param initWord The initialization word of the bitset.
 */
template < Signedness S >
template < typename W2 >
RtBitSet<S>::RtBitSet(size_t _bitCount, const W2& initWord)
{
	setBitCount(_bitCount);
	setValue(initWord);
}

/**
 * Template constructor. The bitset is given an initializer word.
 * @param initWord The initialization word of the bitset.
 */
template < Signedness S >
template < typename W2 >
RtBitSet<S>::RtBitSet(const W2& initWord)
{
	setBitCount(wordBitCount(initWord));
	setValue(initWord);
}

/**
 * Template constructor. The bitset is given its total number of bits and an initializer vector.
 * @param initVect The initialization vector of the bitset.
 */
template < Signedness S >
template < typename W2 >
RtBitSet<S>::RtBitSet(size_t _bitCount, const std::vector<W2>& initVect)
{
	setBitCount(_bitCount);
	setValue(initVect);
}

/**
 * Template constructor. The bitset is given an initializer vector.
 * @param initVect The initialization vector of the bitset.
 */
template < Signedness S >
template < typename W2 >
RtBitSet<S>::RtBitSet(const std::vector<W2>& initVect)
{
	setBitCount(initVect.size()* wordBitCount<W2>());
	setValue(initVect);
}

/**
 * Template constructor. The bitset is given its total number of bits and another bitset to initialize its value.
 * @tparam S2 The signedness of the initialization bitset.
 * @param _bitCount The total number of bit of this RtBitSet instance.
 * @param initRtBitSet The initialization bitset of this bitset.
 */
template < Signedness S >
template < Signedness S2 >
RtBitSet<S>::RtBitSet(size_t _bitCount, const RtBitSet<S2>& initRtBitSet)
{
	setBitCount(_bitCount);
	setValue(initRtBitSet);
}

/**
 * Template constructor. The bitset is given an initializer bitset.
 * @param initRtBitSet The initialization bitset of the bitset.
 */
template < Signedness S >
template < Signedness S2 >
RtBitSet<S>::RtBitSet(const RtBitSet<S2>& initRtBitSet)
{
	setBitCount(initRtBitSet.bitCount());
	setValue(initRtBitSet);
}

/**
 * Initialization function of the RtBitSet. Factors in all initializations that need to be done once the total number
 * of bits of the the bitset is known. This method can be called any other time to modify the bit count of this BitSet.
 */
template < Signedness S >
void RtBitSet<S>::setBitCount(size_t _bitCount)
{
	bitCount_ = _bitCount;
	size_t atomCount_ = bitCount() / atomBitCount();
	if (bitCount() - atomCount_ * atomBitCount() != 0)
	{
		atomCount_ += 1;
	}
	buffer.resize(atomCount_);
	normalize();
	LOG_DEBUG() << "Created run time bitset of " << std::dec << bitCount() << " bits over " << atomCount() << " atoms of " << atomByteCount() << " bytes";
}

/**
 * Normalizes the RtBitSet format, namely by setting the sign bit properly.
 */
template < Signedness S >
void RtBitSet<S>::normalize()
{
	const size_t deadBits = atomBitCount() - (bitCount() % atomBitCount());
	atomLast() <<= deadBits;
	atomLast() >>= deadBits;
}

template < Signedness S >
void RtBitSet<S>::bitNot()
{
	rasterOp<AtomOperator::bitNot, Atom>(bufferPtr<Atom>(), atomCount(), bufferPtr<Atom>(), atomCount());
}

template < Signedness S >
void RtBitSet<S>::shiftLeft(size_t nBits)
{
	iter_Atom itBegin(itAtomBegin());

	size_t nAtomJumps = nBits / atomBitCount();
	size_t nBitJumpsLeft  =  nBits - atomBitCount() * nAtomJumps;
	size_t nBitJumpsRight =  atomBitCount() - nBitJumpsLeft;

	iter_Atom itTarget(itAtomLast());
	iter_Atom itSource(itTarget - nAtomJumps);

	while (itSource > itBegin)
	{
		*itTarget  = *itSource << nBitJumpsLeft;
		*itTarget |= *--itSource >> nBitJumpsRight;
		--itTarget;
	}

	*itTarget = *itSource << nBitJumpsLeft;
	while(--itTarget >= itBegin)
	{
		*itTarget = 0;
	}
}

template < Signedness S >
void RtBitSet<S>::shiftRight(size_t nBits)
{
	iter_Atom itLast(itAtomLast());

	size_t nAtomJumps = nBits / atomBitCount();
	size_t nBitJumpsRight =  nBits - atomBitCount() * nAtomJumps;
	size_t nBitJumpsLeft  =  atomBitCount() - nBitJumpsRight;

	iter_Atom itTarget(itAtomBegin());
	iter_Atom itSource (itTarget + nAtomJumps);

	while (itSource < itLast)
	{
		*itTarget  = static_cast<unsigned_Atom>(*itSource) >> nBitJumpsRight;
		*itTarget |= *++itSource << nBitJumpsLeft;
		++itTarget;
	}

	*itTarget  = *itSource >> nBitJumpsRight;
	Atom padding = *itLast < 0 ? ~0 : 0;
	while(++itTarget <= itLast)
	{
		*itTarget = padding;
	}
}

/**
 * Returns 0 if this bitset is numerically equal to a specified word, 1 if it is numerically greater, -1 otherwise.
 * @tparam S Signedness of this bitset.
 * @tparam W2 Type of the word this bitset is compared to.
 * @param w The word this bitset is compared to.
 */
template < Signedness S >
template < typename W2 >
int RtBitSet<S>::compare(const W2& w) const
{
	if (wordBitCount(w) > bitCount())
	{
		typedef typename boost::conditional
		<
			S == SIGNED,
			typename boost::make_signed <W2>::type,
			typename boost::make_unsigned <W2>::type
		>
		::type ThisWord;

		ThisWord thisWord;
		getValue(thisWord);
		     if (thisWord > w) return  1;
		else if (thisWord < w) return -1;
		                       return  0;
	}
	else
	{
		typename make_RtBitSet<W2>::type bitSet(bitCount(), w);
		return compare(bitSet);
	}
}

/**
 * Returns 0 if this bitset is numerically equal to another specified bitset, 1 if it is numerically greater, -1 otherwise.
 * @tparam S Signedness of this bitset.
 * @tparam W2 Signedness of the bitset this one is compared to.
 * @param w The bitset this one is compared to.
 */
template < Signedness S >
template < Signedness S2 >
int RtBitSet<S>::compare(const RtBitSet<S2>& bitSet) const
{
	if (bitCount() == bitSet.bitCount())
	{
		iter_Atom itBegin = itAtomBegin();
		iter_Atom it = itAtomEnd();
		WordIterator<typename RtBitSet<S2>::Atom> jt = bitSet.itAtomEnd();
		do
		{
			--it; --jt;
			     if (*it > *jt) return  1;
			else if (*it < *jt) return -1;

		} while (it != itBegin);
		return 0;
	}
	else
	{
		if (bitSet.bitCount() > bitCount())
		{
			RtBitSet<S> largerThis(bitSet.bitCount(), *this);
			return largerThis.compare(bitSet); // recursive call
		}
		else
		{
			RtBitSet<S2> largerThat(bitCount(), bitSet);
			return compare(largerThat); // recursive call
		}
	}
}

/**
 * This static method executes in a loop over the source and target array words the same elementary operation (e.g. copy
 * operation). In the case where target array is of a greater size than source array, excess target words are "wiped"
 * according to the signedness of the array word type (if AtomType is signed, the sign bit is extended).
 * @tparam AtomType The type of the elementary element of the arrays.
 * @tparam AtomOp The elementary operation applied (as functor).
 * @param pTarget Pointer to the head of the target array.
 * @param maxAtoms Size of the target array.
 * @param pSource Pointer to the head of the source array.
 * @param nAtoms Size of the source array.
 */
template < Signedness S >
template < class AtomOp, typename AtomType >
void RtBitSet<S>::rasterOp(AtomType* pTarget, size_t maxAtoms, const AtomType* pSource, size_t nAtoms)
{
	AtomType wipeValue = isPositive(*(WordIterator<AtomType>::getEnd(pSource, nAtoms) - 1)) ? 0 : -1;
	std::ptrdiff_t delta = nAtoms - maxAtoms;

	if (delta > 0)
	{
		nAtoms = maxAtoms;
#ifdef WORDS_BIG_ENDIAN
		pSource += delta;
#endif
	}
	else
	{
#ifdef WORDS_BIG_ENDIAN
		AtomType* pWipe = pTarget;
#else
		AtomType* pWipe = pTarget + nAtoms;
#endif
		while (delta++ < 0)
		{
			AtomOp()(*pWipe, wipeValue);
			++pWipe;
		}
	}
	while (nAtoms-- > 0)
	{
		AtomOp()(*pTarget, *pSource);
		++pTarget;
		++pSource;
	}
}

/**
 * Specialized rasterOp method from array of words to bitset.
 * @tparam AtomOp Elementary raster operation (e.g. copy, bitwise or, etc.)
 * @tparam AtomType Elementary word type of the array.
 * @param sourceArray The atom array whose value is operated with this bitset.
 * @param arraySize The size of the array.
 */
template < Signedness S >
template < class AtomOp, typename AtomType >
void RtBitSet<S>::rasterOp(const AtomType* sourceArray, size_t arraySize)
{
	// Define Byte with same sign as source, i.e. AtomType template parameter (here)
	typedef typename boost::conditional
		<
			boost::is_signed<AtomType>::value,
			signed char,
			unsigned char
		>
		::type SourceByte;

	RtBitSet<S>::rasterOp<AtomOp, SourceByte>
	(
		bufferPtr<SourceByte>(), byteCount(), reinterpret_cast<const SourceByte*>(sourceArray), arraySize*sizeof(AtomType)
	);
	normalize();
}

/**
 * Specialized rasterOp method from bitset to array of words.
 * @param targetArray The atom array whose value is operated with this bitset.
 */
template < Signedness S >
template < class AtomOp, typename AtomType >
void RtBitSet<S>::rasterOp(AtomType* targetArray, size_t arraySize) const
{
	RtBitSet<S>::rasterOp<AtomOp, Byte>
	(
		reinterpret_cast<Byte*>(targetArray), arraySize*sizeof(AtomType), bufferPtr<Byte>(), byteCount()
	);
}

/**
 * This method executes the specified elementary operation between a source vector and this bitset (e.g. copy
 * operation). In this case this bitset is modified but not the source vector.
 * @tparam SourceWord The type of the source vector's elementary word.
 * @tparam AtomOp The elementary operation applied (as functor).
 */
template < Signedness S >
template < class AtomOp, typename SourceWord >
void RtBitSet<S>::rasterOp(const std::vector<SourceWord>& sourceVect)
{
	typedef typename boost::conditional
		<
			boost::is_signed<SourceWord>::value,
			signed char,
			unsigned char
		>
		::type SourceByte;

	if (sourceVect.size() > 0)
	{
		RtBitSet<S>::rasterOp<AtomOp, SourceByte>
		(
			bufferPtr<SourceByte>(), byteCount(), reinterpret_cast<const SourceByte*>(&sourceVect[0]), sourceVect.size()*sizeof(SourceWord)
		);
		normalize();
	}
}

/**
 * This method executes the specified elementary operation between this bitset a target vector (e.g. copy operation). In
 * this case the target vector is modified but not this bitset. The size of the target vector is adjusted so that it
 * can contain the whole bitset.
 * @tparam TargetWord The type of the target vector's elementary word.
 * @tparam AtomOp The elementary operation applied (as functor).
 */
template < Signedness S >
template < class AtomOp, typename TargetWord >
void RtBitSet<S>::rasterOp(std::vector<TargetWord>& targetVect) const
{
	size_t lcmBytes = boost::math::lcm(sizeof(TargetWord), byteCount());
	targetVect.resize(lcmBytes / sizeof(TargetWord));
	for (size_t i = 0; i < targetVect.size(); ++i)
	{
		targetVect[i] = 0;
	}
	size_t maxTargetAtoms = lcmBytes / atomByteCount();
	if (targetVect.size() > 0)
	{
		RtBitSet<S>::rasterOp<AtomOp, Atom>
		(
			reinterpret_cast<Atom*>(&targetVect[0]), maxTargetAtoms, bufferPtr<Atom>(), atomCount()
		);
	}
}

/**
 * Specialized rasterOp method from word to bitset method for words of type Atom in its unsigned version.
 * This specialization optimizes the performance of the raster operation by using words bigger than bytes.
 * @param sourceWord The word whose value is operated with this bitset.
 */
template < Signedness S >
template < class AtomOp >
void RtBitSet<S>::rasterOp(const unsigned_Atom& sourceWord)
{
	RtBitSet<S>::rasterOp<AtomOp, unsigned_Atom>(bufferPtr<unsigned_Atom>(), atomCount(), &sourceWord, 1u);
	normalize();
}

/**
 * Specialized rasterOp method from bitset to word for words of type Atom in its unsigned version.
 * This specialization optimizes the performance of the raster operation by using words bigger than bytes.
 * @param targetWord The word whose value is operated with this bitset.
 */
template < Signedness S >
template < class AtomOp >
void RtBitSet<S>::rasterOp(unsigned_Atom& targetWord) const
{
	RtBitSet<S>::rasterOp<AtomOp, unsigned_Atom>(&targetWord, 1u, bufferPtr<unsigned_Atom>(), atomCount());
}

/**
 * Specialized rasterOp method from word to bitset method for words of type Atom in its signed version.
 * This specialization optimizes the performance of the raster operation by using words bigger than bytes.
 * @param sourceWord The word whose value is operated with this bitset.
 */
template < Signedness S >
template < class AtomOp >
void RtBitSet<S>::rasterOp(const signed_Atom& sourceWord)
{
	RtBitSet<S>::rasterOp<AtomOp, signed_Atom>(bufferPtr<signed_Atom>(), atomCount(), &sourceWord, 1u);
	normalize();
}

/**
 * Specialized rasterOp method from bitset to word for words of type Atom in its signed version.
 * This specialization optimizes the performance of the raster operation by using words bigger than bytes.
 * @param sourceWord The word whose value is operated with this bitset.
 */
template < Signedness S >
template < class AtomOp >
void RtBitSet<S>::rasterOp(signed_Atom& targetWord) const
{
	RtBitSet<S>::rasterOp<AtomOp, signed_Atom>(&targetWord, 1u, bufferPtr<signed_Atom>(), atomCount());
}

/**
 * Specialized rasterOp method from bitset to bitset.
 * This specialization optimizes the performance of the raster operation by using words bigger than bytes.
 * @param sourceRtBitSet The bitset whose value is operated with this bitset.
 * @param arraySize The size of the array.
 */
template < Signedness S >
template < class AtomOp, Signedness S_Source >
void RtBitSet<S>::rasterOp(const RtBitSet<S_Source>& sourceRtBitSet)
{
	typedef typename RtBitSet<S_Source>::Atom AtomType;
	RtBitSet<S>::rasterOp<AtomOp, AtomType>
	(
		bufferPtr<AtomType>(), atomCount(), sourceRtBitSet.template bufferPtr<AtomType>(),  sourceRtBitSet.atomCount()
	);
	normalize();
}

template < Signedness S >
void RtBitSet<S>::print(std::ostream& os) const
{
	os << (S == SIGNED ? "(s)" : "(u)");
	iter_Atom it(itAtomEnd());
	iter_Atom itBegin(itAtomBegin());
	os << std::hex << std::setfill('0');
	size_t iBit = atomCount() * sizeof(Atom) * CHAR_BIT - 1;
	while (it != itBegin)
	{
		--it;
		for (int i = atomBitCount()-1; i >= 0; --i)
		{
			if (iBit < bitCount())
			{
				os << ((*it >> i) & 1);
				if (iBit % wordBitCount<Atom>() == 0 and iBit != 0) os << ':';
				else if (iBit % 4 == 0 and iBit != 0) os << '.';
			}
			--iBit;
		}
	}
	os << std::dec;
}

template < Signedness S >
bool RtBitSet<S>::getBit(unsigned pos) const
{
	return BitFieldHelper<Atom>::getBit(atomAt(pos / atomBitCount()), pos % atomBitCount());

}

template < Signedness S >
void RtBitSet<S>::setBit(unsigned pos, bool bitVal)
{
	BitFieldHelper<Atom>::setBit(atomAt(pos / atomBitCount()), pos % atomBitCount(), bitVal);
	normalize();
}

template < Signedness S >
template < typename value_t >
void RtBitSet<S>::getField(unsigned pos, unsigned wid, value_t& value) const
{
	LOG_DEBUG() << "Getting field (" << std::dec << pos << "," << wid << ") in word of " << sizeof(Atom) << " bytes" ;
	std::size_t iAtom = pos / atomBitCount();
	iter_Atom it = itAtomBegin() + iAtom;

	unsigned posInAtom = pos - iAtom * atomBitCount();
	unsigned widInAtom = std::min(wid, atomBitCount() - posInAtom);
	if (widInAtom == wid) // if this is the last part, extraction must have same signedness as value
	{
		value = BitFieldHelper<Atom>::getField(*it++, posInAtom, widInAtom);
	}
	else  // if this is NOT the last part, extraction must be unsigned (to avoid risking sign extension if value_t is signed)
	{
		typename boost::make_unsigned<value_t>::type xtract;
		BitFieldHelper<Atom>::getField(*it++, posInAtom, widInAtom, xtract);
		value = xtract;
		for (unsigned widSum = widInAtom; widSum < wid; widSum += widInAtom)
		{
			widInAtom = std::min(wid - widSum, atomBitCount());
			if (widSum + widInAtom < wid)
			{
				typename boost::make_unsigned<value_t>::type xtract;
				BitFieldHelper<Atom>::getField(*it++, 0, widInAtom, xtract);
				value |= xtract << widSum;
			}
			else // this is the last part => extraction must have same signedness as value
			{
				value_t xtract;
				BitFieldHelper<Atom>::getField(*it++, 0, widInAtom, xtract);
				value |= xtract << widSum;
			}
		}
	}
}

template < Signedness S >
template < typename value_t >
value_t RtBitSet<S>::getField(unsigned pos, unsigned wid) const
{
	value_t val;
	getField(pos, wid, val);
	return val;
}

template < Signedness S >
template < typename value_t >
void RtBitSet<S>::setField(unsigned pos, unsigned wid, const value_t& value)
{
	LOG_DEBUG() << "Setting field (" << std::dec << pos << "," << wid << ") with value 0x" << std::hex << value << " in word of " << sizeof(Atom) << " bytes" ;
	std::size_t iAtom = pos / atomBitCount();
	iter_Atom it = itAtomBegin() + iAtom;

	unsigned posInAtom = pos - iAtom * atomBitCount();
	unsigned widInAtom = std::min(wid, atomBitCount() - posInAtom);

	BitFieldHelper<Atom>::setField(*it++, posInAtom, widInAtom, value);
	for (unsigned widSum = widInAtom; widSum < wid; widSum += widInAtom)
	{
		widInAtom = std::min(wid - widSum, static_cast<unsigned>(atomBitCount()));
		BitFieldHelper<Atom>::setField(*it++, 0, widInAtom, value >> widSum);
	}
	normalize();
}

/**
 * Helper for making bitset with same sign as source word
 */
template < typename W >
struct make_RtBitSet
{
	typedef ::utl::RtBitSet< boost::is_signed<W>::value ? SIGNED : UNSIGNED > type;
};

/**
 * Helper for making bitset with same sign as source vector word
 */
template < typename W >
struct make_RtBitSet < std::vector<W> >
{
	typedef ::utl::RtBitSet< boost::is_signed<W>::value ? SIGNED : UNSIGNED > type;
};

} // namespace utl

template < ::utl::Signedness S >
std::ostream& operator << (std::ostream& os, const ::utl::RtBitSet<S>& bitSet)
{
	bitSet.print(os);
	return os;
}

template < typename WordType, ::utl::Signedness S >
WordType operator | (const WordType& wordLeft, const ::utl::RtBitSet<S>& bitSetRight)
{
	typename ::utl::make_RtBitSet<WordType>::type bitSet(wordLeft);
	bitSet |= bitSetRight;
	WordType w;
	bitSet.getValue(w);
	return w;
}

template < ::utl::Signedness SL, ::utl::Signedness SR >
::utl::RtBitSet<SL> operator | (const ::utl::RtBitSet<SL>& bitSetLeft, const ::utl::RtBitSet<SR>& bitSetRight)
{
	::utl::RtBitSet<SL> bitSet(bitSetLeft);
	bitSet |= bitSetRight;
	return bitSet;
}

template < typename WordType, ::utl::Signedness S >
WordType operator & (const WordType& wordLeft, const ::utl::RtBitSet<S>& bitSetRight)
{
	typename ::utl::make_RtBitSet<WordType>::type bitSet(wordLeft);
	bitSet &= bitSetRight;
	WordType w;
	bitSet.getValue(w);
	return w;
}

template < ::utl::Signedness SL, ::utl::Signedness SR >
::utl::RtBitSet<SL> operator & (const ::utl::RtBitSet<SL>& bitSetLeft, const ::utl::RtBitSet<SR>& bitSetRight)
{
	::utl::RtBitSet<SL> bitSet(bitSetLeft);
	bitSet ^= bitSetRight;
	return bitSet;
}

template < typename WordType, ::utl::Signedness S >
WordType operator ^ (const WordType& wordLeft, const ::utl::RtBitSet<S>& bitSetRight)
{
	typename ::utl::make_RtBitSet<WordType>::type bitSet(wordLeft);
	bitSet ^= bitSetRight;
	WordType w;
	bitSet.getValue(w);
	return w;
}

template < ::utl::Signedness SL, ::utl::Signedness SR >
::utl::RtBitSet<SL> operator ^ (const ::utl::RtBitSet<SL>& bitSetLeft, const ::utl::RtBitSet<SR>& bitSetRight)
{
	::utl::RtBitSet<SL> bitSet(bitSetLeft);
	bitSet ^= bitSetRight;
	return bitSet;
}

template < typename WordType, ::utl::Signedness S >
bool operator > (const WordType& wordLeft, const ::utl::RtBitSet<S>& bitSetRight)
{
	return bitSetRight < wordLeft;
}

template < typename WordType, ::utl::Signedness S >
bool operator < (const WordType& wordLeft, const ::utl::RtBitSet<S>& bitSetRight)
{
	return bitSetRight > wordLeft;
}

template < typename WordType, ::utl::Signedness S >
bool operator >= (const WordType& wordLeft, const ::utl::RtBitSet<S>& bitSetRight)
{
	return bitSetRight <= wordLeft;
}

template < typename WordType, ::utl::Signedness S >
bool operator <= (const WordType& wordLeft, const ::utl::RtBitSet<S>& bitSetRight)
{
	return bitSetRight >= wordLeft;
}

template < typename WordType, ::utl::Signedness S >
bool operator == (const WordType& wordLeft, const ::utl::RtBitSet<S>& bitSetRight)
{
	return bitSetRight == wordLeft;
}

template < typename WordType, ::utl::Signedness S >
bool operator != (const WordType& wordLeft, const ::utl::RtBitSet<S>& bitSetRight)
{
	return bitSetRight != wordLeft;
}

#endif /* utl_RtBitSet_h_INCLUDED */
